/* Generated by Cython 0.29.7 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "/home/travis/miniconda3/envs/ingeotec/lib/python3.6/site-packages/numpy/core/include"
        ],
        "name": "EvoDAG.cython_utils",
        "sources": [
            "EvoDAG/cython_utils.pyx"
        ]
    },
    "module_name": "EvoDAG.cython_utils"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_7"
#define CYTHON_HEX_VERSION 0x001D07F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__EvoDAG__cython_utils
#define __PYX_HAVE_API__EvoDAG__cython_utils
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "EvoDAG/cython_utils.pyx",
  "stringsource",
  "array.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "sparse_array.pxd",
};

/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_11SparseArray_12sparse_array_SparseArray;
struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers;
struct __pyx_obj_6EvoDAG_12cython_utils_Score;
struct __pyx_opt_args_11SparseArray_12sparse_array_11SparseArray_finite;

/* "SparseArray/sparse_array.pxd":19
 * cimport cython
 * from libc cimport math
 * ctypedef double (*two_arguments)(double, double)             # <<<<<<<<<<<<<<
 * ctypedef double (*one_argument)(double)
 * 
 */
typedef double (*__pyx_t_11SparseArray_12sparse_array_two_arguments)(double, double);

/* "SparseArray/sparse_array.pxd":20
 * from libc cimport math
 * ctypedef double (*two_arguments)(double, double)
 * ctypedef double (*one_argument)(double)             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef double (*__pyx_t_11SparseArray_12sparse_array_one_argument)(double);

/* "SparseArray/sparse_array.pxd":176
 *     cpdef SparseArray erfc(self)
 * 
 *     cpdef SparseArray finite(self, bint inplace=?)             # <<<<<<<<<<<<<<
 *     cpdef SparseArray boundaries(self)
 *     cpdef SparseArray copy(self)
 */
struct __pyx_opt_args_11SparseArray_12sparse_array_11SparseArray_finite {
  int __pyx_n;
  int inplace;
};

/* "SparseArray/sparse_array.pxd":116
 * 
 * 
 * cdef class SparseArray:             # <<<<<<<<<<<<<<
 *     cdef public unsigned int non_zero
 *     cdef unsigned int _len
 */
struct __pyx_obj_11SparseArray_12sparse_array_SparseArray {
  PyObject_HEAD
  struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *__pyx_vtab;
  unsigned int non_zero;
  unsigned int _len;
  arrayobject *index;
  arrayobject *data;
};


/* "EvoDAG/cython_utils.pyx":208
 * 
 * 
 * cdef class SelectNumbers:             # <<<<<<<<<<<<<<
 *     cdef array.array data
 *     cdef public Py_ssize_t pos
 */
struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers {
  PyObject_HEAD
  struct __pyx_vtabstruct_6EvoDAG_12cython_utils_SelectNumbers *__pyx_vtab;
  arrayobject *data;
  Py_ssize_t pos;
  Py_ssize_t size;
};


/* "EvoDAG/cython_utils.pyx":238
 * 
 * 
 * cdef class Score:             # <<<<<<<<<<<<<<
 *     cdef public array.array precision
 *     cdef public array.array precision_den
 */
struct __pyx_obj_6EvoDAG_12cython_utils_Score {
  PyObject_HEAD
  struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *__pyx_vtab;
  arrayobject *precision;
  arrayobject *precision_den;
  arrayobject *recall;
  arrayobject *recall_den;
  arrayobject *precision2;
  arrayobject *precision2_den;
  arrayobject *recall2;
  arrayobject *recall2_den;
  double _accuracy;
  double _accuracy2;
  unsigned int y_pos;
  unsigned int hy_pos;
  unsigned int y_size;
  unsigned int hy_size;
  unsigned int nclasses;
};



/* "SparseArray/sparse_array.pxd":116
 * 
 * 
 * cdef class SparseArray:             # <<<<<<<<<<<<<<
 *     cdef public unsigned int non_zero
 *     cdef unsigned int _len
 */

struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray {
  void (*_empty)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, unsigned int, unsigned int);
  void (*fix_size)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, unsigned int);
  double (*getitem)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, unsigned int);
  unsigned int (*union_size)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*union_func)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, __pyx_t_11SparseArray_12sparse_array_two_arguments, __pyx_t_11SparseArray_12sparse_array_one_argument, __pyx_t_11SparseArray_12sparse_array_one_argument, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*add)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*add2)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sub)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*min)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*max)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*div)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*one_argument_func)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, __pyx_t_11SparseArray_12sparse_array_one_argument, int);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*acos)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*asin)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*atan)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*atan2)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*cos)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sin)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*tan)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*cosh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sinh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*tanh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*acosh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*asinh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*atanh)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*hypot)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*exp)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*expm1)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*log)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*log2)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*log10)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*log1p)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*lgamma)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*pow)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sqrt)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sq)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*sign)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*fabs)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*ceil)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*floor)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*trunc)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*erf)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*erfc)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*finite)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch, struct __pyx_opt_args_11SparseArray_12sparse_array_11SparseArray_finite *__pyx_optional_args);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*boundaries)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*copy)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  unsigned int (*intersection_size)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*intersection_func)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, __pyx_t_11SparseArray_12sparse_array_two_arguments, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*full_func)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, __pyx_t_11SparseArray_12sparse_array_two_arguments, __pyx_t_11SparseArray_12sparse_array_one_argument, __pyx_t_11SparseArray_12sparse_array_one_argument, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*mul)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*mul2)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, double, int __pyx_skip_dispatch);
  double (*dot)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  double (*sum)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *(*unit_vector)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  double (*pearson_coefficient)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  double (*cosine_distance)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  double (*SSE)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  double (*SAE)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
  int (*isfinite)(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *__pyx_vtabptr_11SparseArray_12sparse_array_SparseArray;


/* "EvoDAG/cython_utils.pyx":208
 * 
 * 
 * cdef class SelectNumbers:             # <<<<<<<<<<<<<<
 *     cdef array.array data
 *     cdef public Py_ssize_t pos
 */

struct __pyx_vtabstruct_6EvoDAG_12cython_utils_SelectNumbers {
  PyObject *(*get)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, Py_ssize_t, int __pyx_skip_dispatch);
  int (*get_one)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, int __pyx_skip_dispatch);
  int (*empty)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6EvoDAG_12cython_utils_SelectNumbers *__pyx_vtabptr_6EvoDAG_12cython_utils_SelectNumbers;


/* "EvoDAG/cython_utils.pyx":238
 * 
 * 
 * cdef class Score:             # <<<<<<<<<<<<<<
 *     cdef public array.array precision
 *     cdef public array.array precision_den
 */

struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score {
  PyObject *(*_cleanI)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, arrayobject *);
  PyObject *(*_cleand)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, arrayobject *);
  PyObject *(*clean)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *);
  double (*get_klass)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, double *, unsigned int *, unsigned int);
  double (*get_predicted)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, double *, unsigned int *, unsigned int);
  PyObject *(*recall2accuracy)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *);
  PyObject *(*precision_recall)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *);
  PyObject *(*count)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, arrayobject *);
};
static struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *__pyx_vtabptr_6EvoDAG_12cython_utils_Score;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    op->ob_size = size;
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    self->ob_size = n;
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        self->ob_size = n;
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    self->ob_size = n;
    self->allocated = newsize;
    return 0;
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, Py_ssize_t __pyx_v_k, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get_one(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_empty(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score__cleanI(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, arrayobject *__pyx_v_d); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score__cleand(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, arrayobject *__pyx_v_d); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_clean(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto*/
static double __pyx_f_6EvoDAG_12cython_utils_5Score_get_klass(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, double *__pyx_v_data, unsigned int *__pyx_v_index, unsigned int __pyx_v_i); /* proto*/
static double __pyx_f_6EvoDAG_12cython_utils_5Score_get_predicted(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, double *__pyx_v_data, unsigned int *__pyx_v_index, unsigned int __pyx_v_i); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_recall2accuracy(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_precision_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_count(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto*/

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'array' */

/* Module declarations from 'cpython.array' */
static PyTypeObject *__pyx_ptype_7cpython_5array_array = 0;
static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *, Py_ssize_t, int); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'SparseArray.sparse_array' */
static PyTypeObject *__pyx_ptype_11SparseArray_12sparse_array_SparseArray = 0;

/* Module declarations from 'EvoDAG.cython_utils' */
static PyTypeObject *__pyx_ptype_6EvoDAG_12cython_utils_SelectNumbers = 0;
static PyTypeObject *__pyx_ptype_6EvoDAG_12cython_utils_Score = 0;
static double __pyx_f_6EvoDAG_12cython_utils_fitness_SSE(PyObject *, PyObject *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_6EvoDAG_12cython_utils_fitness_SAE(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite(PyObject *, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite_MN(PyObject *, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_mean_std2(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, arrayobject *, arrayobject *, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes(PyObject *, PyObject *, unsigned int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_Nc(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, arrayobject *, arrayobject *, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_MN(PyObject *, PyObject *, unsigned int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_centroid_coef(PyObject *, arrayobject *, arrayobject *, int, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "EvoDAG.cython_utils"
extern int __pyx_module_is_main_EvoDAG__cython_utils;
int __pyx_module_is_main_EvoDAG__cython_utils = 0;

/* Implementation of 'EvoDAG.cython_utils' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_MemoryError;
static const char __pyx_k_I[] = "I";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_hy[] = "_hy";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_lst[] = "lst";
static const char __pyx_k_var[] = "var";
static const char __pyx_k_ytr[] = "_ytr";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_coef[] = "coef";
static const char __pyx_k_hy_2[] = "hy";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "_mask";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_vars[] = "vars";
static const char __pyx_k_Score[] = "Score";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_klass[] = "klass";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_mask_2[] = "mask";
static const char __pyx_k_nclass[] = "nclass";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_get_one[] = "get_one";
static const char __pyx_k_shuffle[] = "shuffle";
static const char __pyx_k_constant[] = "constant";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_nclasses[] = "nclasses";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_SelectNumbers[] = "SelectNumbers";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_n_s_I;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_Score;
static PyObject *__pyx_n_s_SelectNumbers;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_coef;
static PyObject *__pyx_n_s_constant;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_one;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_hy;
static PyObject *__pyx_n_s_hy_2;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_klass;
static PyObject *__pyx_n_s_lst;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask_2;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_nclass;
static PyObject *__pyx_n_s_nclasses;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shuffle;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_var;
static PyObject *__pyx_n_s_vars;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_ytr;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_fitness_SSE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v__mask); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_2fitness_SAE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, PyObject *__pyx_v__mask); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_4naive_bayes_isfinite(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coef, int __pyx_v_nclass); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_6naive_bayes_isfinite_MN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coef, int __pyx_v_nclass); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_8naive_bayes_mean_std2(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_10naive_bayes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_12naive_bayes_Nc(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_14naive_bayes_MN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers___cinit__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_lst); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_2get(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, Py_ssize_t __pyx_v_k); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4get_one(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_6empty(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos___get__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size___get__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score___cinit__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, unsigned int __pyx_v_nclasses); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_2g_recall_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, Py_ssize_t __pyx_v_i, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_4F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, Py_ssize_t __pyx_v_i, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_6macroRecallF1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_8a_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10a_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_12g_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_14g_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_16accuracy(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_18accDotMacroF1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_20g_g_recall_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_22g_F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_24a_F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_9precision___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_6recall___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6EvoDAG_12cython_utils_16centroid_coef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_vars, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_6EvoDAG_12cython_utils_SelectNumbers(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6EvoDAG_12cython_utils_Score(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
/* Late includes */

/* "EvoDAG/cython_utils.pyx":25
 * 
 * 
 * cpdef double fitness_SSE(list _ytr, list _hy, SparseArray _mask):             # <<<<<<<<<<<<<<
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_1fitness_SSE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_6EvoDAG_12cython_utils_fitness_SSE(PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v__mask, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_ytr = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  double __pyx_v_res;
  Py_ssize_t __pyx_v_i;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("fitness_SSE", 0);

  /* "EvoDAG/cython_utils.pyx":27
 * cpdef double fitness_SSE(list _ytr, list _hy, SparseArray _mask):
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i
 *     for i in range(PyList_GET_SIZE(_ytr)):
 */
  __pyx_v_res = 0.0;

  /* "EvoDAG/cython_utils.pyx":29
 *     cdef double res = 0
 *     cdef Py_ssize_t i
 *     for i in range(PyList_GET_SIZE(_ytr)):             # <<<<<<<<<<<<<<
 *         ytr = <SparseArray> PyList_GET_ITEM(_ytr, i)
 *         hy = <SparseArray> PyList_GET_ITEM(_hy, i)
 */
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v__ytr);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "EvoDAG/cython_utils.pyx":30
 *     cdef Py_ssize_t i
 *     for i in range(PyList_GET_SIZE(_ytr)):
 *         ytr = <SparseArray> PyList_GET_ITEM(_ytr, i)             # <<<<<<<<<<<<<<
 *         hy = <SparseArray> PyList_GET_ITEM(_hy, i)
 *         res += -ytr.SSE(hy.mul(_mask))
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v__ytr, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_ytr, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "EvoDAG/cython_utils.pyx":31
 *     for i in range(PyList_GET_SIZE(_ytr)):
 *         ytr = <SparseArray> PyList_GET_ITEM(_ytr, i)
 *         hy = <SparseArray> PyList_GET_ITEM(_hy, i)             # <<<<<<<<<<<<<<
 *         res += -ytr.SSE(hy.mul(_mask))
 *     return res / len(_ytr)
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v__hy, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_hy, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "EvoDAG/cython_utils.pyx":32
 *         ytr = <SparseArray> PyList_GET_ITEM(_ytr, i)
 *         hy = <SparseArray> PyList_GET_ITEM(_hy, i)
 *         res += -ytr.SSE(hy.mul(_mask))             # <<<<<<<<<<<<<<
 *     return res / len(_ytr)
 * 
 */
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_hy->__pyx_vtab)->mul(__pyx_v_hy, __pyx_v__mask, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_res = (__pyx_v_res + (-((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_ytr->__pyx_vtab)->SSE(__pyx_v_ytr, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_5), 0)));
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "EvoDAG/cython_utils.pyx":33
 *         hy = <SparseArray> PyList_GET_ITEM(_hy, i)
 *         res += -ytr.SSE(hy.mul(_mask))
 *     return res / len(_ytr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v__ytr == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 33, __pyx_L1_error)
  }
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v__ytr); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 33, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 33, __pyx_L1_error)
  }
  __pyx_r = (__pyx_v_res / __pyx_t_1);
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":25
 * 
 * 
 * cpdef double fitness_SSE(list _ytr, list _hy, SparseArray _mask):             # <<<<<<<<<<<<<<
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("EvoDAG.cython_utils.fitness_SSE", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ytr);
  __Pyx_XDECREF((PyObject *)__pyx_v_hy);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_1fitness_SSE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_1fitness_SSE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v__ytr = 0;
  PyObject *__pyx_v__hy = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v__mask = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fitness_SSE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ytr,&__pyx_n_s_hy,&__pyx_n_s_mask,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ytr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fitness_SSE", 1, 3, 3, 1); __PYX_ERR(0, 25, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fitness_SSE", 1, 3, 3, 2); __PYX_ERR(0, 25, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fitness_SSE") < 0)) __PYX_ERR(0, 25, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v__ytr = ((PyObject*)values[0]);
    __pyx_v__hy = ((PyObject*)values[1]);
    __pyx_v__mask = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fitness_SSE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 25, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.fitness_SSE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__ytr), (&PyList_Type), 1, "_ytr", 1))) __PYX_ERR(0, 25, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__hy), (&PyList_Type), 1, "_hy", 1))) __PYX_ERR(0, 25, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__mask), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "_mask", 0))) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_fitness_SSE(__pyx_self, __pyx_v__ytr, __pyx_v__hy, __pyx_v__mask);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_fitness_SSE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v__mask) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("fitness_SSE", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6EvoDAG_12cython_utils_fitness_SSE(__pyx_v__ytr, __pyx_v__hy, __pyx_v__mask, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.fitness_SSE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":36
 * 
 * 
 * cpdef double fitness_SAE(list _ytr, list _hy, list _mask):             # <<<<<<<<<<<<<<
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_3fitness_SAE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_6EvoDAG_12cython_utils_fitness_SAE(PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, PyObject *__pyx_v__mask, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_ytr = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_mask = 0;
  double __pyx_v_res;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  __Pyx_RefNannySetupContext("fitness_SAE", 0);

  /* "EvoDAG/cython_utils.pyx":38
 * cpdef double fitness_SAE(list _ytr, list _hy, list _mask):
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0             # <<<<<<<<<<<<<<
 *     for ytr, hy, mask in zip(_ytr, _hy, _mask):
 *         res += -ytr.SAE(hy.mul(mask))
 */
  __pyx_v_res = 0.0;

  /* "EvoDAG/cython_utils.pyx":39
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 *     for ytr, hy, mask in zip(_ytr, _hy, _mask):             # <<<<<<<<<<<<<<
 *         res += -ytr.SAE(hy.mul(mask))
 *     return res / len(_ytr)
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v__ytr);
  __Pyx_GIVEREF(__pyx_v__ytr);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__ytr);
  __Pyx_INCREF(__pyx_v__hy);
  __Pyx_GIVEREF(__pyx_v__hy);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v__hy);
  __Pyx_INCREF(__pyx_v__mask);
  __Pyx_GIVEREF(__pyx_v__mask);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v__mask);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 39, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 39, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 39, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 39, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 39, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_11SparseArray_12sparse_array_SparseArray))))) __PYX_ERR(0, 39, __pyx_L1_error)
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_11SparseArray_12sparse_array_SparseArray))))) __PYX_ERR(0, 39, __pyx_L1_error)
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_11SparseArray_12sparse_array_SparseArray))))) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_ytr, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_hy, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_6));
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_mask, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "EvoDAG/cython_utils.pyx":40
 *     cdef double res = 0
 *     for ytr, hy, mask in zip(_ytr, _hy, _mask):
 *         res += -ytr.SAE(hy.mul(mask))             # <<<<<<<<<<<<<<
 *     return res / len(_ytr)
 * 
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_hy->__pyx_vtab)->mul(__pyx_v_hy, __pyx_v_mask, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_res = (__pyx_v_res + (-((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_ytr->__pyx_vtab)->SAE(__pyx_v_ytr, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2), 0)));
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EvoDAG/cython_utils.pyx":39
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 *     for ytr, hy, mask in zip(_ytr, _hy, _mask):             # <<<<<<<<<<<<<<
 *         res += -ytr.SAE(hy.mul(mask))
 *     return res / len(_ytr)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":41
 *     for ytr, hy, mask in zip(_ytr, _hy, _mask):
 *         res += -ytr.SAE(hy.mul(mask))
 *     return res / len(_ytr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v__ytr == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 41, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_v__ytr); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 41, __pyx_L1_error)
  if (unlikely(__pyx_t_3 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 41, __pyx_L1_error)
  }
  __pyx_r = (__pyx_v_res / __pyx_t_3);
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":36
 * 
 * 
 * cpdef double fitness_SAE(list _ytr, list _hy, list _mask):             # <<<<<<<<<<<<<<
 *     cdef SparseArray ytr, hy, mask
 *     cdef double res = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("EvoDAG.cython_utils.fitness_SAE", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ytr);
  __Pyx_XDECREF((PyObject *)__pyx_v_hy);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_3fitness_SAE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_3fitness_SAE(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v__ytr = 0;
  PyObject *__pyx_v__hy = 0;
  PyObject *__pyx_v__mask = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fitness_SAE (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ytr,&__pyx_n_s_hy,&__pyx_n_s_mask,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ytr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fitness_SAE", 1, 3, 3, 1); __PYX_ERR(0, 36, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("fitness_SAE", 1, 3, 3, 2); __PYX_ERR(0, 36, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fitness_SAE") < 0)) __PYX_ERR(0, 36, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v__ytr = ((PyObject*)values[0]);
    __pyx_v__hy = ((PyObject*)values[1]);
    __pyx_v__mask = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fitness_SAE", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 36, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.fitness_SAE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__ytr), (&PyList_Type), 1, "_ytr", 1))) __PYX_ERR(0, 36, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__hy), (&PyList_Type), 1, "_hy", 1))) __PYX_ERR(0, 36, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v__mask), (&PyList_Type), 1, "_mask", 1))) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_2fitness_SAE(__pyx_self, __pyx_v__ytr, __pyx_v__hy, __pyx_v__mask);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_2fitness_SAE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v__ytr, PyObject *__pyx_v__hy, PyObject *__pyx_v__mask) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("fitness_SAE", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6EvoDAG_12cython_utils_fitness_SAE(__pyx_v__ytr, __pyx_v__hy, __pyx_v__mask, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.fitness_SAE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":44
 * 
 * 
 * cpdef bint naive_bayes_isfinite(list coef, int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j=0
 *     cdef array.array c
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5naive_bayes_isfinite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite(PyObject *__pyx_v_coef, int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  arrayobject *__pyx_v_c = 0;
  double *__pyx_v_c_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("naive_bayes_isfinite", 0);

  /* "EvoDAG/cython_utils.pyx":45
 * 
 * cpdef bint naive_bayes_isfinite(list coef, int nclass):
 *     cdef Py_ssize_t i, j=0             # <<<<<<<<<<<<<<
 *     cdef array.array c
 *     cdef double *c_value
 */
  __pyx_v_j = 0;

  /* "EvoDAG/cython_utils.pyx":48
 *     cdef array.array c
 *     cdef double *c_value
 *     for j in range(1, 3):             # <<<<<<<<<<<<<<
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles
 */
  for (__pyx_t_1 = 1; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "EvoDAG/cython_utils.pyx":49
 *     cdef double *c_value
 *     for j in range(1, 3):
 *         c = <array.array> PyList_GET_ITEM(coef, j)             # <<<<<<<<<<<<<<
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 */
    __pyx_t_2 = PyList_GET_ITEM(__pyx_v_coef, __pyx_v_j);
    __pyx_t_3 = ((PyObject *)__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_c, ((arrayobject *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "EvoDAG/cython_utils.pyx":50
 *     for j in range(1, 3):
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles             # <<<<<<<<<<<<<<
 *         for i in range(nclass):
 *             if c_value[i] == 0:
 */
    __pyx_t_4 = __pyx_v_c->data.as_doubles;
    __pyx_v_c_value = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":51
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):             # <<<<<<<<<<<<<<
 *             if c_value[i] == 0:
 *                 return False
 */
    __pyx_t_5 = __pyx_v_nclass;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "EvoDAG/cython_utils.pyx":52
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 *             if c_value[i] == 0:             # <<<<<<<<<<<<<<
 *                 return False
 *     return True
 */
      __pyx_t_8 = (((__pyx_v_c_value[__pyx_v_i]) == 0.0) != 0);
      if (__pyx_t_8) {

        /* "EvoDAG/cython_utils.pyx":53
 *         for i in range(nclass):
 *             if c_value[i] == 0:
 *                 return False             # <<<<<<<<<<<<<<
 *     return True
 * 
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "EvoDAG/cython_utils.pyx":52
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 *             if c_value[i] == 0:             # <<<<<<<<<<<<<<
 *                 return False
 *     return True
 */
      }
    }
  }

  /* "EvoDAG/cython_utils.pyx":54
 *             if c_value[i] == 0:
 *                 return False
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":44
 * 
 * 
 * cpdef bint naive_bayes_isfinite(list coef, int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j=0
 *     cdef array.array c
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5naive_bayes_isfinite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5naive_bayes_isfinite(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_coef = 0;
  int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes_isfinite (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coef,&__pyx_n_s_nclass,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coef)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_isfinite", 1, 2, 2, 1); __PYX_ERR(0, 44, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes_isfinite") < 0)) __PYX_ERR(0, 44, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coef = ((PyObject*)values[0]);
    __pyx_v_nclass = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nclass == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 44, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes_isfinite", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 44, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_isfinite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coef), (&PyList_Type), 1, "coef", 1))) __PYX_ERR(0, 44, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_4naive_bayes_isfinite(__pyx_self, __pyx_v_coef, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_4naive_bayes_isfinite(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coef, int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_isfinite", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite(__pyx_v_coef, __pyx_v_nclass, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_isfinite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":57
 * 
 * 
 * cpdef bint naive_bayes_isfinite_MN(list coef, int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j=0
 *     cdef array.array c
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_7naive_bayes_isfinite_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite_MN(PyObject *__pyx_v_coef, int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  arrayobject *__pyx_v_c = 0;
  double *__pyx_v_c_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("naive_bayes_isfinite_MN", 0);

  /* "EvoDAG/cython_utils.pyx":58
 * 
 * cpdef bint naive_bayes_isfinite_MN(list coef, int nclass):
 *     cdef Py_ssize_t i, j=0             # <<<<<<<<<<<<<<
 *     cdef array.array c
 *     cdef double *c_value
 */
  __pyx_v_j = 0;

  /* "EvoDAG/cython_utils.pyx":61
 *     cdef array.array c
 *     cdef double *c_value
 *     for j in range(2):             # <<<<<<<<<<<<<<
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "EvoDAG/cython_utils.pyx":62
 *     cdef double *c_value
 *     for j in range(2):
 *         c = <array.array> PyList_GET_ITEM(coef, j)             # <<<<<<<<<<<<<<
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 */
    __pyx_t_2 = PyList_GET_ITEM(__pyx_v_coef, __pyx_v_j);
    __pyx_t_3 = ((PyObject *)__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_c, ((arrayobject *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "EvoDAG/cython_utils.pyx":63
 *     for j in range(2):
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles             # <<<<<<<<<<<<<<
 *         for i in range(nclass):
 *             if not math.isfinite(c_value[i]):
 */
    __pyx_t_4 = __pyx_v_c->data.as_doubles;
    __pyx_v_c_value = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":64
 *         c = <array.array> PyList_GET_ITEM(coef, j)
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):             # <<<<<<<<<<<<<<
 *             if not math.isfinite(c_value[i]):
 *                 return False
 */
    __pyx_t_5 = __pyx_v_nclass;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "EvoDAG/cython_utils.pyx":65
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 *             if not math.isfinite(c_value[i]):             # <<<<<<<<<<<<<<
 *                 return False
 *     return True
 */
      __pyx_t_8 = ((!(isfinite((__pyx_v_c_value[__pyx_v_i])) != 0)) != 0);
      if (__pyx_t_8) {

        /* "EvoDAG/cython_utils.pyx":66
 *         for i in range(nclass):
 *             if not math.isfinite(c_value[i]):
 *                 return False             # <<<<<<<<<<<<<<
 *     return True
 * 
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "EvoDAG/cython_utils.pyx":65
 *         c_value = c.data.as_doubles
 *         for i in range(nclass):
 *             if not math.isfinite(c_value[i]):             # <<<<<<<<<<<<<<
 *                 return False
 *     return True
 */
      }
    }
  }

  /* "EvoDAG/cython_utils.pyx":67
 *             if not math.isfinite(c_value[i]):
 *                 return False
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":57
 * 
 * 
 * cpdef bint naive_bayes_isfinite_MN(list coef, int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j=0
 *     cdef array.array c
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_7naive_bayes_isfinite_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_7naive_bayes_isfinite_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_coef = 0;
  int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes_isfinite_MN (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coef,&__pyx_n_s_nclass,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coef)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_isfinite_MN", 1, 2, 2, 1); __PYX_ERR(0, 57, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes_isfinite_MN") < 0)) __PYX_ERR(0, 57, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coef = ((PyObject*)values[0]);
    __pyx_v_nclass = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nclass == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes_isfinite_MN", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 57, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_isfinite_MN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coef), (&PyList_Type), 1, "coef", 1))) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_6naive_bayes_isfinite_MN(__pyx_self, __pyx_v_coef, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_6naive_bayes_isfinite_MN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coef, int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_isfinite_MN", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6EvoDAG_12cython_utils_naive_bayes_isfinite_MN(__pyx_v_coef, __pyx_v_nclass, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_isfinite_MN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":70
 * 
 * 
 * cpdef list naive_bayes_mean_std2(SparseArray var, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den, std_num
 *     cdef unsigned int *a = var.index.data.as_uints, index
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_9naive_bayes_mean_std2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_mean_std2(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  arrayobject *__pyx_v_mean_num = 0;
  arrayobject *__pyx_v_mean_den = 0;
  arrayobject *__pyx_v_std_num = 0;
  unsigned int *__pyx_v_a;
  unsigned int __pyx_v_index;
  unsigned int *__pyx_v_klass_value;
  double *__pyx_v_a_value;
  double __pyx_v_tmp;
  double *__pyx_v_mean_num_value;
  double *__pyx_v_mean_den_value;
  double *__pyx_v_std_num_value;
  unsigned int *__pyx_v_mask_value;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_k;
  unsigned int __pyx_v_var_end;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int *__pyx_t_1;
  double *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  unsigned int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  double __pyx_t_15;
  __Pyx_RefNannySetupContext("naive_bayes_mean_std2", 0);

  /* "EvoDAG/cython_utils.pyx":72
 * cpdef list naive_bayes_mean_std2(SparseArray var, array.array klass, array.array mask, int nclass):
 *     cdef array.array mean_num, mean_den, std_num
 *     cdef unsigned int *a = var.index.data.as_uints, index             # <<<<<<<<<<<<<<
 *     cdef unsigned int *klass_value = klass.data.as_uints
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 */
  __pyx_t_1 = __pyx_v_var->index->data.as_uints;
  __pyx_v_a = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":73
 *     cdef array.array mean_num, mean_den, std_num
 *     cdef unsigned int *a = var.index.data.as_uints, index
 *     cdef unsigned int *klass_value = klass.data.as_uints             # <<<<<<<<<<<<<<
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 *     cdef double *mean_num_value, *mean_den_value, *std_num_value
 */
  __pyx_t_1 = __pyx_v_klass->data.as_uints;
  __pyx_v_klass_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":74
 *     cdef unsigned int *a = var.index.data.as_uints, index
 *     cdef unsigned int *klass_value = klass.data.as_uints
 *     cdef double *a_value = var.data.data.as_doubles, tmp             # <<<<<<<<<<<<<<
 *     cdef double *mean_num_value, *mean_den_value, *std_num_value
 *     cdef unsigned int *mask_value = mask.data.as_uints
 */
  __pyx_t_2 = __pyx_v_var->data->data.as_doubles;
  __pyx_v_a_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":76
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 *     cdef double *mean_num_value, *mean_den_value, *std_num_value
 *     cdef unsigned int *mask_value = mask.data.as_uints             # <<<<<<<<<<<<<<
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_mask->data.as_uints;
  __pyx_v_mask_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":77
 *     cdef double *mean_num_value, *mean_den_value, *std_num_value
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     mean_num = array.clone(var.data, nclass, zero=True)             # <<<<<<<<<<<<<<
 *     mean_num_value = mean_num.data.as_doubles
 *     std_num = array.clone(var.data, nclass, zero=True)
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_3), __pyx_v_nclass, 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_num = ((arrayobject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "EvoDAG/cython_utils.pyx":78
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles             # <<<<<<<<<<<<<<
 *     std_num = array.clone(var.data, nclass, zero=True)
 *     std_num_value = std_num.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_mean_num->data.as_doubles;
  __pyx_v_mean_num_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":79
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles
 *     std_num = array.clone(var.data, nclass, zero=True)             # <<<<<<<<<<<<<<
 *     std_num_value = std_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_4), __pyx_v_nclass, 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_std_num = ((arrayobject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":80
 *     mean_num_value = mean_num.data.as_doubles
 *     std_num = array.clone(var.data, nclass, zero=True)
 *     std_num_value = std_num.data.as_doubles             # <<<<<<<<<<<<<<
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_std_num->data.as_doubles;
  __pyx_v_std_num_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":81
 *     std_num = array.clone(var.data, nclass, zero=True)
 *     std_num_value = std_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)             # <<<<<<<<<<<<<<
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_3), __pyx_v_nclass, 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_den = ((arrayobject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "EvoDAG/cython_utils.pyx":82
 *     std_num_value = std_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 */
  __pyx_t_2 = __pyx_v_mean_den->data.as_doubles;
  __pyx_v_mean_den_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":83
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0             # <<<<<<<<<<<<<<
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:
 */
  __pyx_v_k = 0;

  /* "EvoDAG/cython_utils.pyx":84
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero             # <<<<<<<<<<<<<<
 *     if var_end == 0:
 *         return [mean_num, std_num, mean_den]
 */
  __pyx_t_5 = __pyx_v_var->non_zero;
  __pyx_v_var_end = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":85
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:             # <<<<<<<<<<<<<<
 *         return [mean_num, std_num, mean_den]
 *     for i in range(len(mask)):
 */
  __pyx_t_6 = ((__pyx_v_var_end == 0) != 0);
  if (__pyx_t_6) {

    /* "EvoDAG/cython_utils.pyx":86
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:
 *         return [mean_num, std_num, mean_den]             # <<<<<<<<<<<<<<
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)__pyx_v_mean_num));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_num));
    PyList_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_mean_num));
    __Pyx_INCREF(((PyObject *)__pyx_v_std_num));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_std_num));
    PyList_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_std_num));
    __Pyx_INCREF(((PyObject *)__pyx_v_mean_den));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_den));
    PyList_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_mean_den));
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "EvoDAG/cython_utils.pyx":85
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:             # <<<<<<<<<<<<<<
 *         return [mean_num, std_num, mean_den]
 *     for i in range(len(mask)):
 */
  }

  /* "EvoDAG/cython_utils.pyx":87
 *     if var_end == 0:
 *         return [mean_num, std_num, mean_den]
 *     for i in range(len(mask)):             # <<<<<<<<<<<<<<
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 */
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 87, __pyx_L1_error)
  }
  __pyx_t_7 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "EvoDAG/cython_utils.pyx":88
 *         return [mean_num, std_num, mean_den]
 *     for i in range(len(mask)):
 *         index = mask_value[i]             # <<<<<<<<<<<<<<
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 */
    __pyx_v_index = (__pyx_v_mask_value[__pyx_v_i]);

    /* "EvoDAG/cython_utils.pyx":89
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 *         for k in range(k, var_end):             # <<<<<<<<<<<<<<
 *             if a[k] >= index:
 *                 break
 */
    __pyx_t_5 = __pyx_v_var_end;
    __pyx_t_10 = __pyx_t_5;
    for (__pyx_t_11 = __pyx_v_k; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_k = __pyx_t_11;

      /* "EvoDAG/cython_utils.pyx":90
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index:
 */
      __pyx_t_6 = (((__pyx_v_a[__pyx_v_k]) >= __pyx_v_index) != 0);
      if (__pyx_t_6) {

        /* "EvoDAG/cython_utils.pyx":91
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 *                 break             # <<<<<<<<<<<<<<
 *         if a[k] == index:
 *             mean_num_value[klass_value[index]] += a_value[k]
 */
        goto __pyx_L7_break;

        /* "EvoDAG/cython_utils.pyx":90
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index:
 */
      }
    }
    __pyx_L7_break:;

    /* "EvoDAG/cython_utils.pyx":92
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index:             # <<<<<<<<<<<<<<
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             mean_den_value[klass_value[index]] += 1
 */
    __pyx_t_6 = (((__pyx_v_a[__pyx_v_k]) == __pyx_v_index) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":93
 *                 break
 *         if a[k] == index:
 *             mean_num_value[klass_value[index]] += a_value[k]             # <<<<<<<<<<<<<<
 *             mean_den_value[klass_value[index]] += 1
 *         else:
 */
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_num_value[__pyx_t_5]) = ((__pyx_v_mean_num_value[__pyx_t_5]) + (__pyx_v_a_value[__pyx_v_k]));

      /* "EvoDAG/cython_utils.pyx":94
 *         if a[k] == index:
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *         else:
 *             mean_den_value[klass_value[index]] += 1
 */
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_den_value[__pyx_t_5]) = ((__pyx_v_mean_den_value[__pyx_t_5]) + 1.0);

      /* "EvoDAG/cython_utils.pyx":92
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index:             # <<<<<<<<<<<<<<
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             mean_den_value[klass_value[index]] += 1
 */
      goto __pyx_L9;
    }

    /* "EvoDAG/cython_utils.pyx":96
 *             mean_den_value[klass_value[index]] += 1
 *         else:
 *             mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         mean_num_value[i] = mean_num_value[i] / mean_den_value[i]
 */
    /*else*/ {
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_den_value[__pyx_t_5]) = ((__pyx_v_mean_den_value[__pyx_t_5]) + 1.0);
    }
    __pyx_L9:;
  }

  /* "EvoDAG/cython_utils.pyx":97
 *         else:
 *             mean_den_value[klass_value[index]] += 1
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         mean_num_value[i] = mean_num_value[i] / mean_den_value[i]
 *     k = 0
 */
  __pyx_t_12 = __pyx_v_nclass;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":98
 *             mean_den_value[klass_value[index]] += 1
 *     for i in range(nclass):
 *         mean_num_value[i] = mean_num_value[i] / mean_den_value[i]             # <<<<<<<<<<<<<<
 *     k = 0
 *     for i in range(len(mask)):
 */
    if (unlikely((__pyx_v_mean_den_value[__pyx_v_i]) == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 98, __pyx_L1_error)
    }
    (__pyx_v_mean_num_value[__pyx_v_i]) = ((__pyx_v_mean_num_value[__pyx_v_i]) / (__pyx_v_mean_den_value[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":99
 *     for i in range(nclass):
 *         mean_num_value[i] = mean_num_value[i] / mean_den_value[i]
 *     k = 0             # <<<<<<<<<<<<<<
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 */
  __pyx_v_k = 0;

  /* "EvoDAG/cython_utils.pyx":100
 *         mean_num_value[i] = mean_num_value[i] / mean_den_value[i]
 *     k = 0
 *     for i in range(len(mask)):             # <<<<<<<<<<<<<<
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 */
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 100, __pyx_L1_error)
  }
  __pyx_t_7 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "EvoDAG/cython_utils.pyx":101
 *     k = 0
 *     for i in range(len(mask)):
 *         index = mask_value[i]             # <<<<<<<<<<<<<<
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 */
    __pyx_v_index = (__pyx_v_mask_value[__pyx_v_i]);

    /* "EvoDAG/cython_utils.pyx":102
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 *         for k in range(k, var_end):             # <<<<<<<<<<<<<<
 *             if a[k] >= index:
 *                 break
 */
    __pyx_t_5 = __pyx_v_var_end;
    __pyx_t_10 = __pyx_t_5;
    for (__pyx_t_11 = __pyx_v_k; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_k = __pyx_t_11;

      /* "EvoDAG/cython_utils.pyx":103
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index:
 */
      __pyx_t_6 = (((__pyx_v_a[__pyx_v_k]) >= __pyx_v_index) != 0);
      if (__pyx_t_6) {

        /* "EvoDAG/cython_utils.pyx":104
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 *                 break             # <<<<<<<<<<<<<<
 *         if a[k] == index:
 *             tmp = a_value[k] - mean_num[klass_value[index]]
 */
        goto __pyx_L15_break;

        /* "EvoDAG/cython_utils.pyx":103
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index:
 */
      }
    }
    __pyx_L15_break:;

    /* "EvoDAG/cython_utils.pyx":105
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index:             # <<<<<<<<<<<<<<
 *             tmp = a_value[k] - mean_num[klass_value[index]]
 *         else:
 */
    __pyx_t_6 = (((__pyx_v_a[__pyx_v_k]) == __pyx_v_index) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":106
 *                 break
 *         if a[k] == index:
 *             tmp = a_value[k] - mean_num[klass_value[index]]             # <<<<<<<<<<<<<<
 *         else:
 *             tmp = mean_num[klass_value[index]]
 */
      __pyx_t_4 = PyFloat_FromDouble((__pyx_v_a_value[__pyx_v_k])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_GetItemInt(((PyObject *)__pyx_v_mean_num), (__pyx_v_klass_value[__pyx_v_index]), unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = PyNumber_Subtract(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_v_tmp = __pyx_t_15;

      /* "EvoDAG/cython_utils.pyx":105
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index:             # <<<<<<<<<<<<<<
 *             tmp = a_value[k] - mean_num[klass_value[index]]
 *         else:
 */
      goto __pyx_L17;
    }

    /* "EvoDAG/cython_utils.pyx":108
 *             tmp = a_value[k] - mean_num[klass_value[index]]
 *         else:
 *             tmp = mean_num[klass_value[index]]             # <<<<<<<<<<<<<<
 *         tmp *= tmp
 *         std_num_value[klass_value[index]] += tmp
 */
    /*else*/ {
      __pyx_t_14 = __Pyx_GetItemInt(((PyObject *)__pyx_v_mean_num), (__pyx_v_klass_value[__pyx_v_index]), unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_v_tmp = __pyx_t_15;
    }
    __pyx_L17:;

    /* "EvoDAG/cython_utils.pyx":109
 *         else:
 *             tmp = mean_num[klass_value[index]]
 *         tmp *= tmp             # <<<<<<<<<<<<<<
 *         std_num_value[klass_value[index]] += tmp
 *     for i in range(nclass):
 */
    __pyx_v_tmp = (__pyx_v_tmp * __pyx_v_tmp);

    /* "EvoDAG/cython_utils.pyx":110
 *             tmp = mean_num[klass_value[index]]
 *         tmp *= tmp
 *         std_num_value[klass_value[index]] += tmp             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         std_num_value[i] = std_num_value[i] / mean_den_value[i]
 */
    __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
    (__pyx_v_std_num_value[__pyx_t_5]) = ((__pyx_v_std_num_value[__pyx_t_5]) + __pyx_v_tmp);
  }

  /* "EvoDAG/cython_utils.pyx":111
 *         tmp *= tmp
 *         std_num_value[klass_value[index]] += tmp
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         std_num_value[i] = std_num_value[i] / mean_den_value[i]
 *     tmp = len(mask)
 */
  __pyx_t_12 = __pyx_v_nclass;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":112
 *         std_num_value[klass_value[index]] += tmp
 *     for i in range(nclass):
 *         std_num_value[i] = std_num_value[i] / mean_den_value[i]             # <<<<<<<<<<<<<<
 *     tmp = len(mask)
 *     for i in range(nclass):
 */
    if (unlikely((__pyx_v_mean_den_value[__pyx_v_i]) == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 112, __pyx_L1_error)
    }
    (__pyx_v_std_num_value[__pyx_v_i]) = ((__pyx_v_std_num_value[__pyx_v_i]) / (__pyx_v_mean_den_value[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":113
 *     for i in range(nclass):
 *         std_num_value[i] = std_num_value[i] / mean_den_value[i]
 *     tmp = len(mask)             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         mean_den_value[i] = mean_den_value[i] / tmp
 */
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 113, __pyx_L1_error)
  }
  __pyx_t_7 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_v_tmp = __pyx_t_7;

  /* "EvoDAG/cython_utils.pyx":114
 *         std_num_value[i] = std_num_value[i] / mean_den_value[i]
 *     tmp = len(mask)
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         mean_den_value[i] = mean_den_value[i] / tmp
 *     return [mean_num, std_num, mean_den]
 */
  __pyx_t_12 = __pyx_v_nclass;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_13; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":115
 *     tmp = len(mask)
 *     for i in range(nclass):
 *         mean_den_value[i] = mean_den_value[i] / tmp             # <<<<<<<<<<<<<<
 *     return [mean_num, std_num, mean_den]
 * 
 */
    if (unlikely(__pyx_v_tmp == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 115, __pyx_L1_error)
    }
    (__pyx_v_mean_den_value[__pyx_v_i]) = ((__pyx_v_mean_den_value[__pyx_v_i]) / __pyx_v_tmp);
  }

  /* "EvoDAG/cython_utils.pyx":116
 *     for i in range(nclass):
 *         mean_den_value[i] = mean_den_value[i] / tmp
 *     return [mean_num, std_num, mean_den]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_14 = PyList_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_mean_num));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_num));
  PyList_SET_ITEM(__pyx_t_14, 0, ((PyObject *)__pyx_v_mean_num));
  __Pyx_INCREF(((PyObject *)__pyx_v_std_num));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_std_num));
  PyList_SET_ITEM(__pyx_t_14, 1, ((PyObject *)__pyx_v_std_num));
  __Pyx_INCREF(((PyObject *)__pyx_v_mean_den));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_den));
  PyList_SET_ITEM(__pyx_t_14, 2, ((PyObject *)__pyx_v_mean_den));
  __pyx_r = ((PyObject*)__pyx_t_14);
  __pyx_t_14 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":70
 * 
 * 
 * cpdef list naive_bayes_mean_std2(SparseArray var, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den, std_num
 *     cdef unsigned int *a = var.index.data.as_uints, index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_mean_std2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_num);
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_den);
  __Pyx_XDECREF((PyObject *)__pyx_v_std_num);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_9naive_bayes_mean_std2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_9naive_bayes_mean_std2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var = 0;
  arrayobject *__pyx_v_klass = 0;
  arrayobject *__pyx_v_mask = 0;
  int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes_mean_std2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_var,&__pyx_n_s_klass,&__pyx_n_s_mask_2,&__pyx_n_s_nclass,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_var)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_klass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_mean_std2", 1, 4, 4, 1); __PYX_ERR(0, 70, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_mean_std2", 1, 4, 4, 2); __PYX_ERR(0, 70, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_mean_std2", 1, 4, 4, 3); __PYX_ERR(0, 70, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes_mean_std2") < 0)) __PYX_ERR(0, 70, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_var = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_klass = ((arrayobject *)values[1]);
    __pyx_v_mask = ((arrayobject *)values[2]);
    __pyx_v_nclass = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nclass == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 70, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes_mean_std2", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 70, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_mean_std2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_var), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "var", 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_klass), __pyx_ptype_7cpython_5array_array, 1, "klass", 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_7cpython_5array_array, 1, "mask", 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_8naive_bayes_mean_std2(__pyx_self, __pyx_v_var, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_8naive_bayes_mean_std2(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_mean_std2", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_naive_bayes_mean_std2(__pyx_v_var, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_mean_std2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":120
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes(list var, list coef, unsigned int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_11naive_bayes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes(PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  PyObject *__pyx_v_l = 0;
  arrayobject *__pyx_v_p_klass = 0;
  arrayobject *__pyx_v_s_klass = 0;
  arrayobject *__pyx_v_m_klass = 0;
  double *__pyx_v_p_klass_value;
  double __pyx_v_a;
  double *__pyx_v_s_klass_value;
  double *__pyx_v_m_klass_value;
  double __pyx_v_c;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_b = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_v = 0;
  CYTHON_UNUSED unsigned int __pyx_v__len;
  unsigned int __pyx_v_nvar;
  PyObject *__pyx_v_res = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  unsigned int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("naive_bayes", 0);

  /* "EvoDAG/cython_utils.pyx":122
 * cpdef list naive_bayes(list var, list coef, unsigned int nclass):
 *     cdef Py_ssize_t i, j
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)             # <<<<<<<<<<<<<<
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 2), s_klass, m_klass
 *     cdef double *p_klass_value = p_klass.data.as_doubles, a, *s_klass_value, *m_klass_value
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_coef, 0);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_l = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":123
 *     cdef Py_ssize_t i, j
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 2), s_klass, m_klass             # <<<<<<<<<<<<<<
 *     cdef double *p_klass_value = p_klass.data.as_doubles, a, *s_klass_value, *m_klass_value
 *     cdef double c
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_l, 2);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_p_klass = ((arrayobject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":124
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 2), s_klass, m_klass
 *     cdef double *p_klass_value = p_klass.data.as_doubles, a, *s_klass_value, *m_klass_value             # <<<<<<<<<<<<<<
 *     cdef double c
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v, tmp
 */
  __pyx_t_3 = __pyx_v_p_klass->data.as_doubles;
  __pyx_v_p_klass_value = __pyx_t_3;

  /* "EvoDAG/cython_utils.pyx":126
 *     cdef double *p_klass_value = p_klass.data.as_doubles, a, *s_klass_value, *m_klass_value
 *     cdef double c
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v, tmp             # <<<<<<<<<<<<<<
 *     cdef unsigned int _len = len(b), nvar = len(var)
 *     cdef list res = []
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_var, 0);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_b = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":127
 *     cdef double c
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v, tmp
 *     cdef unsigned int _len = len(b), nvar = len(var)             # <<<<<<<<<<<<<<
 *     cdef list res = []
 *     for i in range(nclass):
 */
  __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_b)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_v__len = __pyx_t_4;
  if (unlikely(__pyx_v_var == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 127, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_var); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_v_nvar = __pyx_t_4;

  /* "EvoDAG/cython_utils.pyx":128
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v, tmp
 *     cdef unsigned int _len = len(b), nvar = len(var)
 *     cdef list res = []             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         a = math.log(p_klass_value[i])
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_res = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":129
 *     cdef unsigned int _len = len(b), nvar = len(var)
 *     cdef list res = []
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         a = math.log(p_klass_value[i])
 *         b = b.mul2(0)
 */
  __pyx_t_5 = __pyx_v_nclass;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_6; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":130
 *     cdef list res = []
 *     for i in range(nclass):
 *         a = math.log(p_klass_value[i])             # <<<<<<<<<<<<<<
 *         b = b.mul2(0)
 *         c = 0
 */
    __pyx_v_a = log((__pyx_v_p_klass_value[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":131
 *     for i in range(nclass):
 *         a = math.log(p_klass_value[i])
 *         b = b.mul2(0)             # <<<<<<<<<<<<<<
 *         c = 0
 *         for j in range(nvar):
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->mul2(__pyx_v_b, 0.0, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "EvoDAG/cython_utils.pyx":132
 *         a = math.log(p_klass_value[i])
 *         b = b.mul2(0)
 *         c = 0             # <<<<<<<<<<<<<<
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 */
    __pyx_v_c = 0.0;

    /* "EvoDAG/cython_utils.pyx":133
 *         b = b.mul2(0)
 *         c = 0
 *         for j in range(nvar):             # <<<<<<<<<<<<<<
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)
 */
    __pyx_t_7 = __pyx_v_nvar;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "EvoDAG/cython_utils.pyx":134
 *         c = 0
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)             # <<<<<<<<<<<<<<
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_var, __pyx_v_j);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_v, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":135
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)             # <<<<<<<<<<<<<<
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_coef, __pyx_v_j);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_l, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":136
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)             # <<<<<<<<<<<<<<
 *             m_klass_value = m_klass.data.as_doubles
 *             s_klass = <array.array> PyList_GET_ITEM(l, 1)
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_l, 0);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_m_klass, ((arrayobject *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":137
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles             # <<<<<<<<<<<<<<
 *             s_klass = <array.array> PyList_GET_ITEM(l, 1)
 *             s_klass_value = s_klass.data.as_doubles
 */
      __pyx_t_3 = __pyx_v_m_klass->data.as_doubles;
      __pyx_v_m_klass_value = __pyx_t_3;

      /* "EvoDAG/cython_utils.pyx":138
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles
 *             s_klass = <array.array> PyList_GET_ITEM(l, 1)             # <<<<<<<<<<<<<<
 *             s_klass_value = s_klass.data.as_doubles
 *             c += math.log(2 * math.pi * s_klass_value[i])
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_l, 1);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_s_klass, ((arrayobject *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":139
 *             m_klass_value = m_klass.data.as_doubles
 *             s_klass = <array.array> PyList_GET_ITEM(l, 1)
 *             s_klass_value = s_klass.data.as_doubles             # <<<<<<<<<<<<<<
 *             c += math.log(2 * math.pi * s_klass_value[i])
 *             b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))
 */
      __pyx_t_3 = __pyx_v_s_klass->data.as_doubles;
      __pyx_v_s_klass_value = __pyx_t_3;

      /* "EvoDAG/cython_utils.pyx":140
 *             s_klass = <array.array> PyList_GET_ITEM(l, 1)
 *             s_klass_value = s_klass.data.as_doubles
 *             c += math.log(2 * math.pi * s_klass_value[i])             # <<<<<<<<<<<<<<
 *             b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))
 *         b = b.mul2(-0.5)
 */
      __pyx_v_c = (__pyx_v_c + log(((2.0 * M_PI) * (__pyx_v_s_klass_value[__pyx_v_i]))));

      /* "EvoDAG/cython_utils.pyx":141
 *             s_klass_value = s_klass.data.as_doubles
 *             c += math.log(2 * math.pi * s_klass_value[i])
 *             b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))             # <<<<<<<<<<<<<<
 *         b = b.mul2(-0.5)
 *         b = b.add2(a - (0.5 * c))
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_v->__pyx_vtab)->add2(__pyx_v_v, (-(__pyx_v_m_klass_value[__pyx_v_i])), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2)->__pyx_vtab)->sq(((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2), 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10)->__pyx_vtab)->mul2(((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10), (1.0 / (__pyx_v_s_klass_value[__pyx_v_i])), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->add(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2), 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10));
      __pyx_t_10 = 0;
    }

    /* "EvoDAG/cython_utils.pyx":142
 *             c += math.log(2 * math.pi * s_klass_value[i])
 *             b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))
 *         b = b.mul2(-0.5)             # <<<<<<<<<<<<<<
 *         b = b.add2(a - (0.5 * c))
 *         PyList_Append(res, b)
 */
    __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->mul2(__pyx_v_b, -0.5, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "EvoDAG/cython_utils.pyx":143
 *             b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))
 *         b = b.mul2(-0.5)
 *         b = b.add2(a - (0.5 * c))             # <<<<<<<<<<<<<<
 *         PyList_Append(res, b)
 *     return res
 */
    __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->add2(__pyx_v_b, (__pyx_v_a - (0.5 * __pyx_v_c)), 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "EvoDAG/cython_utils.pyx":144
 *         b = b.mul2(-0.5)
 *         b = b.add2(a - (0.5 * c))
 *         PyList_Append(res, b)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_11 = PyList_Append(__pyx_v_res, ((PyObject *)__pyx_v_b)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 144, __pyx_L1_error)
  }

  /* "EvoDAG/cython_utils.pyx":145
 *         b = b.add2(a - (0.5 * c))
 *         PyList_Append(res, b)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":120
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes(list var, list coef, unsigned int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF((PyObject *)__pyx_v_p_klass);
  __Pyx_XDECREF((PyObject *)__pyx_v_s_klass);
  __Pyx_XDECREF((PyObject *)__pyx_v_m_klass);
  __Pyx_XDECREF((PyObject *)__pyx_v_b);
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_11naive_bayes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_11naive_bayes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_var = 0;
  PyObject *__pyx_v_coef = 0;
  unsigned int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_var,&__pyx_n_s_coef,&__pyx_n_s_nclass,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_var)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coef)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes", 1, 3, 3, 1); __PYX_ERR(0, 120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes", 1, 3, 3, 2); __PYX_ERR(0, 120, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes") < 0)) __PYX_ERR(0, 120, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_var = ((PyObject*)values[0]);
    __pyx_v_coef = ((PyObject*)values[1]);
    __pyx_v_nclass = __Pyx_PyInt_As_unsigned_int(values[2]); if (unlikely((__pyx_v_nclass == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 120, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_var), (&PyList_Type), 1, "var", 1))) __PYX_ERR(0, 120, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coef), (&PyList_Type), 1, "coef", 1))) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_10naive_bayes(__pyx_self, __pyx_v_var, __pyx_v_coef, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_10naive_bayes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_naive_bayes(__pyx_v_var, __pyx_v_coef, __pyx_v_nclass, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":149
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes_Nc(SparseArray var, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den
 *     cdef unsigned int *a = var.index.data.as_uints, index
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13naive_bayes_Nc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_Nc(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  arrayobject *__pyx_v_mean_num = 0;
  arrayobject *__pyx_v_mean_den = 0;
  unsigned int *__pyx_v_a;
  unsigned int __pyx_v_index;
  unsigned int *__pyx_v_klass_value;
  double *__pyx_v_a_value;
  double __pyx_v_tmp;
  double *__pyx_v_mean_num_value;
  double *__pyx_v_mean_den_value;
  double __pyx_v_N;
  unsigned int *__pyx_v_mask_value;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_k;
  unsigned int __pyx_v_var_end;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int *__pyx_t_1;
  double *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  unsigned int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("naive_bayes_Nc", 0);

  /* "EvoDAG/cython_utils.pyx":151
 * cpdef list naive_bayes_Nc(SparseArray var, array.array klass, array.array mask, int nclass):
 *     cdef array.array mean_num, mean_den
 *     cdef unsigned int *a = var.index.data.as_uints, index             # <<<<<<<<<<<<<<
 *     cdef unsigned int *klass_value = klass.data.as_uints
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 */
  __pyx_t_1 = __pyx_v_var->index->data.as_uints;
  __pyx_v_a = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":152
 *     cdef array.array mean_num, mean_den
 *     cdef unsigned int *a = var.index.data.as_uints, index
 *     cdef unsigned int *klass_value = klass.data.as_uints             # <<<<<<<<<<<<<<
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 *     cdef double *mean_num_value, *mean_den_value, N=0
 */
  __pyx_t_1 = __pyx_v_klass->data.as_uints;
  __pyx_v_klass_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":153
 *     cdef unsigned int *a = var.index.data.as_uints, index
 *     cdef unsigned int *klass_value = klass.data.as_uints
 *     cdef double *a_value = var.data.data.as_doubles, tmp             # <<<<<<<<<<<<<<
 *     cdef double *mean_num_value, *mean_den_value, N=0
 *     cdef unsigned int *mask_value = mask.data.as_uints
 */
  __pyx_t_2 = __pyx_v_var->data->data.as_doubles;
  __pyx_v_a_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":154
 *     cdef unsigned int *klass_value = klass.data.as_uints
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 *     cdef double *mean_num_value, *mean_den_value, N=0             # <<<<<<<<<<<<<<
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     mean_num = array.clone(var.data, nclass, zero=True)
 */
  __pyx_v_N = 0.0;

  /* "EvoDAG/cython_utils.pyx":155
 *     cdef double *a_value = var.data.data.as_doubles, tmp
 *     cdef double *mean_num_value, *mean_den_value, N=0
 *     cdef unsigned int *mask_value = mask.data.as_uints             # <<<<<<<<<<<<<<
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_mask->data.as_uints;
  __pyx_v_mask_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":156
 *     cdef double *mean_num_value, *mean_den_value, N=0
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     mean_num = array.clone(var.data, nclass, zero=True)             # <<<<<<<<<<<<<<
 *     mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_3), __pyx_v_nclass, 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_num = ((arrayobject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "EvoDAG/cython_utils.pyx":157
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles             # <<<<<<<<<<<<<<
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_mean_num->data.as_doubles;
  __pyx_v_mean_num_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":158
 *     mean_num = array.clone(var.data, nclass, zero=True)
 *     mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)             # <<<<<<<<<<<<<<
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_4), __pyx_v_nclass, 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_mean_den = ((arrayobject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":159
 *     mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 */
  __pyx_t_2 = __pyx_v_mean_den->data.as_doubles;
  __pyx_v_mean_den_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":160
 *     mean_den = array.clone(var.data, nclass, zero=True)
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0             # <<<<<<<<<<<<<<
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:
 */
  __pyx_v_k = 0;

  /* "EvoDAG/cython_utils.pyx":161
 *     mean_den_value = mean_den.data.as_doubles
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero             # <<<<<<<<<<<<<<
 *     if var_end == 0:
 *         return [mean_num, mean_den]
 */
  __pyx_t_5 = __pyx_v_var->non_zero;
  __pyx_v_var_end = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":162
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:             # <<<<<<<<<<<<<<
 *         return [mean_num, mean_den]
 *     for i in range(len(mask)):
 */
  __pyx_t_6 = ((__pyx_v_var_end == 0) != 0);
  if (__pyx_t_6) {

    /* "EvoDAG/cython_utils.pyx":163
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:
 *         return [mean_num, mean_den]             # <<<<<<<<<<<<<<
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_mean_num));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_num));
    PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_mean_num));
    __Pyx_INCREF(((PyObject *)__pyx_v_mean_den));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_den));
    PyList_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_mean_den));
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "EvoDAG/cython_utils.pyx":162
 *     cdef Py_ssize_t i, k=0
 *     cdef unsigned int var_end = var.non_zero
 *     if var_end == 0:             # <<<<<<<<<<<<<<
 *         return [mean_num, mean_den]
 *     for i in range(len(mask)):
 */
  }

  /* "EvoDAG/cython_utils.pyx":164
 *     if var_end == 0:
 *         return [mean_num, mean_den]
 *     for i in range(len(mask)):             # <<<<<<<<<<<<<<
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 */
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 164, __pyx_L1_error)
  }
  __pyx_t_7 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 164, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "EvoDAG/cython_utils.pyx":165
 *         return [mean_num, mean_den]
 *     for i in range(len(mask)):
 *         index = mask_value[i]             # <<<<<<<<<<<<<<
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 */
    __pyx_v_index = (__pyx_v_mask_value[__pyx_v_i]);

    /* "EvoDAG/cython_utils.pyx":166
 *     for i in range(len(mask)):
 *         index = mask_value[i]
 *         for k in range(k, var_end):             # <<<<<<<<<<<<<<
 *             if a[k] >= index:
 *                 break
 */
    __pyx_t_5 = __pyx_v_var_end;
    __pyx_t_10 = __pyx_t_5;
    for (__pyx_t_11 = __pyx_v_k; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_k = __pyx_t_11;

      /* "EvoDAG/cython_utils.pyx":167
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index and a_value[k] > 0:
 */
      __pyx_t_6 = (((__pyx_v_a[__pyx_v_k]) >= __pyx_v_index) != 0);
      if (__pyx_t_6) {

        /* "EvoDAG/cython_utils.pyx":168
 *         for k in range(k, var_end):
 *             if a[k] >= index:
 *                 break             # <<<<<<<<<<<<<<
 *         if a[k] == index and a_value[k] > 0:
 *             mean_num_value[klass_value[index]] += a_value[k]
 */
        goto __pyx_L7_break;

        /* "EvoDAG/cython_utils.pyx":167
 *         index = mask_value[i]
 *         for k in range(k, var_end):
 *             if a[k] >= index:             # <<<<<<<<<<<<<<
 *                 break
 *         if a[k] == index and a_value[k] > 0:
 */
      }
    }
    __pyx_L7_break:;

    /* "EvoDAG/cython_utils.pyx":169
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index and a_value[k] > 0:             # <<<<<<<<<<<<<<
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             N += a_value[k]
 */
    __pyx_t_12 = (((__pyx_v_a[__pyx_v_k]) == __pyx_v_index) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_6 = __pyx_t_12;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_12 = (((__pyx_v_a_value[__pyx_v_k]) > 0.0) != 0);
    __pyx_t_6 = __pyx_t_12;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":170
 *                 break
 *         if a[k] == index and a_value[k] > 0:
 *             mean_num_value[klass_value[index]] += a_value[k]             # <<<<<<<<<<<<<<
 *             N += a_value[k]
 *             mean_den_value[klass_value[index]] += 1
 */
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_num_value[__pyx_t_5]) = ((__pyx_v_mean_num_value[__pyx_t_5]) + (__pyx_v_a_value[__pyx_v_k]));

      /* "EvoDAG/cython_utils.pyx":171
 *         if a[k] == index and a_value[k] > 0:
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             N += a_value[k]             # <<<<<<<<<<<<<<
 *             mean_den_value[klass_value[index]] += 1
 *         else:
 */
      __pyx_v_N = (__pyx_v_N + (__pyx_v_a_value[__pyx_v_k]));

      /* "EvoDAG/cython_utils.pyx":172
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             N += a_value[k]
 *             mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *         else:
 *             mean_den_value[klass_value[index]] += 1
 */
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_den_value[__pyx_t_5]) = ((__pyx_v_mean_den_value[__pyx_t_5]) + 1.0);

      /* "EvoDAG/cython_utils.pyx":169
 *             if a[k] >= index:
 *                 break
 *         if a[k] == index and a_value[k] > 0:             # <<<<<<<<<<<<<<
 *             mean_num_value[klass_value[index]] += a_value[k]
 *             N += a_value[k]
 */
      goto __pyx_L9;
    }

    /* "EvoDAG/cython_utils.pyx":174
 *             mean_den_value[klass_value[index]] += 1
 *         else:
 *             mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         mean_num_value[i] = math.log(mean_num_value[i] / N)
 */
    /*else*/ {
      __pyx_t_5 = (__pyx_v_klass_value[__pyx_v_index]);
      (__pyx_v_mean_den_value[__pyx_t_5]) = ((__pyx_v_mean_den_value[__pyx_t_5]) + 1.0);
    }
    __pyx_L9:;
  }

  /* "EvoDAG/cython_utils.pyx":175
 *         else:
 *             mean_den_value[klass_value[index]] += 1
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         mean_num_value[i] = math.log(mean_num_value[i] / N)
 *     tmp = len(mask)
 */
  __pyx_t_13 = __pyx_v_nclass;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_14; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":176
 *             mean_den_value[klass_value[index]] += 1
 *     for i in range(nclass):
 *         mean_num_value[i] = math.log(mean_num_value[i] / N)             # <<<<<<<<<<<<<<
 *     tmp = len(mask)
 *     for i in range(nclass):
 */
    (__pyx_v_mean_num_value[__pyx_v_i]) = log(((__pyx_v_mean_num_value[__pyx_v_i]) / __pyx_v_N));
  }

  /* "EvoDAG/cython_utils.pyx":177
 *     for i in range(nclass):
 *         mean_num_value[i] = math.log(mean_num_value[i] / N)
 *     tmp = len(mask)             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         mean_den_value[i] = math.log(mean_den_value[i] / tmp)
 */
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 177, __pyx_L1_error)
  }
  __pyx_t_7 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_v_tmp = __pyx_t_7;

  /* "EvoDAG/cython_utils.pyx":178
 *         mean_num_value[i] = math.log(mean_num_value[i] / N)
 *     tmp = len(mask)
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         mean_den_value[i] = math.log(mean_den_value[i] / tmp)
 *     return [mean_num, mean_den]
 */
  __pyx_t_13 = __pyx_v_nclass;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_14; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":179
 *     tmp = len(mask)
 *     for i in range(nclass):
 *         mean_den_value[i] = math.log(mean_den_value[i] / tmp)             # <<<<<<<<<<<<<<
 *     return [mean_num, mean_den]
 * 
 */
    (__pyx_v_mean_den_value[__pyx_v_i]) = log(((__pyx_v_mean_den_value[__pyx_v_i]) / __pyx_v_tmp));
  }

  /* "EvoDAG/cython_utils.pyx":180
 *     for i in range(nclass):
 *         mean_den_value[i] = math.log(mean_den_value[i] / tmp)
 *     return [mean_num, mean_den]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_mean_num));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_num));
  PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_mean_num));
  __Pyx_INCREF(((PyObject *)__pyx_v_mean_den));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mean_den));
  PyList_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_mean_den));
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":149
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes_Nc(SparseArray var, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den
 *     cdef unsigned int *a = var.index.data.as_uints, index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_Nc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_num);
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_den);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13naive_bayes_Nc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13naive_bayes_Nc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var = 0;
  arrayobject *__pyx_v_klass = 0;
  arrayobject *__pyx_v_mask = 0;
  int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes_Nc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_var,&__pyx_n_s_klass,&__pyx_n_s_mask_2,&__pyx_n_s_nclass,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_var)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_klass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_Nc", 1, 4, 4, 1); __PYX_ERR(0, 149, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_Nc", 1, 4, 4, 2); __PYX_ERR(0, 149, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_Nc", 1, 4, 4, 3); __PYX_ERR(0, 149, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes_Nc") < 0)) __PYX_ERR(0, 149, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_var = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_klass = ((arrayobject *)values[1]);
    __pyx_v_mask = ((arrayobject *)values[2]);
    __pyx_v_nclass = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nclass == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes_Nc", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 149, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_Nc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_var), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "var", 0))) __PYX_ERR(0, 149, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_klass), __pyx_ptype_7cpython_5array_array, 1, "klass", 0))) __PYX_ERR(0, 149, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_7cpython_5array_array, 1, "mask", 0))) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_12naive_bayes_Nc(__pyx_self, __pyx_v_var, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_12naive_bayes_Nc(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_Nc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_naive_bayes_Nc(__pyx_v_var, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_Nc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":184
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes_MN(list var, list coef, unsigned int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j, k
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_15naive_bayes_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_naive_bayes_MN(PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_k;
  PyObject *__pyx_v_l = 0;
  arrayobject *__pyx_v_p_klass = 0;
  arrayobject *__pyx_v_m_klass = 0;
  double *__pyx_v_p_klass_value;
  double *__pyx_v_m_klass_value;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_b = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_v = 0;
  unsigned int __pyx_v_nvar;
  PyObject *__pyx_v_res = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  unsigned int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_MN", 0);

  /* "EvoDAG/cython_utils.pyx":186
 * cpdef list naive_bayes_MN(list var, list coef, unsigned int nclass):
 *     cdef Py_ssize_t i, j, k
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)             # <<<<<<<<<<<<<<
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 1), m_klass
 *     cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_coef, 0);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_l = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":187
 *     cdef Py_ssize_t i, j, k
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 1), m_klass             # <<<<<<<<<<<<<<
 *     cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_l, 1);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_p_klass = ((arrayobject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":188
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 1), m_klass
 *     cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value             # <<<<<<<<<<<<<<
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v
 *     cdef unsigned int nvar = len(var)
 */
  __pyx_t_3 = __pyx_v_p_klass->data.as_doubles;
  __pyx_v_p_klass_value = __pyx_t_3;

  /* "EvoDAG/cython_utils.pyx":189
 *     cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 1), m_klass
 *     cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v             # <<<<<<<<<<<<<<
 *     cdef unsigned int nvar = len(var)
 *     cdef list res = []
 */
  __pyx_t_1 = PyList_GET_ITEM(__pyx_v_var, 0);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_b = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":190
 *     cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v
 *     cdef unsigned int nvar = len(var)             # <<<<<<<<<<<<<<
 *     cdef list res = []
 *     for i in range(nclass):
 */
  if (unlikely(__pyx_v_var == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 190, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_var); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_v_nvar = __pyx_t_4;

  /* "EvoDAG/cython_utils.pyx":191
 *     cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v
 *     cdef unsigned int nvar = len(var)
 *     cdef list res = []             # <<<<<<<<<<<<<<
 *     for i in range(nclass):
 *         b = b.mul2(0)
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_res = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":192
 *     cdef unsigned int nvar = len(var)
 *     cdef list res = []
 *     for i in range(nclass):             # <<<<<<<<<<<<<<
 *         b = b.mul2(0)
 *         for j in range(nvar):
 */
  __pyx_t_5 = __pyx_v_nclass;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_6; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":193
 *     cdef list res = []
 *     for i in range(nclass):
 *         b = b.mul2(0)             # <<<<<<<<<<<<<<
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->mul2(__pyx_v_b, 0.0, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "EvoDAG/cython_utils.pyx":194
 *     for i in range(nclass):
 *         b = b.mul2(0)
 *         for j in range(nvar):             # <<<<<<<<<<<<<<
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)
 */
    __pyx_t_7 = __pyx_v_nvar;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "EvoDAG/cython_utils.pyx":195
 *         b = b.mul2(0)
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)             # <<<<<<<<<<<<<<
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_var, __pyx_v_j);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_v, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":196
 *         for j in range(nvar):
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)             # <<<<<<<<<<<<<<
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_coef, __pyx_v_j);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_l, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":197
 *             v = <SparseArray> PyList_GET_ITEM(var, j)
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)             # <<<<<<<<<<<<<<
 *             m_klass_value = m_klass.data.as_doubles
 *             b = b.add(v.mul2(m_klass_value[i]))
 */
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_l, 0);
      __pyx_t_2 = ((PyObject *)__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_m_klass, ((arrayobject *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "EvoDAG/cython_utils.pyx":198
 *             l = <list> PyList_GET_ITEM(coef, j)
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles             # <<<<<<<<<<<<<<
 *             b = b.add(v.mul2(m_klass_value[i]))
 *         b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))
 */
      __pyx_t_3 = __pyx_v_m_klass->data.as_doubles;
      __pyx_v_m_klass_value = __pyx_t_3;

      /* "EvoDAG/cython_utils.pyx":199
 *             m_klass = <array.array> PyList_GET_ITEM(l, 0)
 *             m_klass_value = m_klass.data.as_doubles
 *             b = b.add(v.mul2(m_klass_value[i]))             # <<<<<<<<<<<<<<
 *         b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))
 *         m_klass_value = b.data.data.as_doubles
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_v->__pyx_vtab)->mul2(__pyx_v_v, (__pyx_v_m_klass_value[__pyx_v_i]), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->add(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2), 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10));
      __pyx_t_10 = 0;
    }

    /* "EvoDAG/cython_utils.pyx":200
 *             m_klass_value = m_klass.data.as_doubles
 *             b = b.add(v.mul2(m_klass_value[i]))
 *         b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))             # <<<<<<<<<<<<<<
 *         m_klass_value = b.data.data.as_doubles
 *         for k in range(b.non_zero):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_11SparseArray_12sparse_array_SparseArray), __pyx_n_s_constant); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyFloat_FromDouble((__pyx_v_p_klass_value[__pyx_v_i])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyInt_From_unsigned_int(__pyx_v_b->_len); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = NULL;
    __pyx_t_14 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_14 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_11, ((PyObject *)__pyx_v_b->index), __pyx_t_12};
      __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 3+__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_11, ((PyObject *)__pyx_v_b->index), __pyx_t_12};
      __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 3+__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    {
      __pyx_t_15 = PyTuple_New(3+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_t_11);
      __Pyx_INCREF(((PyObject *)__pyx_v_b->index));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_b->index));
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, ((PyObject *)__pyx_v_b->index));
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_14, __pyx_t_12);
      __pyx_t_11 = 0;
      __pyx_t_12 = 0;
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_15, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_11SparseArray_12sparse_array_SparseArray))))) __PYX_ERR(0, 200, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray *)__pyx_v_b->__pyx_vtab)->add(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_10), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF_SET(__pyx_v_b, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "EvoDAG/cython_utils.pyx":201
 *             b = b.add(v.mul2(m_klass_value[i]))
 *         b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))
 *         m_klass_value = b.data.data.as_doubles             # <<<<<<<<<<<<<<
 *         for k in range(b.non_zero):
 *             m_klass_value[k] = math.exp(m_klass_value[k])
 */
    __pyx_t_3 = __pyx_v_b->data->data.as_doubles;
    __pyx_v_m_klass_value = __pyx_t_3;

    /* "EvoDAG/cython_utils.pyx":202
 *         b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))
 *         m_klass_value = b.data.data.as_doubles
 *         for k in range(b.non_zero):             # <<<<<<<<<<<<<<
 *             m_klass_value[k] = math.exp(m_klass_value[k])
 *         PyList_Append(res, b)
 */
    __pyx_t_7 = __pyx_v_b->non_zero;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_k = __pyx_t_9;

      /* "EvoDAG/cython_utils.pyx":203
 *         m_klass_value = b.data.data.as_doubles
 *         for k in range(b.non_zero):
 *             m_klass_value[k] = math.exp(m_klass_value[k])             # <<<<<<<<<<<<<<
 *         PyList_Append(res, b)
 *     return res
 */
      (__pyx_v_m_klass_value[__pyx_v_k]) = exp((__pyx_v_m_klass_value[__pyx_v_k]));
    }

    /* "EvoDAG/cython_utils.pyx":204
 *         for k in range(b.non_zero):
 *             m_klass_value[k] = math.exp(m_klass_value[k])
 *         PyList_Append(res, b)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_14 = PyList_Append(__pyx_v_res, ((PyObject *)__pyx_v_b)); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 204, __pyx_L1_error)
  }

  /* "EvoDAG/cython_utils.pyx":205
 *             m_klass_value[k] = math.exp(m_klass_value[k])
 *         PyList_Append(res, b)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":184
 * 
 * @cython.cdivision(True)
 * cpdef list naive_bayes_MN(list var, list coef, unsigned int nclass):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i, j, k
 *     cdef list l = <list> PyList_GET_ITEM(coef, 0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_MN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF((PyObject *)__pyx_v_p_klass);
  __Pyx_XDECREF((PyObject *)__pyx_v_m_klass);
  __Pyx_XDECREF((PyObject *)__pyx_v_b);
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_15naive_bayes_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_15naive_bayes_MN(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_var = 0;
  PyObject *__pyx_v_coef = 0;
  unsigned int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("naive_bayes_MN (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_var,&__pyx_n_s_coef,&__pyx_n_s_nclass,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_var)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coef)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_MN", 1, 3, 3, 1); __PYX_ERR(0, 184, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("naive_bayes_MN", 1, 3, 3, 2); __PYX_ERR(0, 184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "naive_bayes_MN") < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_var = ((PyObject*)values[0]);
    __pyx_v_coef = ((PyObject*)values[1]);
    __pyx_v_nclass = __Pyx_PyInt_As_unsigned_int(values[2]); if (unlikely((__pyx_v_nclass == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("naive_bayes_MN", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_MN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_var), (&PyList_Type), 1, "var", 1))) __PYX_ERR(0, 184, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coef), (&PyList_Type), 1, "coef", 1))) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_14naive_bayes_MN(__pyx_self, __pyx_v_var, __pyx_v_coef, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_14naive_bayes_MN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_var, PyObject *__pyx_v_coef, unsigned int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("naive_bayes_MN", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_naive_bayes_MN(__pyx_v_var, __pyx_v_coef, __pyx_v_nclass, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.naive_bayes_MN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":212
 *     cdef public Py_ssize_t pos
 *     cdef public Py_ssize_t size
 *     def __cinit__(self, list lst):             # <<<<<<<<<<<<<<
 *         random.shuffle(lst)
 *         self.data = array.array('I', lst)
 */

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_lst = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lst,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lst)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 212, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_lst = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 212, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lst), (&PyList_Type), 1, "lst", 1))) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers___cinit__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self), __pyx_v_lst);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers___cinit__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_lst) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "EvoDAG/cython_utils.pyx":213
 *     cdef public Py_ssize_t size
 *     def __cinit__(self, list lst):
 *         random.shuffle(lst)             # <<<<<<<<<<<<<<
 *         self.data = array.array('I', lst)
 *         self.size = PyList_GET_SIZE(lst)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_random); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shuffle); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_lst) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_lst);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":214
 *     def __cinit__(self, list lst):
 *         random.shuffle(lst)
 *         self.data = array.array('I', lst)             # <<<<<<<<<<<<<<
 *         self.size = PyList_GET_SIZE(lst)
 *         self.pos = 0
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_I);
  __Pyx_GIVEREF(__pyx_n_s_I);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_I);
  __Pyx_INCREF(__pyx_v_lst);
  __Pyx_GIVEREF(__pyx_v_lst);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_lst);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->data));
  __pyx_v_self->data = ((arrayobject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":215
 *         random.shuffle(lst)
 *         self.data = array.array('I', lst)
 *         self.size = PyList_GET_SIZE(lst)             # <<<<<<<<<<<<<<
 *         self.pos = 0
 * 
 */
  __pyx_v_self->size = PyList_GET_SIZE(__pyx_v_lst);

  /* "EvoDAG/cython_utils.pyx":216
 *         self.data = array.array('I', lst)
 *         self.size = PyList_GET_SIZE(lst)
 *         self.pos = 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get(self, Py_ssize_t k):
 */
  __pyx_v_self->pos = 0;

  /* "EvoDAG/cython_utils.pyx":212
 *     cdef public Py_ssize_t pos
 *     cdef public Py_ssize_t size
 *     def __cinit__(self, list lst):             # <<<<<<<<<<<<<<
 *         random.shuffle(lst)
 *         self.data = array.array('I', lst)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":218
 *         self.pos = 0
 * 
 *     cpdef list get(self, Py_ssize_t k):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t end = self.pos + k, i, pos = self.pos
 *         cdef list res = PyList_New(0)
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3get(PyObject *__pyx_v_self, PyObject *__pyx_arg_k); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, Py_ssize_t __pyx_v_k, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_end;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_res = 0;
  arrayobject *__pyx_v_data = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("get", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3get)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 218, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "EvoDAG/cython_utils.pyx":219
 * 
 *     cpdef list get(self, Py_ssize_t k):
 *         cdef Py_ssize_t end = self.pos + k, i, pos = self.pos             # <<<<<<<<<<<<<<
 *         cdef list res = PyList_New(0)
 *         cdef array.array data = self.data
 */
  __pyx_v_end = (__pyx_v_self->pos + __pyx_v_k);
  __pyx_t_6 = __pyx_v_self->pos;
  __pyx_v_pos = __pyx_t_6;

  /* "EvoDAG/cython_utils.pyx":220
 *     cpdef list get(self, Py_ssize_t k):
 *         cdef Py_ssize_t end = self.pos + k, i, pos = self.pos
 *         cdef list res = PyList_New(0)             # <<<<<<<<<<<<<<
 *         cdef array.array data = self.data
 *         if end > self.size:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_res = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":221
 *         cdef Py_ssize_t end = self.pos + k, i, pos = self.pos
 *         cdef list res = PyList_New(0)
 *         cdef array.array data = self.data             # <<<<<<<<<<<<<<
 *         if end > self.size:
 *             end = self.size
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_data = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":222
 *         cdef list res = PyList_New(0)
 *         cdef array.array data = self.data
 *         if end > self.size:             # <<<<<<<<<<<<<<
 *             end = self.size
 *         self.pos = end
 */
  __pyx_t_7 = ((__pyx_v_end > __pyx_v_self->size) != 0);
  if (__pyx_t_7) {

    /* "EvoDAG/cython_utils.pyx":223
 *         cdef array.array data = self.data
 *         if end > self.size:
 *             end = self.size             # <<<<<<<<<<<<<<
 *         self.pos = end
 *         for i in range(pos, end):
 */
    __pyx_t_6 = __pyx_v_self->size;
    __pyx_v_end = __pyx_t_6;

    /* "EvoDAG/cython_utils.pyx":222
 *         cdef list res = PyList_New(0)
 *         cdef array.array data = self.data
 *         if end > self.size:             # <<<<<<<<<<<<<<
 *             end = self.size
 *         self.pos = end
 */
  }

  /* "EvoDAG/cython_utils.pyx":224
 *         if end > self.size:
 *             end = self.size
 *         self.pos = end             # <<<<<<<<<<<<<<
 *         for i in range(pos, end):
 *             PyList_Append(res, data[i])
 */
  __pyx_v_self->pos = __pyx_v_end;

  /* "EvoDAG/cython_utils.pyx":225
 *             end = self.size
 *         self.pos = end
 *         for i in range(pos, end):             # <<<<<<<<<<<<<<
 *             PyList_Append(res, data[i])
 *         return res
 */
  __pyx_t_6 = __pyx_v_end;
  __pyx_t_8 = __pyx_t_6;
  for (__pyx_t_9 = __pyx_v_pos; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "EvoDAG/cython_utils.pyx":226
 *         self.pos = end
 *         for i in range(pos, end):
 *             PyList_Append(res, data[i])             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt(((PyObject *)__pyx_v_data), __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = PyList_Append(__pyx_v_res, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "EvoDAG/cython_utils.pyx":227
 *         for i in range(pos, end):
 *             PyList_Append(res, data[i])
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cpdef int get_one(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":218
 *         self.pos = 0
 * 
 *     cpdef list get(self, Py_ssize_t k):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t end = self.pos + k, i, pos = self.pos
 *         cdef list res = PyList_New(0)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF((PyObject *)__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3get(PyObject *__pyx_v_self, PyObject *__pyx_arg_k); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3get(PyObject *__pyx_v_self, PyObject *__pyx_arg_k) {
  Py_ssize_t __pyx_v_k;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  assert(__pyx_arg_k); {
    __pyx_v_k = __Pyx_PyIndex_AsSsize_t(__pyx_arg_k); if (unlikely((__pyx_v_k == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_2get(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self), ((Py_ssize_t)__pyx_v_k));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_2get(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, Py_ssize_t __pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get(__pyx_v_self, __pyx_v_k, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":229
 *         return res
 * 
 *     cpdef int get_one(self):             # <<<<<<<<<<<<<<
 *         cdef pos = self.pos
 *         self.pos += 1
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_5get_one(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get_one(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_pos = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("get_one", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_one); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_5get_one)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 229, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "EvoDAG/cython_utils.pyx":230
 * 
 *     cpdef int get_one(self):
 *         cdef pos = self.pos             # <<<<<<<<<<<<<<
 *         self.pos += 1
 *         return self.data[pos]
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pos = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":231
 *     cpdef int get_one(self):
 *         cdef pos = self.pos
 *         self.pos += 1             # <<<<<<<<<<<<<<
 *         return self.data[pos]
 * 
 */
  __pyx_v_self->pos = (__pyx_v_self->pos + 1);

  /* "EvoDAG/cython_utils.pyx":232
 *         cdef pos = self.pos
 *         self.pos += 1
 *         return self.data[pos]             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint empty(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->data), __pyx_v_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":229
 *         return res
 * 
 *     cpdef int get_one(self):             # <<<<<<<<<<<<<<
 *         cdef pos = self.pos
 *         self.pos += 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("EvoDAG.cython_utils.SelectNumbers.get_one", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_5get_one(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_5get_one(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_one (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4get_one(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4get_one(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_one", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get_one(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.get_one", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":234
 *         return self.data[pos]
 * 
 *     cpdef bint empty(self):             # <<<<<<<<<<<<<<
 *         return self.pos == self.size
 * 
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_7empty(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_empty(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("empty", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_7empty)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 234, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "EvoDAG/cython_utils.pyx":235
 * 
 *     cpdef bint empty(self):
 *         return self.pos == self.size             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_self->pos == __pyx_v_self->size);
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":234
 *         return self.data[pos]
 * 
 *     cpdef bint empty(self):             # <<<<<<<<<<<<<<
 *         return self.pos == self.size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("EvoDAG.cython_utils.SelectNumbers.empty", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_7empty(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_7empty(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("empty (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_6empty(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_6empty(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("empty", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_empty(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.empty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":210
 * cdef class SelectNumbers:
 *     cdef array.array data
 *     cdef public Py_ssize_t pos             # <<<<<<<<<<<<<<
 *     cdef public Py_ssize_t size
 *     def __cinit__(self, list lst):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos___get__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.pos.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_3pos_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 210, __pyx_L1_error)
  __pyx_v_self->pos = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.pos.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":211
 *     cdef array.array data
 *     cdef public Py_ssize_t pos
 *     cdef public Py_ssize_t size             # <<<<<<<<<<<<<<
 *     def __cinit__(self, list lst):
 *         random.shuffle(lst)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size___get__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_4size_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
  __pyx_v_self->size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_8__reduce_cython__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_10__setstate_cython__(((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_13SelectNumbers_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.SelectNumbers.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":255
 *     cdef unsigned int nclasses
 * 
 *     def __cinit__(self, unsigned int nclasses):             # <<<<<<<<<<<<<<
 *         cdef  array.array mask = array.array('d', [])
 *         self.precision = array.clone(mask, nclasses, zero=False)
 */

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned int __pyx_v_nclasses;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nclasses,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclasses)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 255, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_nclasses = __Pyx_PyInt_As_unsigned_int(values[0]); if (unlikely((__pyx_v_nclasses == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 255, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score___cinit__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_nclasses);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score___cinit__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, unsigned int __pyx_v_nclasses) {
  arrayobject *__pyx_v_mask = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "EvoDAG/cython_utils.pyx":256
 * 
 *     def __cinit__(self, unsigned int nclasses):
 *         cdef  array.array mask = array.array('d', [])             # <<<<<<<<<<<<<<
 *         self.precision = array.clone(mask, nclasses, zero=False)
 *         self.recall = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_d);
  __Pyx_GIVEREF(__pyx_n_s_d);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_d);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mask = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":257
 *     def __cinit__(self, unsigned int nclasses):
 *         cdef  array.array mask = array.array('d', [])
 *         self.precision = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.recall = array.clone(mask, nclasses, zero=False)
 *         self.precision2 = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision));
  __pyx_v_self->precision = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":258
 *         cdef  array.array mask = array.array('d', [])
 *         self.precision = array.clone(mask, nclasses, zero=False)
 *         self.recall = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.precision2 = array.clone(mask, nclasses, zero=False)
 *         self.recall2 = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall));
  __pyx_v_self->recall = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":259
 *         self.precision = array.clone(mask, nclasses, zero=False)
 *         self.recall = array.clone(mask, nclasses, zero=False)
 *         self.precision2 = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.recall2 = array.clone(mask, nclasses, zero=False)
 *         mask = array.array('I', [])
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2));
  __pyx_v_self->precision2 = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":260
 *         self.recall = array.clone(mask, nclasses, zero=False)
 *         self.precision2 = array.clone(mask, nclasses, zero=False)
 *         self.recall2 = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         mask = array.array('I', [])
 *         self.precision_den = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2));
  __pyx_v_self->recall2 = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":261
 *         self.precision2 = array.clone(mask, nclasses, zero=False)
 *         self.recall2 = array.clone(mask, nclasses, zero=False)
 *         mask = array.array('I', [])             # <<<<<<<<<<<<<<
 *         self.precision_den = array.clone(mask, nclasses, zero=False)
 *         self.recall_den = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_I);
  __Pyx_GIVEREF(__pyx_n_s_I);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_I);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_mask, ((arrayobject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":262
 *         self.recall2 = array.clone(mask, nclasses, zero=False)
 *         mask = array.array('I', [])
 *         self.precision_den = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.recall_den = array.clone(mask, nclasses, zero=False)
 *         self.precision2_den = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision_den));
  __pyx_v_self->precision_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":263
 *         mask = array.array('I', [])
 *         self.precision_den = array.clone(mask, nclasses, zero=False)
 *         self.recall_den = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.precision2_den = array.clone(mask, nclasses, zero=False)
 *         self.recall2_den = array.clone(mask, nclasses, zero=False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall_den));
  __pyx_v_self->recall_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":264
 *         self.precision_den = array.clone(mask, nclasses, zero=False)
 *         self.recall_den = array.clone(mask, nclasses, zero=False)
 *         self.precision2_den = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.recall2_den = array.clone(mask, nclasses, zero=False)
 *         self.nclasses = nclasses
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2_den));
  __pyx_v_self->precision2_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":265
 *         self.recall_den = array.clone(mask, nclasses, zero=False)
 *         self.precision2_den = array.clone(mask, nclasses, zero=False)
 *         self.recall2_den = array.clone(mask, nclasses, zero=False)             # <<<<<<<<<<<<<<
 *         self.nclasses = nclasses
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7cpython_5array_clone(__pyx_v_mask, __pyx_v_nclasses, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2_den));
  __pyx_v_self->recall2_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":266
 *         self.precision2_den = array.clone(mask, nclasses, zero=False)
 *         self.recall2_den = array.clone(mask, nclasses, zero=False)
 *         self.nclasses = nclasses             # <<<<<<<<<<<<<<
 * 
 *     cdef _cleanI(self, array.array d):
 */
  __pyx_v_self->nclasses = __pyx_v_nclasses;

  /* "EvoDAG/cython_utils.pyx":255
 *     cdef unsigned int nclasses
 * 
 *     def __cinit__(self, unsigned int nclasses):             # <<<<<<<<<<<<<<
 *         cdef  array.array mask = array.array('d', [])
 *         self.precision = array.clone(mask, nclasses, zero=False)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":268
 *         self.nclasses = nclasses
 * 
 *     cdef _cleanI(self, array.array d):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         cdef unsigned int *a = d.data.as_uints
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score__cleanI(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, arrayobject *__pyx_v_d) {
  Py_ssize_t __pyx_v_i;
  unsigned int *__pyx_v_a;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int *__pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_cleanI", 0);

  /* "EvoDAG/cython_utils.pyx":269
 * 
 *     cdef _cleanI(self, array.array d):
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         cdef unsigned int *a = d.data.as_uints
 *         for i in range(self.nclasses):
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":270
 *     cdef _cleanI(self, array.array d):
 *         cdef Py_ssize_t i = 0
 *         cdef unsigned int *a = d.data.as_uints             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             a[i] = 0
 */
  __pyx_t_1 = __pyx_v_d->data.as_uints;
  __pyx_v_a = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":271
 *         cdef Py_ssize_t i = 0
 *         cdef unsigned int *a = d.data.as_uints
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             a[i] = 0
 * 
 */
  __pyx_t_2 = __pyx_v_self->nclasses;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":272
 *         cdef unsigned int *a = d.data.as_uints
 *         for i in range(self.nclasses):
 *             a[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _cleand(self, array.array d):
 */
    (__pyx_v_a[__pyx_v_i]) = 0;
  }

  /* "EvoDAG/cython_utils.pyx":268
 *         self.nclasses = nclasses
 * 
 *     cdef _cleanI(self, array.array d):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         cdef unsigned int *a = d.data.as_uints
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":274
 *             a[i] = 0
 * 
 *     cdef _cleand(self, array.array d):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         cdef double *a = d.data.as_doubles
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score__cleand(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, arrayobject *__pyx_v_d) {
  Py_ssize_t __pyx_v_i;
  double *__pyx_v_a;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double *__pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_cleand", 0);

  /* "EvoDAG/cython_utils.pyx":275
 * 
 *     cdef _cleand(self, array.array d):
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         cdef double *a = d.data.as_doubles
 *         for i in range(self.nclasses):
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":276
 *     cdef _cleand(self, array.array d):
 *         cdef Py_ssize_t i = 0
 *         cdef double *a = d.data.as_doubles             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             a[i] = 0
 */
  __pyx_t_1 = __pyx_v_d->data.as_doubles;
  __pyx_v_a = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":277
 *         cdef Py_ssize_t i = 0
 *         cdef double *a = d.data.as_doubles
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             a[i] = 0
 * 
 */
  __pyx_t_2 = __pyx_v_self->nclasses;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "EvoDAG/cython_utils.pyx":278
 *         cdef double *a = d.data.as_doubles
 *         for i in range(self.nclasses):
 *             a[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef clean(self):
 */
    (__pyx_v_a[__pyx_v_i]) = 0.0;
  }

  /* "EvoDAG/cython_utils.pyx":274
 *             a[i] = 0
 * 
 *     cdef _cleand(self, array.array d):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         cdef double *a = d.data.as_doubles
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":280
 *             a[i] = 0
 * 
 *     cdef clean(self):             # <<<<<<<<<<<<<<
 *         self._cleand(self.precision)
 *         self._cleanI(self.precision_den)
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_clean(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("clean", 0);

  /* "EvoDAG/cython_utils.pyx":281
 * 
 *     cdef clean(self):
 *         self._cleand(self.precision)             # <<<<<<<<<<<<<<
 *         self._cleanI(self.precision_den)
 *         self._cleand(self.recall)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->precision);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleand(__pyx_v_self, ((arrayobject *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":282
 *     cdef clean(self):
 *         self._cleand(self.precision)
 *         self._cleanI(self.precision_den)             # <<<<<<<<<<<<<<
 *         self._cleand(self.recall)
 *         self._cleanI(self.recall_den)
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->precision_den);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleanI(__pyx_v_self, ((arrayobject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":283
 *         self._cleand(self.precision)
 *         self._cleanI(self.precision_den)
 *         self._cleand(self.recall)             # <<<<<<<<<<<<<<
 *         self._cleanI(self.recall_den)
 *         self._cleand(self.precision2)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->recall);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleand(__pyx_v_self, ((arrayobject *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":284
 *         self._cleanI(self.precision_den)
 *         self._cleand(self.recall)
 *         self._cleanI(self.recall_den)             # <<<<<<<<<<<<<<
 *         self._cleand(self.precision2)
 *         self._cleanI(self.precision2_den)
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->recall_den);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleanI(__pyx_v_self, ((arrayobject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":285
 *         self._cleand(self.recall)
 *         self._cleanI(self.recall_den)
 *         self._cleand(self.precision2)             # <<<<<<<<<<<<<<
 *         self._cleanI(self.precision2_den)
 *         self._cleand(self.recall2)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->precision2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleand(__pyx_v_self, ((arrayobject *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":286
 *         self._cleanI(self.recall_den)
 *         self._cleand(self.precision2)
 *         self._cleanI(self.precision2_den)             # <<<<<<<<<<<<<<
 *         self._cleand(self.recall2)
 *         self._cleanI(self.recall2_den)
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->precision2_den);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleanI(__pyx_v_self, ((arrayobject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":287
 *         self._cleand(self.precision2)
 *         self._cleanI(self.precision2_den)
 *         self._cleand(self.recall2)             # <<<<<<<<<<<<<<
 *         self._cleanI(self.recall2_den)
 *         self.y_pos = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->recall2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleand(__pyx_v_self, ((arrayobject *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":288
 *         self._cleanI(self.precision2_den)
 *         self._cleand(self.recall2)
 *         self._cleanI(self.recall2_den)             # <<<<<<<<<<<<<<
 *         self.y_pos = 0
 *         self.hy_pos = 0
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->recall2_den);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->_cleanI(__pyx_v_self, ((arrayobject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":289
 *         self._cleand(self.recall2)
 *         self._cleanI(self.recall2_den)
 *         self.y_pos = 0             # <<<<<<<<<<<<<<
 *         self.hy_pos = 0
 *         self._accuracy = 0
 */
  __pyx_v_self->y_pos = 0;

  /* "EvoDAG/cython_utils.pyx":290
 *         self._cleanI(self.recall2_den)
 *         self.y_pos = 0
 *         self.hy_pos = 0             # <<<<<<<<<<<<<<
 *         self._accuracy = 0
 *         self._accuracy2 = 0
 */
  __pyx_v_self->hy_pos = 0;

  /* "EvoDAG/cython_utils.pyx":291
 *         self.y_pos = 0
 *         self.hy_pos = 0
 *         self._accuracy = 0             # <<<<<<<<<<<<<<
 *         self._accuracy2 = 0
 * 
 */
  __pyx_v_self->_accuracy = 0.0;

  /* "EvoDAG/cython_utils.pyx":292
 *         self.hy_pos = 0
 *         self._accuracy = 0
 *         self._accuracy2 = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):
 */
  __pyx_v_self->_accuracy2 = 0.0;

  /* "EvoDAG/cython_utils.pyx":280
 *             a[i] = 0
 * 
 *     cdef clean(self):             # <<<<<<<<<<<<<<
 *         self._cleand(self.precision)
 *         self._cleanI(self.precision_den)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.clean", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":294
 *         self._accuracy2 = 0
 * 
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):             # <<<<<<<<<<<<<<
 *         cdef double res = 0
 *         if self.y_pos >= self.y_size:
 */

static double __pyx_f_6EvoDAG_12cython_utils_5Score_get_klass(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, double *__pyx_v_data, unsigned int *__pyx_v_index, unsigned int __pyx_v_i) {
  double __pyx_v_res;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_klass", 0);

  /* "EvoDAG/cython_utils.pyx":295
 * 
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0             # <<<<<<<<<<<<<<
 *         if self.y_pos >= self.y_size:
 *             return res
 */
  __pyx_v_res = 0.0;

  /* "EvoDAG/cython_utils.pyx":296
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0
 *         if self.y_pos >= self.y_size:             # <<<<<<<<<<<<<<
 *             return res
 *         if index[self.y_pos] == i:
 */
  __pyx_t_1 = ((__pyx_v_self->y_pos >= __pyx_v_self->y_size) != 0);
  if (__pyx_t_1) {

    /* "EvoDAG/cython_utils.pyx":297
 *         cdef double res = 0
 *         if self.y_pos >= self.y_size:
 *             return res             # <<<<<<<<<<<<<<
 *         if index[self.y_pos] == i:
 *             res = data[self.y_pos]
 */
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "EvoDAG/cython_utils.pyx":296
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0
 *         if self.y_pos >= self.y_size:             # <<<<<<<<<<<<<<
 *             return res
 *         if index[self.y_pos] == i:
 */
  }

  /* "EvoDAG/cython_utils.pyx":298
 *         if self.y_pos >= self.y_size:
 *             return res
 *         if index[self.y_pos] == i:             # <<<<<<<<<<<<<<
 *             res = data[self.y_pos]
 *             self.y_pos += 1
 */
  __pyx_t_1 = (((__pyx_v_index[__pyx_v_self->y_pos]) == __pyx_v_i) != 0);
  if (__pyx_t_1) {

    /* "EvoDAG/cython_utils.pyx":299
 *             return res
 *         if index[self.y_pos] == i:
 *             res = data[self.y_pos]             # <<<<<<<<<<<<<<
 *             self.y_pos += 1
 *         return res
 */
    __pyx_v_res = (__pyx_v_data[__pyx_v_self->y_pos]);

    /* "EvoDAG/cython_utils.pyx":300
 *         if index[self.y_pos] == i:
 *             res = data[self.y_pos]
 *             self.y_pos += 1             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
    __pyx_v_self->y_pos = (__pyx_v_self->y_pos + 1);

    /* "EvoDAG/cython_utils.pyx":298
 *         if self.y_pos >= self.y_size:
 *             return res
 *         if index[self.y_pos] == i:             # <<<<<<<<<<<<<<
 *             res = data[self.y_pos]
 *             self.y_pos += 1
 */
  }

  /* "EvoDAG/cython_utils.pyx":301
 *             res = data[self.y_pos]
 *             self.y_pos += 1
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":294
 *         self._accuracy2 = 0
 * 
 *     cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):             # <<<<<<<<<<<<<<
 *         cdef double res = 0
 *         if self.y_pos >= self.y_size:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":303
 *         return res
 * 
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):             # <<<<<<<<<<<<<<
 *         cdef double res = 0
 *         if self.hy_pos >= self.hy_size:
 */

static double __pyx_f_6EvoDAG_12cython_utils_5Score_get_predicted(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, double *__pyx_v_data, unsigned int *__pyx_v_index, unsigned int __pyx_v_i) {
  double __pyx_v_res;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_predicted", 0);

  /* "EvoDAG/cython_utils.pyx":304
 * 
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0             # <<<<<<<<<<<<<<
 *         if self.hy_pos >= self.hy_size:
 *             return res
 */
  __pyx_v_res = 0.0;

  /* "EvoDAG/cython_utils.pyx":305
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0
 *         if self.hy_pos >= self.hy_size:             # <<<<<<<<<<<<<<
 *             return res
 *         if index[self.hy_pos] == i:
 */
  __pyx_t_1 = ((__pyx_v_self->hy_pos >= __pyx_v_self->hy_size) != 0);
  if (__pyx_t_1) {

    /* "EvoDAG/cython_utils.pyx":306
 *         cdef double res = 0
 *         if self.hy_pos >= self.hy_size:
 *             return res             # <<<<<<<<<<<<<<
 *         if index[self.hy_pos] == i:
 *             res = data[self.hy_pos]
 */
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "EvoDAG/cython_utils.pyx":305
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):
 *         cdef double res = 0
 *         if self.hy_pos >= self.hy_size:             # <<<<<<<<<<<<<<
 *             return res
 *         if index[self.hy_pos] == i:
 */
  }

  /* "EvoDAG/cython_utils.pyx":307
 *         if self.hy_pos >= self.hy_size:
 *             return res
 *         if index[self.hy_pos] == i:             # <<<<<<<<<<<<<<
 *             res = data[self.hy_pos]
 *             self.hy_pos += 1
 */
  __pyx_t_1 = (((__pyx_v_index[__pyx_v_self->hy_pos]) == __pyx_v_i) != 0);
  if (__pyx_t_1) {

    /* "EvoDAG/cython_utils.pyx":308
 *             return res
 *         if index[self.hy_pos] == i:
 *             res = data[self.hy_pos]             # <<<<<<<<<<<<<<
 *             self.hy_pos += 1
 *         return res
 */
    __pyx_v_res = (__pyx_v_data[__pyx_v_self->hy_pos]);

    /* "EvoDAG/cython_utils.pyx":309
 *         if index[self.hy_pos] == i:
 *             res = data[self.hy_pos]
 *             self.hy_pos += 1             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
    __pyx_v_self->hy_pos = (__pyx_v_self->hy_pos + 1);

    /* "EvoDAG/cython_utils.pyx":307
 *         if self.hy_pos >= self.hy_size:
 *             return res
 *         if index[self.hy_pos] == i:             # <<<<<<<<<<<<<<
 *             res = data[self.hy_pos]
 *             self.hy_pos += 1
 */
  }

  /* "EvoDAG/cython_utils.pyx":310
 *             res = data[self.hy_pos]
 *             self.hy_pos += 1
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def g_recall_precision(self, Py_ssize_t i, SparseArray y,
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":303
 *         return res
 * 
 *     cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):             # <<<<<<<<<<<<<<
 *         cdef double res = 0
 *         if self.hy_pos >= self.hy_size:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":312
 *         return res
 * 
 *     def g_recall_precision(self, Py_ssize_t i, SparseArray y,             # <<<<<<<<<<<<<<
 *                            SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_3g_recall_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_3g_recall_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("g_recall_precision (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_recall_precision", 1, 4, 4, 1); __PYX_ERR(0, 312, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_recall_precision", 1, 4, 4, 2); __PYX_ERR(0, 312, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_recall_precision", 1, 4, 4, 3); __PYX_ERR(0, 312, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "g_recall_precision") < 0)) __PYX_ERR(0, 312, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_i = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_i == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 312, __pyx_L3_error)
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[2]);
    __pyx_v_index = ((arrayobject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("g_recall_precision", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 312, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_recall_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 312, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 313, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_2g_recall_precision(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_i, __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_2g_recall_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, Py_ssize_t __pyx_v_i, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("g_recall_precision", 0);

  /* "EvoDAG/cython_utils.pyx":314
 *     def g_recall_precision(self, Py_ssize_t i, SparseArray y,
 *                            SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":315
 *                            SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":316
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":317
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":318
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":319
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0
 *         f1 = precision[i] * recall[i]
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":320
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0             # <<<<<<<<<<<<<<
 *         f1 = precision[i] * recall[i]
 *         f12 = precision2[i] * recall2[i]
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":321
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0
 *         f1 = precision[i] * recall[i]             # <<<<<<<<<<<<<<
 *         f12 = precision2[i] * recall2[i]
 *         return f1, f12
 */
  __pyx_v_f1 = ((__pyx_v_precision[__pyx_v_i]) * (__pyx_v_recall[__pyx_v_i]));

  /* "EvoDAG/cython_utils.pyx":322
 *         cdef double f1 = 0, f12 = 0
 *         f1 = precision[i] * recall[i]
 *         f12 = precision2[i] * recall2[i]             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
  __pyx_v_f12 = ((__pyx_v_precision2[__pyx_v_i]) * (__pyx_v_recall2[__pyx_v_i]));

  /* "EvoDAG/cython_utils.pyx":323
 *         f1 = precision[i] * recall[i]
 *         f12 = precision2[i] * recall2[i]
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":312
 *         return res
 * 
 *     def g_recall_precision(self, Py_ssize_t i, SparseArray y,             # <<<<<<<<<<<<<<
 *                            SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_recall_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":326
 * 
 *     @cython.cdivision(True)
 *     def F1(self, Py_ssize_t i, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_5F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_5F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("F1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("F1", 1, 4, 4, 1); __PYX_ERR(0, 326, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("F1", 1, 4, 4, 2); __PYX_ERR(0, 326, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("F1", 1, 4, 4, 3); __PYX_ERR(0, 326, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "F1") < 0)) __PYX_ERR(0, 326, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_i = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_i == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L3_error)
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[2]);
    __pyx_v_index = ((arrayobject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("F1", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 326, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_4F1(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_i, __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_4F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, Py_ssize_t __pyx_v_i, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  double __pyx_v_den;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("F1", 0);

  /* "EvoDAG/cython_utils.pyx":327
 *     @cython.cdivision(True)
 *     def F1(self, Py_ssize_t i, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":328
 *     def F1(self, Py_ssize_t i, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":329
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":330
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":331
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":332
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         den = precision[i] + recall[i]
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":333
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         den = precision[i] + recall[i]
 *         if den > 0:
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":334
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         den = precision[i] + recall[i]             # <<<<<<<<<<<<<<
 *         if den > 0:
 *             f1 = (2 * precision[i] * recall[i]) / den
 */
  __pyx_v_den = ((__pyx_v_precision[__pyx_v_i]) + (__pyx_v_recall[__pyx_v_i]));

  /* "EvoDAG/cython_utils.pyx":335
 *         cdef double f1 = 0, f12 = 0, den
 *         den = precision[i] + recall[i]
 *         if den > 0:             # <<<<<<<<<<<<<<
 *             f1 = (2 * precision[i] * recall[i]) / den
 *         den = precision2[i] + recall2[i]
 */
  __pyx_t_3 = ((__pyx_v_den > 0.0) != 0);
  if (__pyx_t_3) {

    /* "EvoDAG/cython_utils.pyx":336
 *         den = precision[i] + recall[i]
 *         if den > 0:
 *             f1 = (2 * precision[i] * recall[i]) / den             # <<<<<<<<<<<<<<
 *         den = precision2[i] + recall2[i]
 *         if den > 0:
 */
    __pyx_v_f1 = (((2.0 * (__pyx_v_precision[__pyx_v_i])) * (__pyx_v_recall[__pyx_v_i])) / __pyx_v_den);

    /* "EvoDAG/cython_utils.pyx":335
 *         cdef double f1 = 0, f12 = 0, den
 *         den = precision[i] + recall[i]
 *         if den > 0:             # <<<<<<<<<<<<<<
 *             f1 = (2 * precision[i] * recall[i]) / den
 *         den = precision2[i] + recall2[i]
 */
  }

  /* "EvoDAG/cython_utils.pyx":337
 *         if den > 0:
 *             f1 = (2 * precision[i] * recall[i]) / den
 *         den = precision2[i] + recall2[i]             # <<<<<<<<<<<<<<
 *         if den > 0:
 *             f12 = (2 * precision2[i] * recall2[i]) / den
 */
  __pyx_v_den = ((__pyx_v_precision2[__pyx_v_i]) + (__pyx_v_recall2[__pyx_v_i]));

  /* "EvoDAG/cython_utils.pyx":338
 *             f1 = (2 * precision[i] * recall[i]) / den
 *         den = precision2[i] + recall2[i]
 *         if den > 0:             # <<<<<<<<<<<<<<
 *             f12 = (2 * precision2[i] * recall2[i]) / den
 *         return f1, f12
 */
  __pyx_t_3 = ((__pyx_v_den > 0.0) != 0);
  if (__pyx_t_3) {

    /* "EvoDAG/cython_utils.pyx":339
 *         den = precision2[i] + recall2[i]
 *         if den > 0:
 *             f12 = (2 * precision2[i] * recall2[i]) / den             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
    __pyx_v_f12 = (((2.0 * (__pyx_v_precision2[__pyx_v_i])) * (__pyx_v_recall2[__pyx_v_i])) / __pyx_v_den);

    /* "EvoDAG/cython_utils.pyx":338
 *             f1 = (2 * precision[i] * recall[i]) / den
 *         den = precision2[i] + recall2[i]
 *         if den > 0:             # <<<<<<<<<<<<<<
 *             f12 = (2 * precision2[i] * recall2[i]) / den
 *         return f1, f12
 */
  }

  /* "EvoDAG/cython_utils.pyx":340
 *         if den > 0:
 *             f12 = (2 * precision2[i] * recall2[i]) / den
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":326
 * 
 *     @cython.cdivision(True)
 *     def F1(self, Py_ssize_t i, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":343
 * 
 *     @cython.cdivision(True)
 *     def macroRecallF1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_7macroRecallF1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_7macroRecallF1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("macroRecallF1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("macroRecallF1", 1, 3, 3, 1); __PYX_ERR(0, 343, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("macroRecallF1", 1, 3, 3, 2); __PYX_ERR(0, 343, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "macroRecallF1") < 0)) __PYX_ERR(0, 343, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("macroRecallF1", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 343, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.macroRecallF1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 343, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 343, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 343, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_6macroRecallF1(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_6macroRecallF1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  double __pyx_v_den;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("macroRecallF1", 0);

  /* "EvoDAG/cython_utils.pyx":344
 *     @cython.cdivision(True)
 *     def macroRecallF1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":345
 *     def macroRecallF1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":346
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":347
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":348
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":349
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":350
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":351
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 += recall[i]
 *             den = precision2[i] + recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":352
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 += recall[i]             # <<<<<<<<<<<<<<
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 */
    __pyx_v_f1 = (__pyx_v_f1 + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":353
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 *             den = precision2[i] + recall2[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision2[__pyx_v_i]) + (__pyx_v_recall2[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":354
 *             f1 += recall[i]
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":355
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den             # <<<<<<<<<<<<<<
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 */
      __pyx_v_f12 = (__pyx_v_f12 + (((2.0 * (__pyx_v_precision2[__pyx_v_i])) * (__pyx_v_recall2[__pyx_v_i])) / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":354
 *             f1 += recall[i]
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    }
  }

  /* "EvoDAG/cython_utils.pyx":356
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses             # <<<<<<<<<<<<<<
 *         f12 = f12 / self.nclasses
 *         return f1, f12
 */
  __pyx_v_f1 = (__pyx_v_f1 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":357
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
  __pyx_v_f12 = (__pyx_v_f12 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":358
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":343
 * 
 *     @cython.cdivision(True)
 *     def macroRecallF1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.macroRecallF1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":361
 * 
 *     @cython.cdivision(True)
 *     def a_recall(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9a_recall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9a_recall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("a_recall (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_recall", 1, 3, 3, 1); __PYX_ERR(0, 361, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_recall", 1, 3, 3, 2); __PYX_ERR(0, 361, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "a_recall") < 0)) __PYX_ERR(0, 361, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("a_recall", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 361, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_recall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 361, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 361, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 361, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_8a_recall(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_8a_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_recall;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("a_recall", 0);

  /* "EvoDAG/cython_utils.pyx":362
 *     @cython.cdivision(True)
 *     def a_recall(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":363
 *     def a_recall(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":364
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":365
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":366
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":367
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":368
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 += recall[i]
 *             f12 += recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":369
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 += recall[i]             # <<<<<<<<<<<<<<
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses
 */
    __pyx_v_f1 = (__pyx_v_f1 + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":370
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 *             f12 += recall2[i]             # <<<<<<<<<<<<<<
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 */
    __pyx_v_f12 = (__pyx_v_f12 + (__pyx_v_recall2[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":371
 *             f1 += recall[i]
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses             # <<<<<<<<<<<<<<
 *         f12 = f12 / self.nclasses
 *         return f1, f12
 */
  __pyx_v_f1 = (__pyx_v_f1 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":372
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
  __pyx_v_f12 = (__pyx_v_f12 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":373
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":361
 * 
 *     @cython.cdivision(True)
 *     def a_recall(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_recall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":376
 * 
 *     @cython.cdivision(True)
 *     def a_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_11a_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_11a_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("a_precision (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_precision", 1, 3, 3, 1); __PYX_ERR(0, 376, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_precision", 1, 3, 3, 2); __PYX_ERR(0, 376, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "a_precision") < 0)) __PYX_ERR(0, 376, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("a_precision", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 376, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 376, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 376, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 376, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10a_precision(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10a_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_recall;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("a_precision", 0);

  /* "EvoDAG/cython_utils.pyx":377
 *     @cython.cdivision(True)
 *     def a_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":378
 *     def a_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":379
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":380
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":381
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":382
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":383
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 += recall[i]
 *             f12 += recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":384
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 += recall[i]             # <<<<<<<<<<<<<<
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses
 */
    __pyx_v_f1 = (__pyx_v_f1 + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":385
 *         for i in range(self.nclasses):
 *             f1 += recall[i]
 *             f12 += recall2[i]             # <<<<<<<<<<<<<<
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 */
    __pyx_v_f12 = (__pyx_v_f12 + (__pyx_v_recall2[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":386
 *             f1 += recall[i]
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses             # <<<<<<<<<<<<<<
 *         f12 = f12 / self.nclasses
 *         return f1, f12
 */
  __pyx_v_f1 = (__pyx_v_f1 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":387
 *             f12 += recall2[i]
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
  __pyx_v_f12 = (__pyx_v_f12 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":388
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     def g_precision(self, SparseArray y, SparseArray hy, array.array index):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":376
 * 
 *     @cython.cdivision(True)
 *     def a_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":390
 *         return f1, f12
 * 
 *     def g_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_13g_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_13g_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("g_precision (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_precision", 1, 3, 3, 1); __PYX_ERR(0, 390, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_precision", 1, 3, 3, 2); __PYX_ERR(0, 390, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "g_precision") < 0)) __PYX_ERR(0, 390, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("g_precision", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 390, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 390, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 390, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 390, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_12g_precision(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_12g_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_recall;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("g_precision", 0);

  /* "EvoDAG/cython_utils.pyx":391
 * 
 *     def g_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":392
 *     def g_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":393
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":394
 *         self.precision_recall()
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":395
 *         cdef double *recall = self.precision.data.as_doubles
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 1.0;
  __pyx_v_f12 = 1.0;

  /* "EvoDAG/cython_utils.pyx":396
 *         cdef double *recall2 = self.precision2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":397
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 *= recall[i]
 *             f12 *= recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":398
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]             # <<<<<<<<<<<<<<
 *             f12 *= recall2[i]
 *         return f1, f12
 */
    __pyx_v_f1 = (__pyx_v_f1 * (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":399
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]
 *             f12 *= recall2[i]             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
    __pyx_v_f12 = (__pyx_v_f12 * (__pyx_v_recall2[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":400
 *             f1 *= recall[i]
 *             f12 *= recall2[i]
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     def g_recall(self, SparseArray y, SparseArray hy, array.array index):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":390
 *         return f1, f12
 * 
 *     def g_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":402
 *         return f1, f12
 * 
 *     def g_recall(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_15g_recall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_15g_recall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("g_recall (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_recall", 1, 3, 3, 1); __PYX_ERR(0, 402, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_recall", 1, 3, 3, 2); __PYX_ERR(0, 402, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "g_recall") < 0)) __PYX_ERR(0, 402, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("g_recall", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 402, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_recall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 402, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 402, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 402, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_14g_recall(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_14g_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_recall;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("g_recall", 0);

  /* "EvoDAG/cython_utils.pyx":403
 * 
 *     def g_recall(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":404
 *     def g_recall(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":405
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":406
 *         self.precision_recall()
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":407
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 1.0;
  __pyx_v_f12 = 1.0;

  /* "EvoDAG/cython_utils.pyx":408
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":409
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 *= recall[i]
 *             f12 *= recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":410
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]             # <<<<<<<<<<<<<<
 *             f12 *= recall2[i]
 *         return f1, f12
 */
    __pyx_v_f1 = (__pyx_v_f1 * (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":411
 *         for i in range(self.nclasses):
 *             f1 *= recall[i]
 *             f12 *= recall2[i]             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
    __pyx_v_f12 = (__pyx_v_f12 * (__pyx_v_recall2[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":412
 *             f1 *= recall[i]
 *             f12 *= recall2[i]
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":402
 *         return f1, f12
 * 
 *     def g_recall(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_recall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":415
 * 
 *     @cython.cdivision(True)
 *     cdef recall2accuracy(self):             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_recall2accuracy(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  double *__pyx_v_recall;
  double *__pyx_v_recall2;
  unsigned int *__pyx_v_recall_den;
  unsigned int *__pyx_v_recall2_den;
  unsigned int __pyx_v_den;
  unsigned int __pyx_v_den2;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double *__pyx_t_1;
  unsigned int *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("recall2accuracy", 0);

  /* "EvoDAG/cython_utils.pyx":416
 *     @cython.cdivision(True)
 *     cdef recall2accuracy(self):
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 */
  __pyx_t_1 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":417
 *     cdef recall2accuracy(self):
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 */
  __pyx_t_1 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":418
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef unsigned int den = 0, den2 = 0
 */
  __pyx_t_2 = __pyx_v_self->recall_den->data.as_uints;
  __pyx_v_recall_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":419
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int den = 0, den2 = 0
 *         cdef Py_ssize_t i
 */
  __pyx_t_2 = __pyx_v_self->recall2_den->data.as_uints;
  __pyx_v_recall2_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":420
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef unsigned int den = 0, den2 = 0             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):
 */
  __pyx_v_den = 0;
  __pyx_v_den2 = 0;

  /* "EvoDAG/cython_utils.pyx":422
 *         cdef unsigned int den = 0, den2 = 0
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             self._accuracy += recall[i]
 *             self._accuracy2 += recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":423
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):
 *             self._accuracy += recall[i]             # <<<<<<<<<<<<<<
 *             self._accuracy2 += recall2[i]
 *             den += recall_den[i]
 */
    __pyx_v_self->_accuracy = (__pyx_v_self->_accuracy + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":424
 *         for i in range(self.nclasses):
 *             self._accuracy += recall[i]
 *             self._accuracy2 += recall2[i]             # <<<<<<<<<<<<<<
 *             den += recall_den[i]
 *             den2 += recall2_den[i]
 */
    __pyx_v_self->_accuracy2 = (__pyx_v_self->_accuracy2 + (__pyx_v_recall2[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":425
 *             self._accuracy += recall[i]
 *             self._accuracy2 += recall2[i]
 *             den += recall_den[i]             # <<<<<<<<<<<<<<
 *             den2 += recall2_den[i]
 *         self._accuracy = self._accuracy / den
 */
    __pyx_v_den = (__pyx_v_den + (__pyx_v_recall_den[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":426
 *             self._accuracy2 += recall2[i]
 *             den += recall_den[i]
 *             den2 += recall2_den[i]             # <<<<<<<<<<<<<<
 *         self._accuracy = self._accuracy / den
 *         self._accuracy2 = self._accuracy2 / den2
 */
    __pyx_v_den2 = (__pyx_v_den2 + (__pyx_v_recall2_den[__pyx_v_i]));
  }

  /* "EvoDAG/cython_utils.pyx":427
 *             den += recall_den[i]
 *             den2 += recall2_den[i]
 *         self._accuracy = self._accuracy / den             # <<<<<<<<<<<<<<
 *         self._accuracy2 = self._accuracy2 / den2
 * 
 */
  __pyx_v_self->_accuracy = (__pyx_v_self->_accuracy / __pyx_v_den);

  /* "EvoDAG/cython_utils.pyx":428
 *             den2 += recall2_den[i]
 *         self._accuracy = self._accuracy / den
 *         self._accuracy2 = self._accuracy2 / den2             # <<<<<<<<<<<<<<
 * 
 *     def accuracy(self, SparseArray y, SparseArray hy, array.array index):
 */
  __pyx_v_self->_accuracy2 = (__pyx_v_self->_accuracy2 / __pyx_v_den2);

  /* "EvoDAG/cython_utils.pyx":415
 * 
 *     @cython.cdivision(True)
 *     cdef recall2accuracy(self):             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":430
 *         self._accuracy2 = self._accuracy2 / den2
 * 
 *     def accuracy(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.recall2accuracy()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_17accuracy(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_17accuracy(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accuracy (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accuracy", 1, 3, 3, 1); __PYX_ERR(0, 430, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accuracy", 1, 3, 3, 2); __PYX_ERR(0, 430, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accuracy") < 0)) __PYX_ERR(0, 430, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accuracy", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 430, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.accuracy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 430, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 430, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 430, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_16accuracy(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_16accuracy(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("accuracy", 0);

  /* "EvoDAG/cython_utils.pyx":431
 * 
 *     def accuracy(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.recall2accuracy()
 *         return self._accuracy, self._accuracy2
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":432
 *     def accuracy(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.recall2accuracy()             # <<<<<<<<<<<<<<
 *         return self._accuracy, self._accuracy2
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->recall2accuracy(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":433
 *         self.count(y, hy, index)
 *         self.recall2accuracy()
 *         return self._accuracy, self._accuracy2             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_accuracy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_accuracy2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":430
 *         self._accuracy2 = self._accuracy2 / den2
 * 
 *     def accuracy(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.recall2accuracy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.accuracy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":436
 * 
 *     @cython.cdivision(True)
 *     def accDotMacroF1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_19accDotMacroF1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_19accDotMacroF1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accDotMacroF1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accDotMacroF1", 1, 3, 3, 1); __PYX_ERR(0, 436, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accDotMacroF1", 1, 3, 3, 2); __PYX_ERR(0, 436, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accDotMacroF1") < 0)) __PYX_ERR(0, 436, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accDotMacroF1", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 436, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.accDotMacroF1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 436, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 436, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 436, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_18accDotMacroF1(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_18accDotMacroF1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  double __pyx_v_den;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("accDotMacroF1", 0);

  /* "EvoDAG/cython_utils.pyx":437
 *     @cython.cdivision(True)
 *     def accDotMacroF1(self, SparseArray y, SparseArray hy, array.array index):
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":438
 *     def accDotMacroF1(self, SparseArray y, SparseArray hy, array.array index):
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":439
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_1 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":440
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_1 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":441
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         self.count(y, hy, index)
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":442
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.recall2accuracy()
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":443
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.recall2accuracy()
 *         self.precision_recall()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":444
 *         cdef Py_ssize_t i = 0
 *         self.count(y, hy, index)
 *         self.recall2accuracy()             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         for i in range(self.nclasses):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->recall2accuracy(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":445
 *         self.count(y, hy, index)
 *         self.recall2accuracy()
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EvoDAG/cython_utils.pyx":446
 *         self.recall2accuracy()
 *         self.precision_recall()
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             den = precision[i] + recall[i]
 *             if den > 0:
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":447
 *         self.precision_recall()
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision[__pyx_v_i]) + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":448
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":449
 *             den = precision[i] + recall[i]
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den             # <<<<<<<<<<<<<<
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 */
      __pyx_v_f1 = (__pyx_v_f1 + (((2.0 * (__pyx_v_precision[__pyx_v_i])) * (__pyx_v_recall[__pyx_v_i])) / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":448
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 */
    }

    /* "EvoDAG/cython_utils.pyx":450
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision2[__pyx_v_i]) + (__pyx_v_recall2[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":451
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":452
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den             # <<<<<<<<<<<<<<
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 */
      __pyx_v_f12 = (__pyx_v_f12 + (((2.0 * (__pyx_v_precision2[__pyx_v_i])) * (__pyx_v_recall2[__pyx_v_i])) / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":451
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    }
  }

  /* "EvoDAG/cython_utils.pyx":453
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses             # <<<<<<<<<<<<<<
 *         f12 = f12 / self.nclasses
 *         return f1 * self._accuracy, f12 * self._accuracy2
 */
  __pyx_v_f1 = (__pyx_v_f1 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":454
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses             # <<<<<<<<<<<<<<
 *         return f1 * self._accuracy, f12 * self._accuracy2
 * 
 */
  __pyx_v_f12 = (__pyx_v_f12 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":455
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 *         return f1 * self._accuracy, f12 * self._accuracy2             # <<<<<<<<<<<<<<
 * 
 *     def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_f1 * __pyx_v_self->_accuracy)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyFloat_FromDouble((__pyx_v_f12 * __pyx_v_self->_accuracy2)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":436
 * 
 *     @cython.cdivision(True)
 *     def accDotMacroF1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.accDotMacroF1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":457
 *         return f1 * self._accuracy, f12 * self._accuracy2
 * 
 *     def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_21g_g_recall_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_21g_g_recall_precision(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("g_g_recall_precision (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_g_recall_precision", 1, 3, 3, 1); __PYX_ERR(0, 457, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_g_recall_precision", 1, 3, 3, 2); __PYX_ERR(0, 457, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "g_g_recall_precision") < 0)) __PYX_ERR(0, 457, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("g_g_recall_precision", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 457, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_g_recall_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 457, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 457, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_20g_g_recall_precision(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_20g_g_recall_precision(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("g_g_recall_precision", 0);

  /* "EvoDAG/cython_utils.pyx":458
 * 
 *     def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":459
 *     def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":460
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":461
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":462
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":463
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":464
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 1.0;
  __pyx_v_f12 = 1.0;

  /* "EvoDAG/cython_utils.pyx":465
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             f1 *= precision[i] * recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":466
 *         cdef double f1 = 1, f12 = 1
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             f1 *= precision[i] * recall[i]
 *             f12 *= precision2[i] * recall2[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":467
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             f1 *= precision[i] * recall[i]             # <<<<<<<<<<<<<<
 *             f12 *= precision2[i] * recall2[i]
 *         return f1, f12
 */
    __pyx_v_f1 = (__pyx_v_f1 * ((__pyx_v_precision[__pyx_v_i]) * (__pyx_v_recall[__pyx_v_i])));

    /* "EvoDAG/cython_utils.pyx":468
 *         for i in range(self.nclasses):
 *             f1 *= precision[i] * recall[i]
 *             f12 *= precision2[i] * recall2[i]             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
    __pyx_v_f12 = (__pyx_v_f12 * ((__pyx_v_precision2[__pyx_v_i]) * (__pyx_v_recall2[__pyx_v_i])));
  }

  /* "EvoDAG/cython_utils.pyx":469
 *             f1 *= precision[i] * recall[i]
 *             f12 *= precision2[i] * recall2[i]
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     def g_F1(self, SparseArray y, SparseArray hy, array.array index):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":457
 *         return f1 * self._accuracy, f12 * self._accuracy2
 * 
 *     def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_g_recall_precision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":471
 *         return f1, f12
 * 
 *     def g_F1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_23g_F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_23g_F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("g_F1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_F1", 1, 3, 3, 1); __PYX_ERR(0, 471, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("g_F1", 1, 3, 3, 2); __PYX_ERR(0, 471, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "g_F1") < 0)) __PYX_ERR(0, 471, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("g_F1", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 471, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 471, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 471, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 471, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_22g_F1(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_22g_F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  double __pyx_v_den;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  double __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("g_F1", 0);

  /* "EvoDAG/cython_utils.pyx":472
 * 
 *     def g_F1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":473
 *     def g_F1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":474
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":475
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":476
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1, den
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":477
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 1, f12 = 1, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":478
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 1.0;
  __pyx_v_f12 = 1.0;

  /* "EvoDAG/cython_utils.pyx":479
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 1, f12 = 1, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":480
 *         cdef double f1 = 1, f12 = 1, den
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             den = precision[i] + recall[i]
 *             if den > 0:
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":481
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f1 *= (2 * precision[i] * recall[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision[__pyx_v_i]) + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":482
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 *= (2 * precision[i] * recall[i]) / den
 *             else:
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":483
 *             den = precision[i] + recall[i]
 *             if den > 0:
 *                 f1 *= (2 * precision[i] * recall[i]) / den             # <<<<<<<<<<<<<<
 *             else:
 *                 f1 = 0
 */
      __pyx_t_7 = ((2.0 * (__pyx_v_precision[__pyx_v_i])) * (__pyx_v_recall[__pyx_v_i]));
      if (unlikely(__pyx_v_den == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 483, __pyx_L1_error)
      }
      __pyx_v_f1 = (__pyx_v_f1 * (__pyx_t_7 / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":482
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 *= (2 * precision[i] * recall[i]) / den
 *             else:
 */
      goto __pyx_L5;
    }

    /* "EvoDAG/cython_utils.pyx":485
 *                 f1 *= (2 * precision[i] * recall[i]) / den
 *             else:
 *                 f1 = 0             # <<<<<<<<<<<<<<
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 */
    /*else*/ {
      __pyx_v_f1 = 0.0;
    }
    __pyx_L5:;

    /* "EvoDAG/cython_utils.pyx":486
 *             else:
 *                 f1 = 0
 *             den = precision2[i] + recall2[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f12 *= (2 * precision2[i] * recall2[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision2[__pyx_v_i]) + (__pyx_v_recall2[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":487
 *                 f1 = 0
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 *= (2 * precision2[i] * recall2[i]) / den
 *             else:
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":488
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 *                 f12 *= (2 * precision2[i] * recall2[i]) / den             # <<<<<<<<<<<<<<
 *             else:
 *                 f12 = 0
 */
      __pyx_t_7 = ((2.0 * (__pyx_v_precision2[__pyx_v_i])) * (__pyx_v_recall2[__pyx_v_i]));
      if (unlikely(__pyx_v_den == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 488, __pyx_L1_error)
      }
      __pyx_v_f12 = (__pyx_v_f12 * (__pyx_t_7 / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":487
 *                 f1 = 0
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 *= (2 * precision2[i] * recall2[i]) / den
 *             else:
 */
      goto __pyx_L6;
    }

    /* "EvoDAG/cython_utils.pyx":490
 *                 f12 *= (2 * precision2[i] * recall2[i]) / den
 *             else:
 *                 f12 = 0             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
    /*else*/ {
      __pyx_v_f12 = 0.0;
    }
    __pyx_L6:;
  }

  /* "EvoDAG/cython_utils.pyx":491
 *             else:
 *                 f12 = 0
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_9;
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":471
 *         return f1, f12
 * 
 *     def g_F1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.g_F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":494
 * 
 *     @cython.cdivision(True)
 *     def a_F1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_25a_F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_25a_F1(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy = 0;
  arrayobject *__pyx_v_index = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("a_F1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_y,&__pyx_n_s_hy_2,&__pyx_n_s_index,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hy_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_F1", 1, 3, 3, 1); __PYX_ERR(0, 494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("a_F1", 1, 3, 3, 2); __PYX_ERR(0, 494, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "a_F1") < 0)) __PYX_ERR(0, 494, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_y = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[0]);
    __pyx_v_hy = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)values[1]);
    __pyx_v_index = ((arrayobject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("a_F1", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 494, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "y", 0))) __PYX_ERR(0, 494, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hy), __pyx_ptype_11SparseArray_12sparse_array_SparseArray, 1, "hy", 0))) __PYX_ERR(0, 494, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7cpython_5array_array, 1, "index", 0))) __PYX_ERR(0, 494, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_24a_F1(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), __pyx_v_y, __pyx_v_hy, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_24a_F1(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  double __pyx_v_f1;
  double __pyx_v_f12;
  double __pyx_v_den;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("a_F1", 0);

  /* "EvoDAG/cython_utils.pyx":495
 *     @cython.cdivision(True)
 *     def a_F1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)             # <<<<<<<<<<<<<<
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->count(__pyx_v_self, __pyx_v_y, __pyx_v_hy, __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":496
 *     def a_F1(self, SparseArray y, SparseArray hy, array.array index):
 *         self.count(y, hy, index)
 *         self.precision_recall()             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->precision_recall(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":497
 *         self.count(y, hy, index)
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":498
 *         self.precision_recall()
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":499
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 */
  __pyx_t_2 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":500
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_2 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":501
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 */
  __pyx_v_f1 = 0.0;
  __pyx_v_f12 = 0.0;

  /* "EvoDAG/cython_utils.pyx":502
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 */
  __pyx_v_i = 0;

  /* "EvoDAG/cython_utils.pyx":503
 *         cdef double f1 = 0, f12 = 0, den
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             den = precision[i] + recall[i]
 *             if den > 0:
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":504
 *         cdef Py_ssize_t i = 0
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision[__pyx_v_i]) + (__pyx_v_recall[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":505
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":506
 *             den = precision[i] + recall[i]
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den             # <<<<<<<<<<<<<<
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 */
      __pyx_v_f1 = (__pyx_v_f1 + (((2.0 * (__pyx_v_precision[__pyx_v_i])) * (__pyx_v_recall[__pyx_v_i])) / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":505
 *         for i in range(self.nclasses):
 *             den = precision[i] + recall[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 */
    }

    /* "EvoDAG/cython_utils.pyx":507
 *             if den > 0:
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]             # <<<<<<<<<<<<<<
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 */
    __pyx_v_den = ((__pyx_v_precision2[__pyx_v_i]) + (__pyx_v_recall2[__pyx_v_i]));

    /* "EvoDAG/cython_utils.pyx":508
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    __pyx_t_6 = ((__pyx_v_den > 0.0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":509
 *             den = precision2[i] + recall2[i]
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den             # <<<<<<<<<<<<<<
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 */
      __pyx_v_f12 = (__pyx_v_f12 + (((2.0 * (__pyx_v_precision2[__pyx_v_i])) * (__pyx_v_recall2[__pyx_v_i])) / __pyx_v_den));

      /* "EvoDAG/cython_utils.pyx":508
 *                 f1 += (2 * precision[i] * recall[i]) / den
 *             den = precision2[i] + recall2[i]
 *             if den > 0:             # <<<<<<<<<<<<<<
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 */
    }
  }

  /* "EvoDAG/cython_utils.pyx":510
 *             if den > 0:
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses             # <<<<<<<<<<<<<<
 *         f12 = f12 / self.nclasses
 *         return f1, f12
 */
  __pyx_v_f1 = (__pyx_v_f1 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":511
 *                 f12 += (2 * precision2[i] * recall2[i]) / den
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses             # <<<<<<<<<<<<<<
 *         return f1, f12
 * 
 */
  __pyx_v_f12 = (__pyx_v_f12 / __pyx_v_self->nclasses);

  /* "EvoDAG/cython_utils.pyx":512
 *         f1 = f1 / self.nclasses
 *         f12 = f12 / self.nclasses
 *         return f1, f12             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_f1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_f12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":494
 * 
 *     @cython.cdivision(True)
 *     def a_F1(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.count(y, hy, index)
 *         self.precision_recall()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.a_F1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":515
 * 
 *     @cython.cdivision(True)
 *     cdef precision_recall(self):             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_precision_recall(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  unsigned int *__pyx_v_precision_den;
  unsigned int *__pyx_v_recall_den;
  unsigned int *__pyx_v_precision2_den;
  unsigned int *__pyx_v_recall2_den;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double *__pyx_t_1;
  unsigned int *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("precision_recall", 0);

  /* "EvoDAG/cython_utils.pyx":516
 *     @cython.cdivision(True)
 *     cdef precision_recall(self):
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":517
 *     cdef precision_recall(self):
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_1 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":518
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 */
  __pyx_t_1 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":519
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 */
  __pyx_t_1 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":520
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 */
  __pyx_t_2 = __pyx_v_self->precision_den->data.as_uints;
  __pyx_v_precision_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":521
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 */
  __pyx_t_2 = __pyx_v_self->recall_den->data.as_uints;
  __pyx_v_recall_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":522
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef Py_ssize_t i
 */
  __pyx_t_2 = __pyx_v_self->precision2_den->data.as_uints;
  __pyx_v_precision2_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":523
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):
 */
  __pyx_t_2 = __pyx_v_self->recall2_den->data.as_uints;
  __pyx_v_recall2_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":525
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):             # <<<<<<<<<<<<<<
 *             if precision_den[i] > 0:
 *                 precision[i] = precision[i] / precision_den[i]
 */
  __pyx_t_3 = __pyx_v_self->nclasses;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "EvoDAG/cython_utils.pyx":526
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):
 *             if precision_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 precision[i] = precision[i] / precision_den[i]
 *             else:
 */
    __pyx_t_6 = (((__pyx_v_precision_den[__pyx_v_i]) > 0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":527
 *         for i in range(self.nclasses):
 *             if precision_den[i] > 0:
 *                 precision[i] = precision[i] / precision_den[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 precision[i] = 0
 */
      (__pyx_v_precision[__pyx_v_i]) = ((__pyx_v_precision[__pyx_v_i]) / (__pyx_v_precision_den[__pyx_v_i]));

      /* "EvoDAG/cython_utils.pyx":526
 *         cdef Py_ssize_t i
 *         for i in range(self.nclasses):
 *             if precision_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 precision[i] = precision[i] / precision_den[i]
 *             else:
 */
      goto __pyx_L5;
    }

    /* "EvoDAG/cython_utils.pyx":529
 *                 precision[i] = precision[i] / precision_den[i]
 *             else:
 *                 precision[i] = 0             # <<<<<<<<<<<<<<
 *             if recall_den[i] > 0:
 *                 recall[i] = recall[i] / recall_den[i]
 */
    /*else*/ {
      (__pyx_v_precision[__pyx_v_i]) = 0.0;
    }
    __pyx_L5:;

    /* "EvoDAG/cython_utils.pyx":530
 *             else:
 *                 precision[i] = 0
 *             if recall_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 recall[i] = recall[i] / recall_den[i]
 *             else:
 */
    __pyx_t_6 = (((__pyx_v_recall_den[__pyx_v_i]) > 0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":531
 *                 precision[i] = 0
 *             if recall_den[i] > 0:
 *                 recall[i] = recall[i] / recall_den[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 recall[i] = 0
 */
      (__pyx_v_recall[__pyx_v_i]) = ((__pyx_v_recall[__pyx_v_i]) / (__pyx_v_recall_den[__pyx_v_i]));

      /* "EvoDAG/cython_utils.pyx":530
 *             else:
 *                 precision[i] = 0
 *             if recall_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 recall[i] = recall[i] / recall_den[i]
 *             else:
 */
      goto __pyx_L6;
    }

    /* "EvoDAG/cython_utils.pyx":533
 *                 recall[i] = recall[i] / recall_den[i]
 *             else:
 *                 recall[i] = 0             # <<<<<<<<<<<<<<
 *             if precision2_den[i] > 0:
 *                 precision2[i] = precision2[i] / precision2_den[i]
 */
    /*else*/ {
      (__pyx_v_recall[__pyx_v_i]) = 0.0;
    }
    __pyx_L6:;

    /* "EvoDAG/cython_utils.pyx":534
 *             else:
 *                 recall[i] = 0
 *             if precision2_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 precision2[i] = precision2[i] / precision2_den[i]
 *             else:
 */
    __pyx_t_6 = (((__pyx_v_precision2_den[__pyx_v_i]) > 0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":535
 *                 recall[i] = 0
 *             if precision2_den[i] > 0:
 *                 precision2[i] = precision2[i] / precision2_den[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 precision2[i] = 0
 */
      (__pyx_v_precision2[__pyx_v_i]) = ((__pyx_v_precision2[__pyx_v_i]) / (__pyx_v_precision2_den[__pyx_v_i]));

      /* "EvoDAG/cython_utils.pyx":534
 *             else:
 *                 recall[i] = 0
 *             if precision2_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 precision2[i] = precision2[i] / precision2_den[i]
 *             else:
 */
      goto __pyx_L7;
    }

    /* "EvoDAG/cython_utils.pyx":537
 *                 precision2[i] = precision2[i] / precision2_den[i]
 *             else:
 *                 precision2[i] = 0             # <<<<<<<<<<<<<<
 *             if recall2_den[i] > 0:
 *                 recall2[i] = recall2[i] / recall2_den[i]
 */
    /*else*/ {
      (__pyx_v_precision2[__pyx_v_i]) = 0.0;
    }
    __pyx_L7:;

    /* "EvoDAG/cython_utils.pyx":538
 *             else:
 *                 precision2[i] = 0
 *             if recall2_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 recall2[i] = recall2[i] / recall2_den[i]
 *             else:
 */
    __pyx_t_6 = (((__pyx_v_recall2_den[__pyx_v_i]) > 0) != 0);
    if (__pyx_t_6) {

      /* "EvoDAG/cython_utils.pyx":539
 *                 precision2[i] = 0
 *             if recall2_den[i] > 0:
 *                 recall2[i] = recall2[i] / recall2_den[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 recall2[i] = 0
 */
      (__pyx_v_recall2[__pyx_v_i]) = ((__pyx_v_recall2[__pyx_v_i]) / (__pyx_v_recall2_den[__pyx_v_i]));

      /* "EvoDAG/cython_utils.pyx":538
 *             else:
 *                 precision2[i] = 0
 *             if recall2_den[i] > 0:             # <<<<<<<<<<<<<<
 *                 recall2[i] = recall2[i] / recall2_den[i]
 *             else:
 */
      goto __pyx_L8;
    }

    /* "EvoDAG/cython_utils.pyx":541
 *                 recall2[i] = recall2[i] / recall2_den[i]
 *             else:
 *                 recall2[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    /*else*/ {
      (__pyx_v_recall2[__pyx_v_i]) = 0.0;
    }
    __pyx_L8:;
  }

  /* "EvoDAG/cython_utils.pyx":515
 * 
 *     @cython.cdivision(True)
 *     cdef precision_recall(self):             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":544
 * 
 *     @cython.cdivision(True)
 *     cdef count(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.clean()
 *         cdef unsigned int *index_value = index.data.as_uints
 */

static PyObject *__pyx_f_6EvoDAG_12cython_utils_5Score_count(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_y, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_hy, arrayobject *__pyx_v_index) {
  unsigned int *__pyx_v_index_value;
  int __pyx_v_flag;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_size;
  unsigned int __pyx_v_k;
  unsigned int __pyx_v_end;
  double *__pyx_v_y_data;
  unsigned int *__pyx_v_y_index;
  double *__pyx_v_hy_data;
  unsigned int *__pyx_v_hy_index;
  double *__pyx_v_precision;
  double *__pyx_v_recall;
  double *__pyx_v_precision2;
  double *__pyx_v_recall2;
  unsigned int *__pyx_v_precision_den;
  unsigned int *__pyx_v_recall_den;
  unsigned int *__pyx_v_precision2_den;
  unsigned int *__pyx_v_recall2_den;
  int __pyx_v__y;
  int __pyx_v__hy;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned int *__pyx_t_2;
  unsigned int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  double *__pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("count", 0);

  /* "EvoDAG/cython_utils.pyx":545
 *     @cython.cdivision(True)
 *     cdef count(self, SparseArray y, SparseArray hy, array.array index):
 *         self.clean()             # <<<<<<<<<<<<<<
 *         cdef unsigned int *index_value = index.data.as_uints
 *         cdef bint flag
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->clean(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":546
 *     cdef count(self, SparseArray y, SparseArray hy, array.array index):
 *         self.clean()
 *         cdef unsigned int *index_value = index.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef bint flag
 *         cdef unsigned int i = 0, size = y._len
 */
  __pyx_t_2 = __pyx_v_index->data.as_uints;
  __pyx_v_index_value = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":548
 *         cdef unsigned int *index_value = index.data.as_uints
 *         cdef bint flag
 *         cdef unsigned int i = 0, size = y._len             # <<<<<<<<<<<<<<
 *         cdef unsigned int k = 0, end = len(index)
 *         cdef double *y_data = y.data.data.as_doubles
 */
  __pyx_v_i = 0;
  __pyx_t_3 = __pyx_v_y->_len;
  __pyx_v_size = __pyx_t_3;

  /* "EvoDAG/cython_utils.pyx":549
 *         cdef bint flag
 *         cdef unsigned int i = 0, size = y._len
 *         cdef unsigned int k = 0, end = len(index)             # <<<<<<<<<<<<<<
 *         cdef double *y_data = y.data.data.as_doubles
 *         cdef unsigned int *y_index = y.index.data.as_uints
 */
  __pyx_v_k = 0;
  if (unlikely(((PyObject *)__pyx_v_index) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_t_4 = Py_SIZE(((PyObject *)__pyx_v_index)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_v_end = __pyx_t_4;

  /* "EvoDAG/cython_utils.pyx":550
 *         cdef unsigned int i = 0, size = y._len
 *         cdef unsigned int k = 0, end = len(index)
 *         cdef double *y_data = y.data.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef unsigned int *y_index = y.index.data.as_uints
 *         cdef double *hy_data = hy.data.data.as_doubles
 */
  __pyx_t_5 = __pyx_v_y->data->data.as_doubles;
  __pyx_v_y_data = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":551
 *         cdef unsigned int k = 0, end = len(index)
 *         cdef double *y_data = y.data.data.as_doubles
 *         cdef unsigned int *y_index = y.index.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef double *hy_data = hy.data.data.as_doubles
 *         cdef unsigned int *hy_index = hy.index.data.as_uints
 */
  __pyx_t_2 = __pyx_v_y->index->data.as_uints;
  __pyx_v_y_index = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":552
 *         cdef double *y_data = y.data.data.as_doubles
 *         cdef unsigned int *y_index = y.index.data.as_uints
 *         cdef double *hy_data = hy.data.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef unsigned int *hy_index = hy.index.data.as_uints
 *         cdef double *precision = self.precision.data.as_doubles
 */
  __pyx_t_5 = __pyx_v_hy->data->data.as_doubles;
  __pyx_v_hy_data = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":553
 *         cdef unsigned int *y_index = y.index.data.as_uints
 *         cdef double *hy_data = hy.data.data.as_doubles
 *         cdef unsigned int *hy_index = hy.index.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 */
  __pyx_t_2 = __pyx_v_hy->index->data.as_uints;
  __pyx_v_hy_index = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":554
 *         cdef double *hy_data = hy.data.data.as_doubles
 *         cdef unsigned int *hy_index = hy.index.data.as_uints
 *         cdef double *precision = self.precision.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 */
  __pyx_t_5 = __pyx_v_self->precision->data.as_doubles;
  __pyx_v_precision = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":555
 *         cdef unsigned int *hy_index = hy.index.data.as_uints
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 */
  __pyx_t_5 = __pyx_v_self->recall->data.as_doubles;
  __pyx_v_recall = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":556
 *         cdef double *precision = self.precision.data.as_doubles
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 */
  __pyx_t_5 = __pyx_v_self->precision2->data.as_doubles;
  __pyx_v_precision2 = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":557
 *         cdef double *recall = self.recall.data.as_doubles
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles             # <<<<<<<<<<<<<<
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 */
  __pyx_t_5 = __pyx_v_self->recall2->data.as_doubles;
  __pyx_v_recall2 = __pyx_t_5;

  /* "EvoDAG/cython_utils.pyx":558
 *         cdef double *precision2 = self.precision2.data.as_doubles
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 */
  __pyx_t_2 = __pyx_v_self->precision_den->data.as_uints;
  __pyx_v_precision_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":559
 *         cdef double *recall2 = self.recall2.data.as_doubles
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 */
  __pyx_t_2 = __pyx_v_self->recall_den->data.as_uints;
  __pyx_v_recall_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":560
 *         cdef unsigned int *precision_den = self.precision_den.data.as_uints
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef int _y, _hy
 */
  __pyx_t_2 = __pyx_v_self->precision2_den->data.as_uints;
  __pyx_v_precision2_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":561
 *         cdef unsigned int *recall_den = self.recall_den.data.as_uints
 *         cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints             # <<<<<<<<<<<<<<
 *         cdef int _y, _hy
 *         self.y_size = y.non_zero
 */
  __pyx_t_2 = __pyx_v_self->recall2_den->data.as_uints;
  __pyx_v_recall2_den = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":563
 *         cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
 *         cdef int _y, _hy
 *         self.y_size = y.non_zero             # <<<<<<<<<<<<<<
 *         self.hy_size = hy.non_zero
 *         for i in range(size):
 */
  __pyx_t_3 = __pyx_v_y->non_zero;
  __pyx_v_self->y_size = __pyx_t_3;

  /* "EvoDAG/cython_utils.pyx":564
 *         cdef int _y, _hy
 *         self.y_size = y.non_zero
 *         self.hy_size = hy.non_zero             # <<<<<<<<<<<<<<
 *         for i in range(size):
 *             _y = <int> self.get_klass(y_data, y_index, i)
 */
  __pyx_t_3 = __pyx_v_hy->non_zero;
  __pyx_v_self->hy_size = __pyx_t_3;

  /* "EvoDAG/cython_utils.pyx":565
 *         self.y_size = y.non_zero
 *         self.hy_size = hy.non_zero
 *         for i in range(size):             # <<<<<<<<<<<<<<
 *             _y = <int> self.get_klass(y_data, y_index, i)
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_6 = __pyx_t_3;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "EvoDAG/cython_utils.pyx":566
 *         self.hy_size = hy.non_zero
 *         for i in range(size):
 *             _y = <int> self.get_klass(y_data, y_index, i)             # <<<<<<<<<<<<<<
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)
 *             flag = False
 */
    __pyx_v__y = ((int)((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->get_klass(__pyx_v_self, __pyx_v_y_data, __pyx_v_y_index, __pyx_v_i));

    /* "EvoDAG/cython_utils.pyx":567
 *         for i in range(size):
 *             _y = <int> self.get_klass(y_data, y_index, i)
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)             # <<<<<<<<<<<<<<
 *             flag = False
 *             if k < end:
 */
    __pyx_v__hy = ((int)((struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score *)__pyx_v_self->__pyx_vtab)->get_predicted(__pyx_v_self, __pyx_v_hy_data, __pyx_v_hy_index, __pyx_v_i));

    /* "EvoDAG/cython_utils.pyx":568
 *             _y = <int> self.get_klass(y_data, y_index, i)
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)
 *             flag = False             # <<<<<<<<<<<<<<
 *             if k < end:
 *                 if index_value[k] == i:
 */
    __pyx_v_flag = 0;

    /* "EvoDAG/cython_utils.pyx":569
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)
 *             flag = False
 *             if k < end:             # <<<<<<<<<<<<<<
 *                 if index_value[k] == i:
 *                     flag = True
 */
    __pyx_t_8 = ((__pyx_v_k < __pyx_v_end) != 0);
    if (__pyx_t_8) {

      /* "EvoDAG/cython_utils.pyx":570
 *             flag = False
 *             if k < end:
 *                 if index_value[k] == i:             # <<<<<<<<<<<<<<
 *                     flag = True
 *                     k += 1
 */
      __pyx_t_8 = (((__pyx_v_index_value[__pyx_v_k]) == __pyx_v_i) != 0);
      if (__pyx_t_8) {

        /* "EvoDAG/cython_utils.pyx":571
 *             if k < end:
 *                 if index_value[k] == i:
 *                     flag = True             # <<<<<<<<<<<<<<
 *                     k += 1
 *             if flag:
 */
        __pyx_v_flag = 1;

        /* "EvoDAG/cython_utils.pyx":572
 *                 if index_value[k] == i:
 *                     flag = True
 *                     k += 1             # <<<<<<<<<<<<<<
 *             if flag:
 *                 precision_den[_hy] += 1
 */
        __pyx_v_k = (__pyx_v_k + 1);

        /* "EvoDAG/cython_utils.pyx":570
 *             flag = False
 *             if k < end:
 *                 if index_value[k] == i:             # <<<<<<<<<<<<<<
 *                     flag = True
 *                     k += 1
 */
      }

      /* "EvoDAG/cython_utils.pyx":569
 *             _hy = <int> self.get_predicted(hy_data, hy_index, i)
 *             flag = False
 *             if k < end:             # <<<<<<<<<<<<<<
 *                 if index_value[k] == i:
 *                     flag = True
 */
    }

    /* "EvoDAG/cython_utils.pyx":573
 *                     flag = True
 *                     k += 1
 *             if flag:             # <<<<<<<<<<<<<<
 *                 precision_den[_hy] += 1
 *                 recall_den[_y] += 1
 */
    __pyx_t_8 = (__pyx_v_flag != 0);
    if (__pyx_t_8) {

      /* "EvoDAG/cython_utils.pyx":574
 *                     k += 1
 *             if flag:
 *                 precision_den[_hy] += 1             # <<<<<<<<<<<<<<
 *                 recall_den[_y] += 1
 *                 if _y == _hy:
 */
      __pyx_t_9 = __pyx_v__hy;
      (__pyx_v_precision_den[__pyx_t_9]) = ((__pyx_v_precision_den[__pyx_t_9]) + 1);

      /* "EvoDAG/cython_utils.pyx":575
 *             if flag:
 *                 precision_den[_hy] += 1
 *                 recall_den[_y] += 1             # <<<<<<<<<<<<<<
 *                 if _y == _hy:
 *                     precision[_y] += 1
 */
      __pyx_t_9 = __pyx_v__y;
      (__pyx_v_recall_den[__pyx_t_9]) = ((__pyx_v_recall_den[__pyx_t_9]) + 1);

      /* "EvoDAG/cython_utils.pyx":576
 *                 precision_den[_hy] += 1
 *                 recall_den[_y] += 1
 *                 if _y == _hy:             # <<<<<<<<<<<<<<
 *                     precision[_y] += 1
 *                     recall[_y] += 1
 */
      __pyx_t_8 = ((__pyx_v__y == __pyx_v__hy) != 0);
      if (__pyx_t_8) {

        /* "EvoDAG/cython_utils.pyx":577
 *                 recall_den[_y] += 1
 *                 if _y == _hy:
 *                     precision[_y] += 1             # <<<<<<<<<<<<<<
 *                     recall[_y] += 1
 *             else:
 */
        __pyx_t_9 = __pyx_v__y;
        (__pyx_v_precision[__pyx_t_9]) = ((__pyx_v_precision[__pyx_t_9]) + 1.0);

        /* "EvoDAG/cython_utils.pyx":578
 *                 if _y == _hy:
 *                     precision[_y] += 1
 *                     recall[_y] += 1             # <<<<<<<<<<<<<<
 *             else:
 *                 precision2_den[_hy] += 1
 */
        __pyx_t_9 = __pyx_v__y;
        (__pyx_v_recall[__pyx_t_9]) = ((__pyx_v_recall[__pyx_t_9]) + 1.0);

        /* "EvoDAG/cython_utils.pyx":576
 *                 precision_den[_hy] += 1
 *                 recall_den[_y] += 1
 *                 if _y == _hy:             # <<<<<<<<<<<<<<
 *                     precision[_y] += 1
 *                     recall[_y] += 1
 */
      }

      /* "EvoDAG/cython_utils.pyx":573
 *                     flag = True
 *                     k += 1
 *             if flag:             # <<<<<<<<<<<<<<
 *                 precision_den[_hy] += 1
 *                 recall_den[_y] += 1
 */
      goto __pyx_L7;
    }

    /* "EvoDAG/cython_utils.pyx":580
 *                     recall[_y] += 1
 *             else:
 *                 precision2_den[_hy] += 1             # <<<<<<<<<<<<<<
 *                 recall2_den[_y] += 1
 *                 if _y == _hy:
 */
    /*else*/ {
      __pyx_t_9 = __pyx_v__hy;
      (__pyx_v_precision2_den[__pyx_t_9]) = ((__pyx_v_precision2_den[__pyx_t_9]) + 1);

      /* "EvoDAG/cython_utils.pyx":581
 *             else:
 *                 precision2_den[_hy] += 1
 *                 recall2_den[_y] += 1             # <<<<<<<<<<<<<<
 *                 if _y == _hy:
 *                     precision2[_y] += 1
 */
      __pyx_t_9 = __pyx_v__y;
      (__pyx_v_recall2_den[__pyx_t_9]) = ((__pyx_v_recall2_den[__pyx_t_9]) + 1);

      /* "EvoDAG/cython_utils.pyx":582
 *                 precision2_den[_hy] += 1
 *                 recall2_den[_y] += 1
 *                 if _y == _hy:             # <<<<<<<<<<<<<<
 *                     precision2[_y] += 1
 *                     recall2[_y] += 1
 */
      __pyx_t_8 = ((__pyx_v__y == __pyx_v__hy) != 0);
      if (__pyx_t_8) {

        /* "EvoDAG/cython_utils.pyx":583
 *                 recall2_den[_y] += 1
 *                 if _y == _hy:
 *                     precision2[_y] += 1             # <<<<<<<<<<<<<<
 *                     recall2[_y] += 1
 * 
 */
        __pyx_t_9 = __pyx_v__y;
        (__pyx_v_precision2[__pyx_t_9]) = ((__pyx_v_precision2[__pyx_t_9]) + 1.0);

        /* "EvoDAG/cython_utils.pyx":584
 *                 if _y == _hy:
 *                     precision2[_y] += 1
 *                     recall2[_y] += 1             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
        __pyx_t_9 = __pyx_v__y;
        (__pyx_v_recall2[__pyx_t_9]) = ((__pyx_v_recall2[__pyx_t_9]) + 1.0);

        /* "EvoDAG/cython_utils.pyx":582
 *                 precision2_den[_hy] += 1
 *                 recall2_den[_y] += 1
 *                 if _y == _hy:             # <<<<<<<<<<<<<<
 *                     precision2[_y] += 1
 *                     recall2[_y] += 1
 */
      }
    }
    __pyx_L7:;
  }

  /* "EvoDAG/cython_utils.pyx":544
 * 
 *     @cython.cdivision(True)
 *     cdef count(self, SparseArray y, SparseArray hy, array.array index):             # <<<<<<<<<<<<<<
 *         self.clean()
 *         cdef unsigned int *index_value = index.data.as_uints
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":239
 * 
 * cdef class Score:
 *     cdef public array.array precision             # <<<<<<<<<<<<<<
 *     cdef public array.array precision_den
 *     cdef public array.array recall
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_9precision___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->precision));
  __pyx_r = ((PyObject *)__pyx_v_self->precision);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision));
  __pyx_v_self->precision = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.precision.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9precision_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->precision);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision));
  __pyx_v_self->precision = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":240
 * cdef class Score:
 *     cdef public array.array precision
 *     cdef public array.array precision_den             # <<<<<<<<<<<<<<
 *     cdef public array.array recall
 *     cdef public array.array recall_den
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->precision_den));
  __pyx_r = ((PyObject *)__pyx_v_self->precision_den);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision_den));
  __pyx_v_self->precision_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.precision_den.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_13precision_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->precision_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision_den));
  __pyx_v_self->precision_den = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":241
 *     cdef public array.array precision
 *     cdef public array.array precision_den
 *     cdef public array.array recall             # <<<<<<<<<<<<<<
 *     cdef public array.array recall_den
 *     cdef public array.array precision2
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_6recall___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->recall));
  __pyx_r = ((PyObject *)__pyx_v_self->recall);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall));
  __pyx_v_self->recall = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.recall.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_6recall_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->recall);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall));
  __pyx_v_self->recall = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":242
 *     cdef public array.array precision_den
 *     cdef public array.array recall
 *     cdef public array.array recall_den             # <<<<<<<<<<<<<<
 *     cdef public array.array precision2
 *     cdef public array.array precision2_den
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->recall_den));
  __pyx_r = ((PyObject *)__pyx_v_self->recall_den);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 242, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall_den));
  __pyx_v_self->recall_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.recall_den.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10recall_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->recall_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall_den));
  __pyx_v_self->recall_den = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":243
 *     cdef public array.array recall
 *     cdef public array.array recall_den
 *     cdef public array.array precision2             # <<<<<<<<<<<<<<
 *     cdef public array.array precision2_den
 *     cdef public array.array recall2
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->precision2));
  __pyx_r = ((PyObject *)__pyx_v_self->precision2);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2));
  __pyx_v_self->precision2 = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.precision2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10precision2_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->precision2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2));
  __pyx_v_self->precision2 = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":244
 *     cdef public array.array recall_den
 *     cdef public array.array precision2
 *     cdef public array.array precision2_den             # <<<<<<<<<<<<<<
 *     cdef public array.array recall2
 *     cdef public array.array recall2_den
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->precision2_den));
  __pyx_r = ((PyObject *)__pyx_v_self->precision2_den);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->precision2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2_den));
  __pyx_v_self->precision2_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.precision2_den.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_14precision2_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->precision2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->precision2_den));
  __pyx_v_self->precision2_den = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":245
 *     cdef public array.array precision2
 *     cdef public array.array precision2_den
 *     cdef public array.array recall2             # <<<<<<<<<<<<<<
 *     cdef public array.array recall2_den
 *     cdef public double _accuracy
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->recall2));
  __pyx_r = ((PyObject *)__pyx_v_self->recall2);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2));
  __pyx_v_self->recall2 = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.recall2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_7recall2_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->recall2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2));
  __pyx_v_self->recall2 = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":246
 *     cdef public array.array precision2_den
 *     cdef public array.array recall2
 *     cdef public array.array recall2_den             # <<<<<<<<<<<<<<
 *     cdef public double _accuracy
 *     cdef public double _accuracy2
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->recall2_den));
  __pyx_r = ((PyObject *)__pyx_v_self->recall2_den);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_5array_array))))) __PYX_ERR(0, 246, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->recall2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2_den));
  __pyx_v_self->recall2_den = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.recall2_den.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_4__del__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_11recall2_den_4__del__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->recall2_den);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->recall2_den));
  __pyx_v_self->recall2_den = ((arrayobject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":247
 *     cdef public array.array recall2
 *     cdef public array.array recall2_den
 *     cdef public double _accuracy             # <<<<<<<<<<<<<<
 *     cdef public double _accuracy2
 *     cdef unsigned int y_pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_accuracy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score._accuracy.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_9_accuracy_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L1_error)
  __pyx_v_self->_accuracy = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score._accuracy.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":248
 *     cdef public array.array recall2_den
 *     cdef public double _accuracy
 *     cdef public double _accuracy2             # <<<<<<<<<<<<<<
 *     cdef unsigned int y_pos
 *     cdef unsigned int hy_pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2___get__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2___get__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_accuracy2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score._accuracy2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2_2__set__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6EvoDAG_12cython_utils_5Score_10_accuracy2_2__set__(struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 248, __pyx_L1_error)
  __pyx_v_self->_accuracy2 = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score._accuracy2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_26__reduce_cython__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_5Score_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_5Score_28__setstate_cython__(((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_5Score_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6EvoDAG_12cython_utils_Score *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.Score.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EvoDAG/cython_utils.pyx":587
 * 
 * @cython.cdivision(True)
 * cpdef list centroid_coef(list vars, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den, coef
 *     cdef unsigned int *mask_value = mask.data.as_uints
 */

static PyObject *__pyx_pw_6EvoDAG_12cython_utils_17centroid_coef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6EvoDAG_12cython_utils_centroid_coef(PyObject *__pyx_v_vars, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass, CYTHON_UNUSED int __pyx_skip_dispatch) {
  arrayobject *__pyx_v_mean_num = 0;
  arrayobject *__pyx_v_mean_den = 0;
  arrayobject *__pyx_v_coef = 0;
  unsigned int *__pyx_v_mask_value;
  unsigned int *__pyx_v_a;
  unsigned int *__pyx_v_klass_value;
  unsigned int __pyx_v_index;
  double *__pyx_v_a_value;
  double *__pyx_v_coef_value;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_k;
  Py_ssize_t __pyx_v_nvars;
  Py_ssize_t __pyx_v_nmask;
  PyObject *__pyx_v_res = 0;
  struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *__pyx_v_var = 0;
  double *__pyx_v_mean_num_value;
  double *__pyx_v_mean_den_value;
  unsigned int __pyx_v_var_end;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  double *__pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  unsigned int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  unsigned int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  __Pyx_RefNannySetupContext("centroid_coef", 0);

  /* "EvoDAG/cython_utils.pyx":589
 * cpdef list centroid_coef(list vars, array.array klass, array.array mask, int nclass):
 *     cdef array.array mean_num, mean_den, coef
 *     cdef unsigned int *mask_value = mask.data.as_uints             # <<<<<<<<<<<<<<
 *     cdef unsigned int *a, *klass_value = klass.data.as_uints, index
 *     cdef double *a_value, *coef_value
 */
  __pyx_t_1 = __pyx_v_mask->data.as_uints;
  __pyx_v_mask_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":590
 *     cdef array.array mean_num, mean_den, coef
 *     cdef unsigned int *mask_value = mask.data.as_uints
 *     cdef unsigned int *a, *klass_value = klass.data.as_uints, index             # <<<<<<<<<<<<<<
 *     cdef double *a_value, *coef_value
 *     cdef Py_ssize_t i, j, k, nvars = PyList_GET_SIZE(vars), nmask = len(mask)
 */
  __pyx_t_1 = __pyx_v_klass->data.as_uints;
  __pyx_v_klass_value = __pyx_t_1;

  /* "EvoDAG/cython_utils.pyx":592
 *     cdef unsigned int *a, *klass_value = klass.data.as_uints, index
 *     cdef double *a_value, *coef_value
 *     cdef Py_ssize_t i, j, k, nvars = PyList_GET_SIZE(vars), nmask = len(mask)             # <<<<<<<<<<<<<<
 *     cdef list res = []
 *     cdef SparseArray var = <SparseArray> PyList_GET_ITEM(vars, 0)
 */
  __pyx_v_nvars = PyList_GET_SIZE(__pyx_v_vars);
  if (unlikely(((PyObject *)__pyx_v_mask) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 592, __pyx_L1_error)
  }
  __pyx_t_2 = Py_SIZE(((PyObject *)__pyx_v_mask)); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 592, __pyx_L1_error)
  __pyx_v_nmask = __pyx_t_2;

  /* "EvoDAG/cython_utils.pyx":593
 *     cdef double *a_value, *coef_value
 *     cdef Py_ssize_t i, j, k, nvars = PyList_GET_SIZE(vars), nmask = len(mask)
 *     cdef list res = []             # <<<<<<<<<<<<<<
 *     cdef SparseArray var = <SparseArray> PyList_GET_ITEM(vars, 0)
 *     mean_num = array.clone(var.data, nclass, zero=False)
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_res = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":594
 *     cdef Py_ssize_t i, j, k, nvars = PyList_GET_SIZE(vars), nmask = len(mask)
 *     cdef list res = []
 *     cdef SparseArray var = <SparseArray> PyList_GET_ITEM(vars, 0)             # <<<<<<<<<<<<<<
 *     mean_num = array.clone(var.data, nclass, zero=False)
 *     cdef double *mean_num_value = mean_num.data.as_doubles
 */
  __pyx_t_4 = PyList_GET_ITEM(__pyx_v_vars, 0);
  __pyx_t_3 = ((PyObject *)__pyx_t_4);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_var = ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":595
 *     cdef list res = []
 *     cdef SparseArray var = <SparseArray> PyList_GET_ITEM(vars, 0)
 *     mean_num = array.clone(var.data, nclass, zero=False)             # <<<<<<<<<<<<<<
 *     cdef double *mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=False)
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_3), __pyx_v_nclass, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_num = ((arrayobject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "EvoDAG/cython_utils.pyx":596
 *     cdef SparseArray var = <SparseArray> PyList_GET_ITEM(vars, 0)
 *     mean_num = array.clone(var.data, nclass, zero=False)
 *     cdef double *mean_num_value = mean_num.data.as_doubles             # <<<<<<<<<<<<<<
 *     mean_den = array.clone(var.data, nclass, zero=False)
 *     cdef double *mean_den_value = mean_den.data.as_doubles
 */
  __pyx_t_6 = __pyx_v_mean_num->data.as_doubles;
  __pyx_v_mean_num_value = __pyx_t_6;

  /* "EvoDAG/cython_utils.pyx":597
 *     mean_num = array.clone(var.data, nclass, zero=False)
 *     cdef double *mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=False)             # <<<<<<<<<<<<<<
 *     cdef double *mean_den_value = mean_den.data.as_doubles
 *     cdef unsigned int var_end
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_var->data);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_5), __pyx_v_nclass, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_mean_den = ((arrayobject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "EvoDAG/cython_utils.pyx":598
 *     cdef double *mean_num_value = mean_num.data.as_doubles
 *     mean_den = array.clone(var.data, nclass, zero=False)
 *     cdef double *mean_den_value = mean_den.data.as_doubles             # <<<<<<<<<<<<<<
 *     cdef unsigned int var_end
 *     for k in range(nclass):
 */
  __pyx_t_6 = __pyx_v_mean_den->data.as_doubles;
  __pyx_v_mean_den_value = __pyx_t_6;

  /* "EvoDAG/cython_utils.pyx":600
 *     cdef double *mean_den_value = mean_den.data.as_doubles
 *     cdef unsigned int var_end
 *     for k in range(nclass):             # <<<<<<<<<<<<<<
 *         coef = array.clone(var.data, nvars, zero=True)
 *         PyList_Append(res, coef)
 */
  __pyx_t_7 = __pyx_v_nclass;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_8; __pyx_t_2+=1) {
    __pyx_v_k = __pyx_t_2;

    /* "EvoDAG/cython_utils.pyx":601
 *     cdef unsigned int var_end
 *     for k in range(nclass):
 *         coef = array.clone(var.data, nvars, zero=True)             # <<<<<<<<<<<<<<
 *         PyList_Append(res, coef)
 *     for i in range(nvars):
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_var->data);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_3), __pyx_v_nvars, 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_coef, ((arrayobject *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "EvoDAG/cython_utils.pyx":602
 *     for k in range(nclass):
 *         coef = array.clone(var.data, nvars, zero=True)
 *         PyList_Append(res, coef)             # <<<<<<<<<<<<<<
 *     for i in range(nvars):
 *         for k in range(nclass):
 */
    __pyx_t_9 = PyList_Append(__pyx_v_res, ((PyObject *)__pyx_v_coef)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 602, __pyx_L1_error)
  }

  /* "EvoDAG/cython_utils.pyx":603
 *         coef = array.clone(var.data, nvars, zero=True)
 *         PyList_Append(res, coef)
 *     for i in range(nvars):             # <<<<<<<<<<<<<<
 *         for k in range(nclass):
 *             mean_num_value[k] = 0
 */
  __pyx_t_2 = __pyx_v_nvars;
  __pyx_t_10 = __pyx_t_2;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "EvoDAG/cython_utils.pyx":604
 *         PyList_Append(res, coef)
 *     for i in range(nvars):
 *         for k in range(nclass):             # <<<<<<<<<<<<<<
 *             mean_num_value[k] = 0
 *             mean_den_value[k] = 0
 */
    __pyx_t_7 = __pyx_v_nclass;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_8; __pyx_t_12+=1) {
      __pyx_v_k = __pyx_t_12;

      /* "EvoDAG/cython_utils.pyx":605
 *     for i in range(nvars):
 *         for k in range(nclass):
 *             mean_num_value[k] = 0             # <<<<<<<<<<<<<<
 *             mean_den_value[k] = 0
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)
 */
      (__pyx_v_mean_num_value[__pyx_v_k]) = 0.0;

      /* "EvoDAG/cython_utils.pyx":606
 *         for k in range(nclass):
 *             mean_num_value[k] = 0
 *             mean_den_value[k] = 0             # <<<<<<<<<<<<<<
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)
 *         var_end = var.non_zero
 */
      (__pyx_v_mean_den_value[__pyx_v_k]) = 0.0;
    }

    /* "EvoDAG/cython_utils.pyx":607
 *             mean_num_value[k] = 0
 *             mean_den_value[k] = 0
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)             # <<<<<<<<<<<<<<
 *         var_end = var.non_zero
 *         if var_end == 0:
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_vars, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_var, ((struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "EvoDAG/cython_utils.pyx":608
 *             mean_den_value[k] = 0
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)
 *         var_end = var.non_zero             # <<<<<<<<<<<<<<
 *         if var_end == 0:
 *             continue
 */
    __pyx_t_13 = __pyx_v_var->non_zero;
    __pyx_v_var_end = __pyx_t_13;

    /* "EvoDAG/cython_utils.pyx":609
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)
 *         var_end = var.non_zero
 *         if var_end == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         a = var.index.data.as_uints
 */
    __pyx_t_14 = ((__pyx_v_var_end == 0) != 0);
    if (__pyx_t_14) {

      /* "EvoDAG/cython_utils.pyx":610
 *         var_end = var.non_zero
 *         if var_end == 0:
 *             continue             # <<<<<<<<<<<<<<
 *         a = var.index.data.as_uints
 *         a_value = var.data.data.as_doubles
 */
      goto __pyx_L5_continue;

      /* "EvoDAG/cython_utils.pyx":609
 *         var = <SparseArray> PyList_GET_ITEM(vars, i)
 *         var_end = var.non_zero
 *         if var_end == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         a = var.index.data.as_uints
 */
    }

    /* "EvoDAG/cython_utils.pyx":611
 *         if var_end == 0:
 *             continue
 *         a = var.index.data.as_uints             # <<<<<<<<<<<<<<
 *         a_value = var.data.data.as_doubles
 *         k = 0
 */
    __pyx_t_1 = __pyx_v_var->index->data.as_uints;
    __pyx_v_a = __pyx_t_1;

    /* "EvoDAG/cython_utils.pyx":612
 *             continue
 *         a = var.index.data.as_uints
 *         a_value = var.data.data.as_doubles             # <<<<<<<<<<<<<<
 *         k = 0
 *         for j in range(nmask):
 */
    __pyx_t_6 = __pyx_v_var->data->data.as_doubles;
    __pyx_v_a_value = __pyx_t_6;

    /* "EvoDAG/cython_utils.pyx":613
 *         a = var.index.data.as_uints
 *         a_value = var.data.data.as_doubles
 *         k = 0             # <<<<<<<<<<<<<<
 *         for j in range(nmask):
 *             index = mask_value[j]
 */
    __pyx_v_k = 0;

    /* "EvoDAG/cython_utils.pyx":614
 *         a_value = var.data.data.as_doubles
 *         k = 0
 *         for j in range(nmask):             # <<<<<<<<<<<<<<
 *             index = mask_value[j]
 *             for k in range(k, var_end):
 */
    __pyx_t_12 = __pyx_v_nmask;
    __pyx_t_15 = __pyx_t_12;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "EvoDAG/cython_utils.pyx":615
 *         k = 0
 *         for j in range(nmask):
 *             index = mask_value[j]             # <<<<<<<<<<<<<<
 *             for k in range(k, var_end):
 *                 if a[k] >= index:
 */
      __pyx_v_index = (__pyx_v_mask_value[__pyx_v_j]);

      /* "EvoDAG/cython_utils.pyx":616
 *         for j in range(nmask):
 *             index = mask_value[j]
 *             for k in range(k, var_end):             # <<<<<<<<<<<<<<
 *                 if a[k] >= index:
 *                     break
 */
      __pyx_t_13 = __pyx_v_var_end;
      __pyx_t_17 = __pyx_t_13;
      for (__pyx_t_18 = __pyx_v_k; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_k = __pyx_t_18;

        /* "EvoDAG/cython_utils.pyx":617
 *             index = mask_value[j]
 *             for k in range(k, var_end):
 *                 if a[k] >= index:             # <<<<<<<<<<<<<<
 *                     break
 *             if a[k] == index:
 */
        __pyx_t_14 = (((__pyx_v_a[__pyx_v_k]) >= __pyx_v_index) != 0);
        if (__pyx_t_14) {

          /* "EvoDAG/cython_utils.pyx":618
 *             for k in range(k, var_end):
 *                 if a[k] >= index:
 *                     break             # <<<<<<<<<<<<<<
 *             if a[k] == index:
 *                 mean_num_value[klass_value[index]] += a_value[k]
 */
          goto __pyx_L13_break;

          /* "EvoDAG/cython_utils.pyx":617
 *             index = mask_value[j]
 *             for k in range(k, var_end):
 *                 if a[k] >= index:             # <<<<<<<<<<<<<<
 *                     break
 *             if a[k] == index:
 */
        }
      }
      __pyx_L13_break:;

      /* "EvoDAG/cython_utils.pyx":619
 *                 if a[k] >= index:
 *                     break
 *             if a[k] == index:             # <<<<<<<<<<<<<<
 *                 mean_num_value[klass_value[index]] += a_value[k]
 *                 mean_den_value[klass_value[index]] += 1
 */
      __pyx_t_14 = (((__pyx_v_a[__pyx_v_k]) == __pyx_v_index) != 0);
      if (__pyx_t_14) {

        /* "EvoDAG/cython_utils.pyx":620
 *                     break
 *             if a[k] == index:
 *                 mean_num_value[klass_value[index]] += a_value[k]             # <<<<<<<<<<<<<<
 *                 mean_den_value[klass_value[index]] += 1
 *             else:
 */
        __pyx_t_13 = (__pyx_v_klass_value[__pyx_v_index]);
        (__pyx_v_mean_num_value[__pyx_t_13]) = ((__pyx_v_mean_num_value[__pyx_t_13]) + (__pyx_v_a_value[__pyx_v_k]));

        /* "EvoDAG/cython_utils.pyx":621
 *             if a[k] == index:
 *                 mean_num_value[klass_value[index]] += a_value[k]
 *                 mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *             else:
 *                 mean_den_value[klass_value[index]] += 1
 */
        __pyx_t_13 = (__pyx_v_klass_value[__pyx_v_index]);
        (__pyx_v_mean_den_value[__pyx_t_13]) = ((__pyx_v_mean_den_value[__pyx_t_13]) + 1.0);

        /* "EvoDAG/cython_utils.pyx":619
 *                 if a[k] >= index:
 *                     break
 *             if a[k] == index:             # <<<<<<<<<<<<<<
 *                 mean_num_value[klass_value[index]] += a_value[k]
 *                 mean_den_value[klass_value[index]] += 1
 */
        goto __pyx_L15;
      }

      /* "EvoDAG/cython_utils.pyx":623
 *                 mean_den_value[klass_value[index]] += 1
 *             else:
 *                 mean_den_value[klass_value[index]] += 1             # <<<<<<<<<<<<<<
 *         for k in range(nclass):
 *             coef = <array.array> PyList_GET_ITEM(res, k)
 */
      /*else*/ {
        __pyx_t_13 = (__pyx_v_klass_value[__pyx_v_index]);
        (__pyx_v_mean_den_value[__pyx_t_13]) = ((__pyx_v_mean_den_value[__pyx_t_13]) + 1.0);
      }
      __pyx_L15:;
    }

    /* "EvoDAG/cython_utils.pyx":624
 *             else:
 *                 mean_den_value[klass_value[index]] += 1
 *         for k in range(nclass):             # <<<<<<<<<<<<<<
 *             coef = <array.array> PyList_GET_ITEM(res, k)
 *             coef_value = coef.data.as_doubles
 */
    __pyx_t_7 = __pyx_v_nclass;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_8; __pyx_t_12+=1) {
      __pyx_v_k = __pyx_t_12;

      /* "EvoDAG/cython_utils.pyx":625
 *                 mean_den_value[klass_value[index]] += 1
 *         for k in range(nclass):
 *             coef = <array.array> PyList_GET_ITEM(res, k)             # <<<<<<<<<<<<<<
 *             coef_value = coef.data.as_doubles
 *             coef_value[i] = - mean_num_value[k] / mean_den_value[k]
 */
      __pyx_t_4 = PyList_GET_ITEM(__pyx_v_res, __pyx_v_k);
      __pyx_t_5 = ((PyObject *)__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_coef, ((arrayobject *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "EvoDAG/cython_utils.pyx":626
 *         for k in range(nclass):
 *             coef = <array.array> PyList_GET_ITEM(res, k)
 *             coef_value = coef.data.as_doubles             # <<<<<<<<<<<<<<
 *             coef_value[i] = - mean_num_value[k] / mean_den_value[k]
 *     return res
 */
      __pyx_t_6 = __pyx_v_coef->data.as_doubles;
      __pyx_v_coef_value = __pyx_t_6;

      /* "EvoDAG/cython_utils.pyx":627
 *             coef = <array.array> PyList_GET_ITEM(res, k)
 *             coef_value = coef.data.as_doubles
 *             coef_value[i] = - mean_num_value[k] / mean_den_value[k]             # <<<<<<<<<<<<<<
 *     return res
 */
      (__pyx_v_coef_value[__pyx_v_i]) = ((-(__pyx_v_mean_num_value[__pyx_v_k])) / (__pyx_v_mean_den_value[__pyx_v_k]));
    }
    __pyx_L5_continue:;
  }

  /* "EvoDAG/cython_utils.pyx":628
 *             coef_value = coef.data.as_doubles
 *             coef_value[i] = - mean_num_value[k] / mean_den_value[k]
 *     return res             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EvoDAG/cython_utils.pyx":587
 * 
 * @cython.cdivision(True)
 * cpdef list centroid_coef(list vars, array.array klass, array.array mask, int nclass):             # <<<<<<<<<<<<<<
 *     cdef array.array mean_num, mean_den, coef
 *     cdef unsigned int *mask_value = mask.data.as_uints
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EvoDAG.cython_utils.centroid_coef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_num);
  __Pyx_XDECREF((PyObject *)__pyx_v_mean_den);
  __Pyx_XDECREF((PyObject *)__pyx_v_coef);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF((PyObject *)__pyx_v_var);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_17centroid_coef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6EvoDAG_12cython_utils_17centroid_coef(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_vars = 0;
  arrayobject *__pyx_v_klass = 0;
  arrayobject *__pyx_v_mask = 0;
  int __pyx_v_nclass;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("centroid_coef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vars,&__pyx_n_s_klass,&__pyx_n_s_mask_2,&__pyx_n_s_nclass,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_vars)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_klass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("centroid_coef", 1, 4, 4, 1); __PYX_ERR(0, 587, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("centroid_coef", 1, 4, 4, 2); __PYX_ERR(0, 587, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nclass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("centroid_coef", 1, 4, 4, 3); __PYX_ERR(0, 587, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "centroid_coef") < 0)) __PYX_ERR(0, 587, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_vars = ((PyObject*)values[0]);
    __pyx_v_klass = ((arrayobject *)values[1]);
    __pyx_v_mask = ((arrayobject *)values[2]);
    __pyx_v_nclass = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nclass == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("centroid_coef", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 587, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EvoDAG.cython_utils.centroid_coef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vars), (&PyList_Type), 1, "vars", 1))) __PYX_ERR(0, 587, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_klass), __pyx_ptype_7cpython_5array_array, 1, "klass", 0))) __PYX_ERR(0, 587, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_7cpython_5array_array, 1, "mask", 0))) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_r = __pyx_pf_6EvoDAG_12cython_utils_16centroid_coef(__pyx_self, __pyx_v_vars, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6EvoDAG_12cython_utils_16centroid_coef(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_vars, arrayobject *__pyx_v_klass, arrayobject *__pyx_v_mask, int __pyx_v_nclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("centroid_coef", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6EvoDAG_12cython_utils_centroid_coef(__pyx_v_vars, __pyx_v_klass, __pyx_v_mask, __pyx_v_nclass, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EvoDAG.cython_utils.centroid_coef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  PyObject *__pyx_v_item_count = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  char __pyx_t_7;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "array.pxd":98
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */
  __pyx_t_1 = PyInt_FromSsize_t(Py_SIZE(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "array.pxd":100
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "array.pxd":101
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */
  __pyx_t_2 = __pyx_v_self->data.as_chars;
  __pyx_v_info->buf = __pyx_t_2;

  /* "array.pxd":102
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */
  __pyx_v_info->readonly = 0;

  /* "array.pxd":103
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */
  __pyx_v_info->ndim = 1;

  /* "array.pxd":104
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */
  __pyx_t_3 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "array.pxd":105
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_info->itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_item_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_info->len = __pyx_t_5;

  /* "array.pxd":107
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  __pyx_t_6 = ((!(__pyx_v_info->shape != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "array.pxd":109
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */
    PyErr_NoMemory(); __PYX_ERR(2, 109, __pyx_L1_error)

    /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  }

  /* "array.pxd":110
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_item_count); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 110, __pyx_L1_error)
  (__pyx_v_info->shape[0]) = __pyx_t_5;

  /* "array.pxd":111
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "array.pxd":113
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "array.pxd":114
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->ob_descr->typecode;
  (__pyx_v_info->format[0]) = __pyx_t_7;

  /* "array.pxd":115
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */
  (__pyx_v_info->format[1]) = 0;

  /* "array.pxd":116
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_item_count);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "array.pxd":119
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */
  PyObject_Free(__pyx_v_info->shape);

  /* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = NULL;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "array.pxd":134
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_3 = (__pyx_v_zero != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "array.pxd":136
 *     op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  }

  /* "array.pxd":137
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = NULL;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "array.pxd":141
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":142
 *     """ make a copy of an array. """
 *     op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "array.pxd":143
 *     op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize;
  Py_ssize_t __pyx_v_origsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("extend_buffer", 0);

  /* "array.pxd":149
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "array.pxd":150
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "array.pxd":151
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 151, __pyx_L1_error)

  /* "array.pxd":152
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "array.pxd":153
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  __pyx_t_1 = ((__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode) != 0);
  if (__pyx_t_1) {

    /* "array.pxd":158
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 158, __pyx_L1_error)

    /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  }

  /* "array.pxd":159
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 159, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zero", 0);

  /* "array.pxd":163
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "SparseArray/sparse_array.pxd":23
 * 
 * 
 * cdef inline double non_op(double a):             # <<<<<<<<<<<<<<
 *     return a
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_non_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("non_op", 0);

  /* "SparseArray/sparse_array.pxd":24
 * 
 * cdef inline double non_op(double a):
 *     return a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_a;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":23
 * 
 * 
 * cdef inline double non_op(double a):             # <<<<<<<<<<<<<<
 *     return a
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":27
 * 
 * 
 * cdef inline double add_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a + b
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_add_op(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_op", 0);

  /* "SparseArray/sparse_array.pxd":28
 * 
 * cdef inline double add_op(double a, double b):
 *     return a + b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a + __pyx_v_b);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":27
 * 
 * 
 * cdef inline double add_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a + b
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":31
 * 
 * 
 * cdef inline double sub_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a - b
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_sub_op(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sub_op", 0);

  /* "SparseArray/sparse_array.pxd":32
 * 
 * cdef inline double sub_op(double a, double b):
 *     return a - b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a - __pyx_v_b);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":31
 * 
 * 
 * cdef inline double sub_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a - b
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":35
 * 
 * 
 * cdef inline double minus_op(double a):             # <<<<<<<<<<<<<<
 *     return -a
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_minus_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("minus_op", 0);

  /* "SparseArray/sparse_array.pxd":36
 * 
 * cdef inline double minus_op(double a):
 *     return -a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (-__pyx_v_a);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":35
 * 
 * 
 * cdef inline double minus_op(double a):             # <<<<<<<<<<<<<<
 *     return -a
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":39
 * 
 * 
 * cdef inline double zero_min_op(double a):             # <<<<<<<<<<<<<<
 *     return math.fmin(a, 0)
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_zero_min_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zero_min_op", 0);

  /* "SparseArray/sparse_array.pxd":40
 * 
 * cdef inline double zero_min_op(double a):
 *     return math.fmin(a, 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = fmin(__pyx_v_a, 0.0);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":39
 * 
 * 
 * cdef inline double zero_min_op(double a):             # <<<<<<<<<<<<<<
 *     return math.fmin(a, 0)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":43
 * 
 * 
 * cdef inline double zero_max_op(double a):             # <<<<<<<<<<<<<<
 *     return math.fmax(a, 0)
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_zero_max_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zero_max_op", 0);

  /* "SparseArray/sparse_array.pxd":44
 * 
 * cdef inline double zero_max_op(double a):
 *     return math.fmax(a, 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = fmax(__pyx_v_a, 0.0);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":43
 * 
 * 
 * cdef inline double zero_max_op(double a):             # <<<<<<<<<<<<<<
 *     return math.fmax(a, 0)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":47
 * 
 * 
 * cdef inline double div_left_op(double a):             # <<<<<<<<<<<<<<
 *     return math.INFINITY
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_div_left_op(CYTHON_UNUSED double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("div_left_op", 0);

  /* "SparseArray/sparse_array.pxd":48
 * 
 * cdef inline double div_left_op(double a):
 *     return math.INFINITY             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = INFINITY;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":47
 * 
 * 
 * cdef inline double div_left_op(double a):             # <<<<<<<<<<<<<<
 *     return math.INFINITY
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":51
 * 
 * 
 * cdef inline double div_right_op(double a):             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_div_right_op(CYTHON_UNUSED double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("div_right_op", 0);

  /* "SparseArray/sparse_array.pxd":52
 * 
 * cdef inline double div_right_op(double a):
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":51
 * 
 * 
 * cdef inline double div_right_op(double a):             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":55
 * 
 * 
 * cdef inline double atan2_left_op(double a):             # <<<<<<<<<<<<<<
 *     return math.atan2(a, 0)
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_atan2_left_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("atan2_left_op", 0);

  /* "SparseArray/sparse_array.pxd":56
 * 
 * cdef inline double atan2_left_op(double a):
 *     return math.atan2(a, 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = atan2(__pyx_v_a, 0.0);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":55
 * 
 * 
 * cdef inline double atan2_left_op(double a):             # <<<<<<<<<<<<<<
 *     return math.atan2(a, 0)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":59
 * 
 * 
 * cdef inline double atan2_right_op(double a):             # <<<<<<<<<<<<<<
 *     return math.atan2(0, a)
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_atan2_right_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("atan2_right_op", 0);

  /* "SparseArray/sparse_array.pxd":60
 * 
 * cdef inline double atan2_right_op(double a):
 *     return math.atan2(0, a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = atan2(0.0, __pyx_v_a);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":59
 * 
 * 
 * cdef inline double atan2_right_op(double a):             # <<<<<<<<<<<<<<
 *     return math.atan2(0, a)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":64
 * 
 * @cython.cdivision(True)
 * cdef inline double div_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a / b
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_div_op(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("div_op", 0);

  /* "SparseArray/sparse_array.pxd":65
 * @cython.cdivision(True)
 * cdef inline double div_op(double a, double b):
 *     return a / b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a / __pyx_v_b);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":64
 * 
 * @cython.cdivision(True)
 * cdef inline double div_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a / b
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":68
 * 
 * 
 * cdef inline double sq_op(double a):             # <<<<<<<<<<<<<<
 *     return a * a
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_sq_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sq_op", 0);

  /* "SparseArray/sparse_array.pxd":69
 * 
 * cdef inline double sq_op(double a):
 *     return a * a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a * __pyx_v_a);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":68
 * 
 * 
 * cdef inline double sq_op(double a):             # <<<<<<<<<<<<<<
 *     return a * a
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":72
 * 
 * 
 * cdef inline double sign_op(double a):             # <<<<<<<<<<<<<<
 *     if a > 0:
 *         return 1
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_sign_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("sign_op", 0);

  /* "SparseArray/sparse_array.pxd":73
 * 
 * cdef inline double sign_op(double a):
 *     if a > 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif a < 0:
 */
  __pyx_t_1 = ((__pyx_v_a > 0.0) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":74
 * cdef inline double sign_op(double a):
 *     if a > 0:
 *         return 1             # <<<<<<<<<<<<<<
 *     elif a < 0:
 *         return -1
 */
    __pyx_r = 1.0;
    goto __pyx_L0;

    /* "SparseArray/sparse_array.pxd":73
 * 
 * cdef inline double sign_op(double a):
 *     if a > 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif a < 0:
 */
  }

  /* "SparseArray/sparse_array.pxd":75
 *     if a > 0:
 *         return 1
 *     elif a < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 0
 */
  __pyx_t_1 = ((__pyx_v_a < 0.0) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":76
 *         return 1
 *     elif a < 0:
 *         return -1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = -1.0;
    goto __pyx_L0;

    /* "SparseArray/sparse_array.pxd":75
 *     if a > 0:
 *         return 1
 *     elif a < 0:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 0
 */
  }

  /* "SparseArray/sparse_array.pxd":77
 *     elif a < 0:
 *         return -1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":72
 * 
 * 
 * cdef inline double sign_op(double a):             # <<<<<<<<<<<<<<
 *     if a > 0:
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":80
 * 
 * 
 * cdef inline double finite_op(double a):             # <<<<<<<<<<<<<<
 *     if math.isfinite(a):
 *         return a
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_finite_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("finite_op", 0);

  /* "SparseArray/sparse_array.pxd":81
 * 
 * cdef inline double finite_op(double a):
 *     if math.isfinite(a):             # <<<<<<<<<<<<<<
 *         return a
 *     return 0
 */
  __pyx_t_1 = (isfinite(__pyx_v_a) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":82
 * cdef inline double finite_op(double a):
 *     if math.isfinite(a):
 *         return a             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;

    /* "SparseArray/sparse_array.pxd":81
 * 
 * cdef inline double finite_op(double a):
 *     if math.isfinite(a):             # <<<<<<<<<<<<<<
 *         return a
 *     return 0
 */
  }

  /* "SparseArray/sparse_array.pxd":83
 *     if math.isfinite(a):
 *         return a
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":80
 * 
 * 
 * cdef inline double finite_op(double a):             # <<<<<<<<<<<<<<
 *     if math.isfinite(a):
 *         return a
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":86
 * 
 * 
 * cdef inline double mul_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a * b
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_mul_op(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul_op", 0);

  /* "SparseArray/sparse_array.pxd":87
 * 
 * cdef inline double mul_op(double a, double b):
 *     return a * b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a * __pyx_v_b);
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":86
 * 
 * 
 * cdef inline double mul_op(double a, double b):             # <<<<<<<<<<<<<<
 *     return a * b
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":90
 * 
 * 
 * cdef inline double boundaries_op(double a):             # <<<<<<<<<<<<<<
 *     if a > 1:
 *         return 1
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_boundaries_op(double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("boundaries_op", 0);

  /* "SparseArray/sparse_array.pxd":91
 * 
 * cdef inline double boundaries_op(double a):
 *     if a > 1:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif a < -1:
 */
  __pyx_t_1 = ((__pyx_v_a > 1.0) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":92
 * cdef inline double boundaries_op(double a):
 *     if a > 1:
 *         return 1             # <<<<<<<<<<<<<<
 *     elif a < -1:
 *         return -1
 */
    __pyx_r = 1.0;
    goto __pyx_L0;

    /* "SparseArray/sparse_array.pxd":91
 * 
 * cdef inline double boundaries_op(double a):
 *     if a > 1:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif a < -1:
 */
  }

  /* "SparseArray/sparse_array.pxd":93
 *     if a > 1:
 *         return 1
 *     elif a < -1:             # <<<<<<<<<<<<<<
 *         return -1
 *     return a
 */
  __pyx_t_1 = ((__pyx_v_a < -1.0) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":94
 *         return 1
 *     elif a < -1:
 *         return -1             # <<<<<<<<<<<<<<
 *     return a
 * 
 */
    __pyx_r = -1.0;
    goto __pyx_L0;

    /* "SparseArray/sparse_array.pxd":93
 *     if a > 1:
 *         return 1
 *     elif a < -1:             # <<<<<<<<<<<<<<
 *         return -1
 *     return a
 */
  }

  /* "SparseArray/sparse_array.pxd":95
 *     elif a < -1:
 *         return -1
 *     return a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_a;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":90
 * 
 * 
 * cdef inline double boundaries_op(double a):             # <<<<<<<<<<<<<<
 *     if a > 1:
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":98
 * 
 * 
 * cdef inline double pow_left_op(double a):             # <<<<<<<<<<<<<<
 *     return 1.0
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_pow_left_op(CYTHON_UNUSED double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pow_left_op", 0);

  /* "SparseArray/sparse_array.pxd":99
 * 
 * cdef inline double pow_left_op(double a):
 *     return 1.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1.0;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":98
 * 
 * 
 * cdef inline double pow_left_op(double a):             # <<<<<<<<<<<<<<
 *     return 1.0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":102
 * 
 * 
 * cdef inline double pow_right_op(double a):             # <<<<<<<<<<<<<<
 *     return 0.0
 * 
 */

static CYTHON_INLINE double __pyx_f_11SparseArray_12sparse_array_pow_right_op(CYTHON_UNUSED double __pyx_v_a) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pow_right_op", 0);

  /* "SparseArray/sparse_array.pxd":103
 * 
 * cdef inline double pow_right_op(double a):
 *     return 0.0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "SparseArray/sparse_array.pxd":102
 * 
 * 
 * cdef inline double pow_right_op(double a):             # <<<<<<<<<<<<<<
 *     return 0.0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SparseArray/sparse_array.pxd":106
 * 
 * 
 * cdef inline void set_value(unsigned int *output_index,             # <<<<<<<<<<<<<<
 *                            double *output_data,
 *                            Py_ssize_t *c, unsigned int k,
 */

static CYTHON_INLINE void __pyx_f_11SparseArray_12sparse_array_set_value(unsigned int *__pyx_v_output_index, double *__pyx_v_output_data, Py_ssize_t *__pyx_v_c, unsigned int __pyx_v_k, double __pyx_v_x) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("set_value", 0);

  /* "SparseArray/sparse_array.pxd":110
 *                            Py_ssize_t *c, unsigned int k,
 *                            double x):
 *     if x != 0:             # <<<<<<<<<<<<<<
 *         output_data[c[0]] = x
 *         output_index[c[0]] = k
 */
  __pyx_t_1 = ((__pyx_v_x != 0.0) != 0);
  if (__pyx_t_1) {

    /* "SparseArray/sparse_array.pxd":111
 *                            double x):
 *     if x != 0:
 *         output_data[c[0]] = x             # <<<<<<<<<<<<<<
 *         output_index[c[0]] = k
 *         c[0] = c[0] + 1
 */
    (__pyx_v_output_data[(__pyx_v_c[0])]) = __pyx_v_x;

    /* "SparseArray/sparse_array.pxd":112
 *     if x != 0:
 *         output_data[c[0]] = x
 *         output_index[c[0]] = k             # <<<<<<<<<<<<<<
 *         c[0] = c[0] + 1
 * 
 */
    (__pyx_v_output_index[(__pyx_v_c[0])]) = __pyx_v_k;

    /* "SparseArray/sparse_array.pxd":113
 *         output_data[c[0]] = x
 *         output_index[c[0]] = k
 *         c[0] = c[0] + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_c[0]) = ((__pyx_v_c[0]) + 1);

    /* "SparseArray/sparse_array.pxd":110
 *                            Py_ssize_t *c, unsigned int k,
 *                            double x):
 *     if x != 0:             # <<<<<<<<<<<<<<
 *         output_data[c[0]] = x
 *         output_index[c[0]] = k
 */
  }

  /* "SparseArray/sparse_array.pxd":106
 * 
 * 
 * cdef inline void set_value(unsigned int *output_index,             # <<<<<<<<<<<<<<
 *                            double *output_data,
 *                            Py_ssize_t *c, unsigned int k,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}
static struct __pyx_vtabstruct_6EvoDAG_12cython_utils_SelectNumbers __pyx_vtable_6EvoDAG_12cython_utils_SelectNumbers;

static PyObject *__pyx_tp_new_6EvoDAG_12cython_utils_SelectNumbers(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)o);
  p->__pyx_vtab = __pyx_vtabptr_6EvoDAG_12cython_utils_SelectNumbers;
  p->data = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6EvoDAG_12cython_utils_SelectNumbers(PyObject *o) {
  struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *p = (struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6EvoDAG_12cython_utils_SelectNumbers(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *p = (struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)o;
  if (p->data) {
    e = (*v)(((PyObject *)p->data), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6EvoDAG_12cython_utils_SelectNumbers(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *p = (struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *)o;
  tmp = ((PyObject*)p->data);
  p->data = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_13SelectNumbers_pos(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_13SelectNumbers_pos(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3pos_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_13SelectNumbers_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_13SelectNumbers_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_4size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6EvoDAG_12cython_utils_SelectNumbers[] = {
  {"get", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_3get, METH_O, 0},
  {"get_one", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_5get_one, METH_NOARGS, 0},
  {"empty", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_7empty, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_13SelectNumbers_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6EvoDAG_12cython_utils_SelectNumbers[] = {
  {(char *)"pos", __pyx_getprop_6EvoDAG_12cython_utils_13SelectNumbers_pos, __pyx_setprop_6EvoDAG_12cython_utils_13SelectNumbers_pos, (char *)0, 0},
  {(char *)"size", __pyx_getprop_6EvoDAG_12cython_utils_13SelectNumbers_size, __pyx_setprop_6EvoDAG_12cython_utils_13SelectNumbers_size, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6EvoDAG_12cython_utils_SelectNumbers = {
  PyVarObject_HEAD_INIT(0, 0)
  "EvoDAG.cython_utils.SelectNumbers", /*tp_name*/
  sizeof(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6EvoDAG_12cython_utils_SelectNumbers, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6EvoDAG_12cython_utils_SelectNumbers, /*tp_traverse*/
  __pyx_tp_clear_6EvoDAG_12cython_utils_SelectNumbers, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6EvoDAG_12cython_utils_SelectNumbers, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6EvoDAG_12cython_utils_SelectNumbers, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6EvoDAG_12cython_utils_SelectNumbers, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6EvoDAG_12cython_utils_Score __pyx_vtable_6EvoDAG_12cython_utils_Score;

static PyObject *__pyx_tp_new_6EvoDAG_12cython_utils_Score(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6EvoDAG_12cython_utils_Score *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6EvoDAG_12cython_utils_Score *)o);
  p->__pyx_vtab = __pyx_vtabptr_6EvoDAG_12cython_utils_Score;
  p->precision = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->precision_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->recall = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->recall_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->precision2 = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->precision2_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->recall2 = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  p->recall2_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6EvoDAG_12cython_utils_5Score_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6EvoDAG_12cython_utils_Score(PyObject *o) {
  struct __pyx_obj_6EvoDAG_12cython_utils_Score *p = (struct __pyx_obj_6EvoDAG_12cython_utils_Score *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->precision);
  Py_CLEAR(p->precision_den);
  Py_CLEAR(p->recall);
  Py_CLEAR(p->recall_den);
  Py_CLEAR(p->precision2);
  Py_CLEAR(p->precision2_den);
  Py_CLEAR(p->recall2);
  Py_CLEAR(p->recall2_den);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6EvoDAG_12cython_utils_Score(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6EvoDAG_12cython_utils_Score *p = (struct __pyx_obj_6EvoDAG_12cython_utils_Score *)o;
  if (p->precision) {
    e = (*v)(((PyObject *)p->precision), a); if (e) return e;
  }
  if (p->precision_den) {
    e = (*v)(((PyObject *)p->precision_den), a); if (e) return e;
  }
  if (p->recall) {
    e = (*v)(((PyObject *)p->recall), a); if (e) return e;
  }
  if (p->recall_den) {
    e = (*v)(((PyObject *)p->recall_den), a); if (e) return e;
  }
  if (p->precision2) {
    e = (*v)(((PyObject *)p->precision2), a); if (e) return e;
  }
  if (p->precision2_den) {
    e = (*v)(((PyObject *)p->precision2_den), a); if (e) return e;
  }
  if (p->recall2) {
    e = (*v)(((PyObject *)p->recall2), a); if (e) return e;
  }
  if (p->recall2_den) {
    e = (*v)(((PyObject *)p->recall2_den), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6EvoDAG_12cython_utils_Score(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6EvoDAG_12cython_utils_Score *p = (struct __pyx_obj_6EvoDAG_12cython_utils_Score *)o;
  tmp = ((PyObject*)p->precision);
  p->precision = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->precision_den);
  p->precision_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->recall);
  p->recall = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->recall_den);
  p->recall_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->precision2);
  p->precision2 = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->precision2_den);
  p->precision2_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->recall2);
  p->recall2 = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->recall2_den);
  p->recall2_den = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_precision(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_9precision_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_precision_den(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision_den(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_13precision_den_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_recall(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_6recall_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_recall_den(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall_den(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_10recall_den_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_precision2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_10precision2_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_precision2_den(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision2_den(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_14precision2_den_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_recall2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_7recall2_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score_recall2_den(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall2_den(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_3__set__(o, v);
  }
  else {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_11recall2_den_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score__accuracy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score__accuracy(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_9_accuracy_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6EvoDAG_12cython_utils_5Score__accuracy2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_1__get__(o);
}

static int __pyx_setprop_6EvoDAG_12cython_utils_5Score__accuracy2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6EvoDAG_12cython_utils_5Score_10_accuracy2_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6EvoDAG_12cython_utils_Score[] = {
  {"g_recall_precision", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_3g_recall_precision, METH_VARARGS|METH_KEYWORDS, 0},
  {"F1", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_5F1, METH_VARARGS|METH_KEYWORDS, 0},
  {"macroRecallF1", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_7macroRecallF1, METH_VARARGS|METH_KEYWORDS, 0},
  {"a_recall", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_9a_recall, METH_VARARGS|METH_KEYWORDS, 0},
  {"a_precision", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_11a_precision, METH_VARARGS|METH_KEYWORDS, 0},
  {"g_precision", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_13g_precision, METH_VARARGS|METH_KEYWORDS, 0},
  {"g_recall", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_15g_recall, METH_VARARGS|METH_KEYWORDS, 0},
  {"accuracy", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_17accuracy, METH_VARARGS|METH_KEYWORDS, 0},
  {"accDotMacroF1", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_19accDotMacroF1, METH_VARARGS|METH_KEYWORDS, 0},
  {"g_g_recall_precision", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_21g_g_recall_precision, METH_VARARGS|METH_KEYWORDS, 0},
  {"g_F1", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_23g_F1, METH_VARARGS|METH_KEYWORDS, 0},
  {"a_F1", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5Score_25a_F1, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_5Score_27__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6EvoDAG_12cython_utils_5Score_29__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6EvoDAG_12cython_utils_Score[] = {
  {(char *)"precision", __pyx_getprop_6EvoDAG_12cython_utils_5Score_precision, __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision, (char *)0, 0},
  {(char *)"precision_den", __pyx_getprop_6EvoDAG_12cython_utils_5Score_precision_den, __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision_den, (char *)0, 0},
  {(char *)"recall", __pyx_getprop_6EvoDAG_12cython_utils_5Score_recall, __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall, (char *)0, 0},
  {(char *)"recall_den", __pyx_getprop_6EvoDAG_12cython_utils_5Score_recall_den, __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall_den, (char *)0, 0},
  {(char *)"precision2", __pyx_getprop_6EvoDAG_12cython_utils_5Score_precision2, __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision2, (char *)0, 0},
  {(char *)"precision2_den", __pyx_getprop_6EvoDAG_12cython_utils_5Score_precision2_den, __pyx_setprop_6EvoDAG_12cython_utils_5Score_precision2_den, (char *)0, 0},
  {(char *)"recall2", __pyx_getprop_6EvoDAG_12cython_utils_5Score_recall2, __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall2, (char *)0, 0},
  {(char *)"recall2_den", __pyx_getprop_6EvoDAG_12cython_utils_5Score_recall2_den, __pyx_setprop_6EvoDAG_12cython_utils_5Score_recall2_den, (char *)0, 0},
  {(char *)"_accuracy", __pyx_getprop_6EvoDAG_12cython_utils_5Score__accuracy, __pyx_setprop_6EvoDAG_12cython_utils_5Score__accuracy, (char *)0, 0},
  {(char *)"_accuracy2", __pyx_getprop_6EvoDAG_12cython_utils_5Score__accuracy2, __pyx_setprop_6EvoDAG_12cython_utils_5Score__accuracy2, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6EvoDAG_12cython_utils_Score = {
  PyVarObject_HEAD_INIT(0, 0)
  "EvoDAG.cython_utils.Score", /*tp_name*/
  sizeof(struct __pyx_obj_6EvoDAG_12cython_utils_Score), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6EvoDAG_12cython_utils_Score, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6EvoDAG_12cython_utils_Score, /*tp_traverse*/
  __pyx_tp_clear_6EvoDAG_12cython_utils_Score, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6EvoDAG_12cython_utils_Score, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6EvoDAG_12cython_utils_Score, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6EvoDAG_12cython_utils_Score, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"fitness_SSE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_1fitness_SSE, METH_VARARGS|METH_KEYWORDS, 0},
  {"fitness_SAE", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_3fitness_SAE, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes_isfinite", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_5naive_bayes_isfinite, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes_isfinite_MN", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_7naive_bayes_isfinite_MN, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes_mean_std2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_9naive_bayes_mean_std2, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_11naive_bayes, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes_Nc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_13naive_bayes_Nc, METH_VARARGS|METH_KEYWORDS, 0},
  {"naive_bayes_MN", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_15naive_bayes_MN, METH_VARARGS|METH_KEYWORDS, 0},
  {"centroid_coef", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6EvoDAG_12cython_utils_17centroid_coef, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cython_utils(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cython_utils},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cython_utils",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_Score, __pyx_k_Score, sizeof(__pyx_k_Score), 0, 0, 1, 1},
  {&__pyx_n_s_SelectNumbers, __pyx_k_SelectNumbers, sizeof(__pyx_k_SelectNumbers), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_coef, __pyx_k_coef, sizeof(__pyx_k_coef), 0, 0, 1, 1},
  {&__pyx_n_s_constant, __pyx_k_constant, sizeof(__pyx_k_constant), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_one, __pyx_k_get_one, sizeof(__pyx_k_get_one), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hy, __pyx_k_hy, sizeof(__pyx_k_hy), 0, 0, 1, 1},
  {&__pyx_n_s_hy_2, __pyx_k_hy_2, sizeof(__pyx_k_hy_2), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_klass, __pyx_k_klass, sizeof(__pyx_k_klass), 0, 0, 1, 1},
  {&__pyx_n_s_lst, __pyx_k_lst, sizeof(__pyx_k_lst), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask_2, __pyx_k_mask_2, sizeof(__pyx_k_mask_2), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_nclass, __pyx_k_nclass, sizeof(__pyx_k_nclass), 0, 0, 1, 1},
  {&__pyx_n_s_nclasses, __pyx_k_nclasses, sizeof(__pyx_k_nclasses), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shuffle, __pyx_k_shuffle, sizeof(__pyx_k_shuffle), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_var, __pyx_k_var, sizeof(__pyx_k_var), 0, 0, 1, 1},
  {&__pyx_n_s_vars, __pyx_k_vars, sizeof(__pyx_k_vars), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_ytr, __pyx_k_ytr, sizeof(__pyx_k_ytr), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 39, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 109, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6EvoDAG_12cython_utils_SelectNumbers = &__pyx_vtable_6EvoDAG_12cython_utils_SelectNumbers;
  __pyx_vtable_6EvoDAG_12cython_utils_SelectNumbers.get = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get;
  __pyx_vtable_6EvoDAG_12cython_utils_SelectNumbers.get_one = (int (*)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, int __pyx_skip_dispatch))__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_get_one;
  __pyx_vtable_6EvoDAG_12cython_utils_SelectNumbers.empty = (int (*)(struct __pyx_obj_6EvoDAG_12cython_utils_SelectNumbers *, int __pyx_skip_dispatch))__pyx_f_6EvoDAG_12cython_utils_13SelectNumbers_empty;
  if (PyType_Ready(&__pyx_type_6EvoDAG_12cython_utils_SelectNumbers) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_type_6EvoDAG_12cython_utils_SelectNumbers.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6EvoDAG_12cython_utils_SelectNumbers.tp_dictoffset && __pyx_type_6EvoDAG_12cython_utils_SelectNumbers.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6EvoDAG_12cython_utils_SelectNumbers.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6EvoDAG_12cython_utils_SelectNumbers.tp_dict, __pyx_vtabptr_6EvoDAG_12cython_utils_SelectNumbers) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SelectNumbers, (PyObject *)&__pyx_type_6EvoDAG_12cython_utils_SelectNumbers) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6EvoDAG_12cython_utils_SelectNumbers) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_ptype_6EvoDAG_12cython_utils_SelectNumbers = &__pyx_type_6EvoDAG_12cython_utils_SelectNumbers;
  __pyx_vtabptr_6EvoDAG_12cython_utils_Score = &__pyx_vtable_6EvoDAG_12cython_utils_Score;
  __pyx_vtable_6EvoDAG_12cython_utils_Score._cleanI = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, arrayobject *))__pyx_f_6EvoDAG_12cython_utils_5Score__cleanI;
  __pyx_vtable_6EvoDAG_12cython_utils_Score._cleand = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, arrayobject *))__pyx_f_6EvoDAG_12cython_utils_5Score__cleand;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.clean = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *))__pyx_f_6EvoDAG_12cython_utils_5Score_clean;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.get_klass = (double (*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, double *, unsigned int *, unsigned int))__pyx_f_6EvoDAG_12cython_utils_5Score_get_klass;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.get_predicted = (double (*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, double *, unsigned int *, unsigned int))__pyx_f_6EvoDAG_12cython_utils_5Score_get_predicted;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.recall2accuracy = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *))__pyx_f_6EvoDAG_12cython_utils_5Score_recall2accuracy;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.precision_recall = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *))__pyx_f_6EvoDAG_12cython_utils_5Score_precision_recall;
  __pyx_vtable_6EvoDAG_12cython_utils_Score.count = (PyObject *(*)(struct __pyx_obj_6EvoDAG_12cython_utils_Score *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, struct __pyx_obj_11SparseArray_12sparse_array_SparseArray *, arrayobject *))__pyx_f_6EvoDAG_12cython_utils_5Score_count;
  if (PyType_Ready(&__pyx_type_6EvoDAG_12cython_utils_Score) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __pyx_type_6EvoDAG_12cython_utils_Score.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6EvoDAG_12cython_utils_Score.tp_dictoffset && __pyx_type_6EvoDAG_12cython_utils_Score.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6EvoDAG_12cython_utils_Score.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6EvoDAG_12cython_utils_Score.tp_dict, __pyx_vtabptr_6EvoDAG_12cython_utils_Score) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Score, (PyObject *)&__pyx_type_6EvoDAG_12cython_utils_Score) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6EvoDAG_12cython_utils_Score) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __pyx_ptype_6EvoDAG_12cython_utils_Score = &__pyx_type_6EvoDAG_12cython_utils_Score;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType(__pyx_t_1, "array", "array", sizeof(arrayobject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(2, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("SparseArray.sparse_array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_11SparseArray_12sparse_array_SparseArray = __Pyx_ImportType(__pyx_t_1, "SparseArray.sparse_array", "SparseArray", sizeof(struct __pyx_obj_11SparseArray_12sparse_array_SparseArray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_11SparseArray_12sparse_array_SparseArray) __PYX_ERR(6, 116, __pyx_L1_error)
  __pyx_vtabptr_11SparseArray_12sparse_array_SparseArray = (struct __pyx_vtabstruct_11SparseArray_12sparse_array_SparseArray*)__Pyx_GetVtable(__pyx_ptype_11SparseArray_12sparse_array_SparseArray->tp_dict); if (unlikely(!__pyx_vtabptr_11SparseArray_12sparse_array_SparseArray)) __PYX_ERR(6, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcython_utils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcython_utils(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cython_utils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cython_utils(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cython_utils(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cython_utils' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cython_utils(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cython_utils", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_EvoDAG__cython_utils) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "EvoDAG.cython_utils")) {
      if (unlikely(PyDict_SetItemString(modules, "EvoDAG.cython_utils", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "EvoDAG/cython_utils.pyx":21
 * from cpython cimport list
 * from libc cimport math
 * import random             # <<<<<<<<<<<<<<
 * cimport cython
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EvoDAG/cython_utils.pyx":1
 * # Copyright 2017 Mario Graff Guerrero             # <<<<<<<<<<<<<<
 * 
 * # Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SparseArray/sparse_array.pxd":106
 * 
 * 
 * cdef inline void set_value(unsigned int *output_index,             # <<<<<<<<<<<<<<
 *                            double *output_data,
 *                            Py_ssize_t *c, unsigned int k,
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init EvoDAG.cython_utils", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init EvoDAG.cython_utils");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
